@@ -178,18 +178,10 @@ namespace System.Web.UI.WebControls
 							string target = spot.Target.Length > 0 ? spot.Target : Target;
 							if (!String.IsNullOrEmpty (target))
 								writer.AddAttribute (HtmlTextWriterAttribute.Target, target);
-#if NET_4_0
 							if (enabled) {
-#endif
-#if TARGET_J2EE
-								string navUrl = ResolveClientUrl (spot.NavigateUrl, String.Compare (target, "_blank", StringComparison.InvariantCultureIgnoreCase) != 0);
-#else
 								string navUrl = ResolveClientUrl (spot.NavigateUrl);
-#endif
 								writer.AddAttribute (HtmlTextWriterAttribute.Href, navUrl);
-#if NET_4_0
 							}
-#endif
 							break;
 						case HotSpotMode.PostBack:
 							writer.AddAttribute (HtmlTextWriterAttribute.Href, Page.ClientScript.GetPostBackClientHyperlink (this, n.ToString(), true));
