@@ -27,6 +27,8 @@ using NUnit.Framework;
 
 using MonoTests.System.Xml.TestClasses;
 
+using MonoTests.Helpers;
+
 namespace MonoTests.System.XmlSerialization
 {
 	[TestFixture]
@@ -46,7 +48,7 @@ namespace MonoTests.System.XmlSerialization
 			Assert.AreEqual (string.Format(CultureInfo.InvariantCulture,
 				"{0}{0}" +
 				"/// <remarks/>{0}" +
-				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"System.Xml\", \"{1}\")]{0}" +
+				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"nunit-lite-console\", \"0.0.0.0\")]{0}" +
 				"[System.SerializableAttribute()]{0}" +
 				"[System.Diagnostics.DebuggerStepThroughAttribute()]{0}" +
 				"[System.ComponentModel.DesignerCategoryAttribute(\"code\")]{0}" +
@@ -64,7 +66,7 @@ namespace MonoTests.System.XmlSerialization
 				"            this.namesField = value;{0}" +
 				"        }}{0}" +
 				"    }}{0}" +
-				"}}{0}", Environment.NewLine, XmlFileVersion), sw.ToString (), "#2");
+				"}}{0}", Environment.NewLine), sw.ToString (), "#2");
 
 
 			codeNamespace = ExportCode (typeof (ArrayClass[]));
@@ -76,7 +78,7 @@ namespace MonoTests.System.XmlSerialization
 			Assert.AreEqual (string.Format (CultureInfo.InvariantCulture,
 				"{0}{0}" +
 				"/// <remarks/>{0}" +
-				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"System.Xml\", \"{1}\")]{0}" +
+				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"nunit-lite-console\", \"0.0.0.0\")]{0}" +
 				"[System.SerializableAttribute()]{0}" +
 				"[System.Diagnostics.DebuggerStepThroughAttribute()]{0}" +
 				"[System.ComponentModel.DesignerCategoryAttribute(\"code\")]{0}" +
@@ -93,7 +95,7 @@ namespace MonoTests.System.XmlSerialization
 				"            this.namesField = value;{0}" +
 				"        }}{0}" +
 				"    }}{0}" +
-				"}}{0}", Environment.NewLine, XmlFileVersion), sw.ToString (), "#4");
+				"}}{0}", Environment.NewLine), sw.ToString (), "#4");
 		}
 
 		[Test]
@@ -110,7 +112,7 @@ namespace MonoTests.System.XmlSerialization
 			Assert.AreEqual (string.Format (CultureInfo.InvariantCulture,
 				"{0}{0}" +
 				"/// <remarks/>{0}" +
-				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"System.Xml\", \"{1}\")]{0}" +
+				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"nunit-lite-console\", \"0.0.0.0\")]{0}" +
 				"[System.SerializableAttribute()]{0}" +
 				"[System.Diagnostics.DebuggerStepThroughAttribute()]{0}" +
 				"[System.ComponentModel.DesignerCategoryAttribute(\"code\")]{0}" +
@@ -128,7 +130,7 @@ namespace MonoTests.System.XmlSerialization
 				"            this.itemsField = value;{0}" +
 				"        }}{0}" +
 				"    }}{0}" +
-				"}}{0}", Environment.NewLine, XmlFileVersion), sw.ToString (), "#2");
+				"}}{0}", Environment.NewLine), sw.ToString (), "#2");
 		}
 
 		[Test]
@@ -145,7 +147,7 @@ namespace MonoTests.System.XmlSerialization
 			Assert.AreEqual (string.Format (CultureInfo.InvariantCulture,
 				"{0}{0}" +
 				"/// <remarks/>{0}" +
-				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"System.Xml\", \"{1}\")]{0}" +
+				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"nunit-lite-console\", \"0.0.0.0\")]{0}" +
 				"[System.SerializableAttribute()]{0}" +
 				"[System.Diagnostics.DebuggerStepThroughAttribute()]{0}" +
 				"[System.ComponentModel.DesignerCategoryAttribute(\"code\")]{0}" +
@@ -163,7 +165,7 @@ namespace MonoTests.System.XmlSerialization
 				"            this.cdataField = value;{0}" +
 				"        }}{0}" +
 				"    }}{0}" +
-				"}}{0}", Environment.NewLine, XmlFileVersion), sw.ToString (), "#2");
+				"}}{0}", Environment.NewLine), sw.ToString (), "#2");
 		}
 
 		[Test]
@@ -182,7 +184,7 @@ namespace MonoTests.System.XmlSerialization
 			Assert.AreEqual (string.Format (CultureInfo.InvariantCulture,
 				"{0}{0}" +
 				"/// <remarks/>{0}" +
-				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"System.Xml\", \"{1}\")]{0}" +
+				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"nunit-lite-console\", \"0.0.0.0\")]{0}" +
 				"[System.SerializableAttribute()]{0}" +
 				"[System.Diagnostics.DebuggerStepThroughAttribute()]{0}" +
 				"[System.ComponentModel.DesignerCategoryAttribute(\"code\")]{0}" +
@@ -204,7 +206,7 @@ namespace MonoTests.System.XmlSerialization
 				"            this.myChoiceField = value;{0}" +
 				"        }}{0}" +
 				"    }}{0}" +
-				"}}{0}", Environment.NewLine, XmlFileVersion), sw.ToString (), "#2");
+				"}}{0}", Environment.NewLine), sw.ToString (), "#2");
 		}
 
 		[Test]
@@ -222,7 +224,7 @@ namespace MonoTests.System.XmlSerialization
 			Assert.AreEqual (string.Format (CultureInfo.InvariantCulture,
 				"{0}{0}" +
 				"/// <remarks/>{0}" +
-				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"System.Xml\", \"{1}\")]{0}" +
+				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"nunit-lite-console\", \"0.0.0.0\")]{0}" +
 				"[System.SerializableAttribute()]{0}" +
 				"[System.Diagnostics.DebuggerStepThroughAttribute()]{0}" +
 				"[System.ComponentModel.DesignerCategoryAttribute(\"code\")]{0}" +
@@ -403,7 +405,7 @@ namespace MonoTests.System.XmlSerialization
 				"{0}" +
 				"/// <remarks/>{0}" +
 				"[System.FlagsAttribute()]{0}" +
-				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"System.Xml\", \"{1}\")]{0}" +
+				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"nunit-lite-console\", \"0.0.0.0\")]{0}" +
 				"[System.SerializableAttribute()]{0}" +
 				"public enum MapModifiers {{{0}" +
 				"    {0}" +
@@ -414,7 +416,7 @@ namespace MonoTests.System.XmlSerialization
 				"    /// <remarks/>{0}" +
 				"    [System.Xml.Serialization.XmlEnumAttribute(\"protected\")]{0}" +
 				"    Protected = 2,{0}" +
-				"}}{0}", Environment.NewLine, XmlFileVersion), sw.ToString (), "#2");
+				"}}{0}", Environment.NewLine), sw.ToString (), "#2");
 		}
 
 		[Test]
@@ -431,7 +433,7 @@ namespace MonoTests.System.XmlSerialization
 			Assert.AreEqual (string.Format (CultureInfo.InvariantCulture,
 				"{0}{0}" +
 				"/// <remarks/>{0}" +
-				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"System.Xml\", \"{1}\")]{0}" +
+				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"nunit-lite-console\", \"0.0.0.0\")]{0}" +
 				"[System.SerializableAttribute()]{0}" +
 				"[System.Xml.Serialization.XmlTypeAttribute(IncludeInSchema=false)]{0}" +
 				"[System.Xml.Serialization.XmlRootAttribute(Namespace=\"\", IsNullable=false)]{0}" +
@@ -446,7 +448,7 @@ namespace MonoTests.System.XmlSerialization
 				"    {0}" +
 				"    /// <remarks/>{0}" +
 				"    ChoiceTwo,{0}" +
-				"}}{0}", Environment.NewLine, XmlFileVersion), sw.ToString (), "#2");
+				"}}{0}", Environment.NewLine), sw.ToString (), "#2");
 
 			codeNamespace = ExportCode (typeof (ItemChoiceType[]));
 			Assert.IsNotNull (codeNamespace, "#3");
@@ -457,7 +459,7 @@ namespace MonoTests.System.XmlSerialization
 			Assert.AreEqual (string.Format (CultureInfo.InvariantCulture,
 				"{0}{0}" +
 				"/// <remarks/>{0}" +
-				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"System.Xml\", \"{1}\")]{0}" +
+				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"nunit-lite-console\", \"0.0.0.0\")]{0}" +
 				"[System.SerializableAttribute()]{0}" +
 				"[System.Xml.Serialization.XmlTypeAttribute(IncludeInSchema=false)]{0}" +
 				"public enum ItemChoiceType {{{0}" +
@@ -471,7 +473,7 @@ namespace MonoTests.System.XmlSerialization
 				"    {0}" +
 				"    /// <remarks/>{0}" +
 				"    ChoiceTwo,{0}" +
-				"}}{0}", Environment.NewLine, XmlFileVersion), sw.ToString (), "#4");
+				"}}{0}", Environment.NewLine), sw.ToString (), "#4");
 		}
 
 		[Test]
@@ -488,7 +490,7 @@ namespace MonoTests.System.XmlSerialization
 			Assert.AreEqual (string.Format (CultureInfo.InvariantCulture,
 				"{0}{0}" +
 				"/// <remarks/>{0}" +
-				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"System.Xml\", \"{1}\")]{0}" +
+				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"nunit-lite-console\", \"0.0.0.0\")]{0}" +
 				"[System.SerializableAttribute()]{0}" +
 				"[System.Diagnostics.DebuggerStepThroughAttribute()]{0}" +
 				"[System.ComponentModel.DesignerCategoryAttribute(\"code\")]{0}" +
@@ -509,7 +511,7 @@ namespace MonoTests.System.XmlSerialization
 				"}}{0}" +
 				"{0}" +
 				"/// <remarks/>{0}" +
-				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"System.Xml\", \"{1}\")]{0}" +
+				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"nunit-lite-console\", \"0.0.0.0\")]{0}" +
 				"[System.SerializableAttribute()]{0}" +
 				"[System.Diagnostics.DebuggerStepThroughAttribute()]{0}" +
 				"[System.ComponentModel.DesignerCategoryAttribute(\"code\")]{0}" +
@@ -526,7 +528,7 @@ namespace MonoTests.System.XmlSerialization
 				"            this.somethingField = value;{0}" +
 				"        }}{0}" +
 				"    }}{0}" +
-				"}}{0}", Environment.NewLine, XmlFileVersion), sw.ToString (), "#2");
+				"}}{0}", Environment.NewLine), sw.ToString (), "#2");
 		}
 
 		[Test]
@@ -544,7 +546,7 @@ namespace MonoTests.System.XmlSerialization
 			Assert.AreEqual (string.Format (CultureInfo.InvariantCulture,
 				"{0}{0}" +
 				"/// <remarks/>{0}" +
-				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"System.Xml\", \"{1}\")]{0}" +
+				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"nunit-lite-console\", \"0.0.0.0\")]{0}" +
 				"[System.SerializableAttribute()]{0}" +
 				"[System.Diagnostics.DebuggerStepThroughAttribute()]{0}" +
 				"[System.ComponentModel.DesignerCategoryAttribute(\"code\")]{0}" +
@@ -578,7 +580,7 @@ namespace MonoTests.System.XmlSerialization
 				"}}{0}" +
 				"{0}" +
 				"/// <remarks/>{0}" +
-				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"System.Xml\", \"{1}\")]{0}" +
+				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"nunit-lite-console\", \"0.0.0.0\")]{0}" +
 				"[System.SerializableAttribute()]{0}" +
 				"[System.Diagnostics.DebuggerStepThroughAttribute()]{0}" +
 				"[System.ComponentModel.DesignerCategoryAttribute(\"code\")]{0}" +
@@ -667,7 +669,7 @@ namespace MonoTests.System.XmlSerialization
 				"{0}" +
 				"/// <remarks/>{0}" +
 				"[System.FlagsAttribute()]{0}" +
-				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"System.Xml\", \"{1}\")]{0}" +
+				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"nunit-lite-console\", \"0.0.0.0\")]{0}" +
 				"[System.SerializableAttribute()]{0}" +
 				"[System.Xml.Serialization.XmlTypeAttribute(Namespace=\"some:urn\")]{0}" +
 				"public enum FlagEnum {{{0}" +
@@ -786,7 +788,7 @@ namespace MonoTests.System.XmlSerialization
 				"{0}" +
 				"/// <remarks/>{0}" +
 				"[System.FlagsAttribute()]{0}" +
-				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"System.Xml\", \"{1}\")]{0}" +
+				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"nunit-lite-console\", \"0.0.0.0\")]{0}" +
 				"[System.SerializableAttribute()]{0}" +
 				"[System.Xml.Serialization.XmlTypeAttribute(Namespace=\"urn:myNS\")]{0}" +
 				"public enum FlagEnum {{{0}" +
@@ -806,7 +808,7 @@ namespace MonoTests.System.XmlSerialization
 				"{0}" +
 				"/// <remarks/>{0}" +
 				"[System.FlagsAttribute()]{0}" +
-				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"System.Xml\", \"{1}\")]{0}" +
+				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"nunit-lite-console\", \"0.0.0.0\")]{0}" +
 				"[System.SerializableAttribute()]{0}" +
 				"[System.Xml.Serialization.XmlTypeAttribute(Namespace=\"urn:myNS\")]{0}" +
 				"public enum FlagEnum_Encoded {{{0}" +
@@ -819,7 +821,7 @@ namespace MonoTests.System.XmlSerialization
 				"    {0}" +
 				"    /// <remarks/>{0}" +
 				"    e4 = 4,{0}" +
-				"}}{0}", Environment.NewLine, XmlFileVersion), sw.ToString (), "#2");
+				"}}{0}", Environment.NewLine), sw.ToString (), "#2");
 		}
 
 		[Test]
@@ -836,7 +838,7 @@ namespace MonoTests.System.XmlSerialization
 			Assert.AreEqual (string.Format (CultureInfo.InvariantCulture,
 				"{0}{0}" +
 				"/// <remarks/>{0}" +
-				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"System.Xml\", \"{1}\")]{0}" +
+				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"nunit-lite-console\", \"0.0.0.0\")]{0}" +
 				"[System.SerializableAttribute()]{0}" +
 				"[System.Diagnostics.DebuggerStepThroughAttribute()]{0}" +
 				"[System.ComponentModel.DesignerCategoryAttribute(\"code\")]{0}" +
@@ -855,7 +857,7 @@ namespace MonoTests.System.XmlSerialization
 				"            this.somethingField = value;{0}" +
 				"        }}{0}" +
 				"    }}{0}" +
-				"}}{0}", Environment.NewLine, XmlFileVersion), sw.ToString (), "#2");
+				"}}{0}", Environment.NewLine), sw.ToString (), "#2");
 		}
 
 		[Test]
@@ -909,7 +911,7 @@ namespace MonoTests.System.XmlSerialization
 				"{0}{0}" +
 				"/// <remarks/>{0}" +
 				"[System.FlagsAttribute()]{0}" +
-				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"System.Xml\", \"{1}\")]{0}" +
+				"[System.CodeDom.Compiler.GeneratedCodeAttribute(\"nunit-lite-console\", \"0.0.0.0\")]{0}" +
 				"[System.SerializableAttribute()]{0}" +
 				"[System.Xml.Serialization.XmlRootAttribute(Namespace=\"\", IsNullable=false)]{0}" +
 				"public enum ZeroFlagEnum {{{0}" +
@@ -925,13 +927,13 @@ namespace MonoTests.System.XmlSerialization
 				"    /// <remarks/>{0}" +
 				"    [System.Xml.Serialization.XmlEnumAttribute(\"tns:t<w>o\")]{0}" +
 				"    e2 = 4,{0}" +
-				"}}{0}", Environment.NewLine, XmlFileVersion), sw.ToString (), "#2");
+				"}}{0}", Environment.NewLine), sw.ToString (), "#2");
 		}
 
 		[Test]
 		public void DuplicateIdentifiers ()
 		{
-			XmlSchema xs = XmlSchema.Read (File.OpenText ("Test/XmlFiles/xsd/82078.xsd"), null);
+			XmlSchema xs = XmlSchema.Read (File.OpenText (TestResourceHelper.GetFullPathOfResource ("Test/XmlFiles/xsd/82078.xsd")), null);
 
 			XmlSchemas xss = new XmlSchemas ();
 			xss.Add (xs);
@@ -1005,15 +1007,6 @@ namespace MonoTests.System.XmlSerialization
 			return codeNamespace;
 		}
 
-		string XmlFileVersion {
-			get {
-				Assembly xmlAsm = typeof (XmlDocument).Assembly;
-				AssemblyFileVersionAttribute afv = (AssemblyFileVersionAttribute)
-					Attribute.GetCustomAttribute (xmlAsm, typeof (AssemblyFileVersionAttribute));
-				return afv.Version;
-			}
-		}
-
 		[XmlRootAttribute ("root", Namespace="urn:aNS", IsNullable=false)]
 		public class Root
 		{
