@@ -16,6 +16,8 @@ using System.Xml.XPath;
 using System.Xml.Xsl;
 using NUnit.Framework;
 
+using MonoTests.Helpers;
+
 namespace MonoTests.System.Xml.Xsl
 {
 	[TestFixture]
@@ -40,8 +42,8 @@ namespace MonoTests.System.Xml.Xsl
 		{
 			string output = Path.Combine (Path.GetTempPath (), "result.xml");
 			doc.LoadXml ("<root/>");
-			xslt.Load ("Test/XmlFiles/xsl/empty.xsl");
-			xslt.Transform ("Test/XmlFiles/xsl/empty.xsl", output);
+			xslt.Load (TestResourceHelper.GetFullPathOfResource ("Test/XmlFiles/xsl/empty.xsl"));
+			xslt.Transform (TestResourceHelper.GetFullPathOfResource ("Test/XmlFiles/xsl/empty.xsl"), output);
 			result.Load (output);
 			Assert.AreEqual (2, result.ChildNodes.Count, "count");
 		}
@@ -90,8 +92,8 @@ namespace MonoTests.System.Xml.Xsl
 			xslt.Load (new XPathDocument (new XmlTextReader (xsl, XmlNodeType.Document, null)));
 		}
 
-		[Test()]
-		[Category ("NotWorking")] // it depends on "mcs" existence
+		[Test]
+	    [Category ("MobileNotWorking")]
 		public void MsxslTest() {
 			string _styleSheet = @"
 			<xslt:stylesheet xmlns:xslt=""http://www.w3.org/1999/XSL/Transform"" version=""1.0"" 
@@ -380,15 +382,15 @@ xmlns:xsl='http://www.w3.org/1999/XSL/Transform' xmlns:msxsl='urn:schemas-micros
     ";
 
 			XmlDocument d = new XmlDocument ();
-			d.Load ("Test/XmlFiles/xsl/91834.xml");
+			d.Load (TestResourceHelper.GetFullPathOfResource ("Test/XmlFiles/xsl/91834.xml"));
 
 			XslTransform t = new XslTransform ();
-			t.Load ("Test/XmlFiles/xsl/91834.xsl");
+			t.Load (TestResourceHelper.GetFullPathOfResource ("Test/XmlFiles/xsl/91834.xsl"));
 
 			StringWriter sw_raw = new StringWriter ();
 			t.Transform (d, null, sw_raw);
 
-			Assert.AreEqual (ref_out, sw_raw.ToString ().Replace ("\r\n", "\n"));
+			Assert.AreEqual (ref_out.Replace ("\r\n", "\n"), sw_raw.ToString ().Replace ("\r\n", "\n"));
 		}
 
 		// http://support.microsoft.com/default.aspx?scid=kb;en-us;829014
@@ -1053,7 +1055,6 @@ xmlns:xsl='http://www.w3.org/1999/XSL/Transform' xmlns:msxsl='urn:schemas-micros
 		}
 
 		[Test]
-		[Category ("NotWorking")] // bug #77081: mono does not output newline and indentation for non-html elements
 		public void Output_Indent_Html_DocType ()
 		{
 			XsltArgumentList xsltArgs = new XsltArgumentList ();
@@ -1092,7 +1093,7 @@ xmlns:xsl='http://www.w3.org/1999/XSL/Transform' xmlns:msxsl='urn:schemas-micros
 				"    <else>{0}" +
 				"    </else>{0}" +
 				"  </something>{0}" +
-				"</test>", Environment.NewLine), sw.ToString (), "#1");
+				"</test>", end_of_line), sw.ToString (), "#1");
 
 			// set indent to no
 			sw.GetStringBuilder ().Length = 0;
@@ -1120,7 +1121,7 @@ xmlns:xsl='http://www.w3.org/1999/XSL/Transform' xmlns:msxsl='urn:schemas-micros
 				"    <else>{0}" +
 				"    </else>{0}" +
 				"  </something>{0}" +
-				"</test>", Environment.NewLine), sw.ToString (), "#3");
+				"</test>", end_of_line), sw.ToString (), "#3");
 		}
 
 		[Test]
@@ -1186,7 +1187,6 @@ xmlns:xsl='http://www.w3.org/1999/XSL/Transform' xmlns:msxsl='urn:schemas-micros
 		}
 
 		[Test]
-		[Category ("NotWorking")] // bug #77081: mono does not output newline and indentation for non-html elements
 		public void Output_Indent_Html ()
 		{
 			XsltArgumentList xsltArgs = new XsltArgumentList ();
@@ -1240,7 +1240,7 @@ xmlns:xsl='http://www.w3.org/1999/XSL/Transform' xmlns:msxsl='urn:schemas-micros
 				"      </whatever>{0}" +
 				"    </p>{0}" +
 				"  </body>{0}" +
-				"</html>", Environment.NewLine), sw.ToString (), "#1");
+				"</html>", end_of_line), sw.ToString (), "#1");
 
 			// set indent to no
 			sw.GetStringBuilder ().Length = 0;
@@ -1276,7 +1276,7 @@ xmlns:xsl='http://www.w3.org/1999/XSL/Transform' xmlns:msxsl='urn:schemas-micros
 				"      </whatever>{0}" +
 				"    </p>{0}" +
 				"  </body>{0}" +
-				"</html>", Environment.NewLine), sw.ToString (), "#3");
+				"</html>", end_of_line), sw.ToString (), "#3");
 		}
 
 		[Test]
@@ -1914,9 +1914,9 @@ Services
 ";
 
 			XslTransform t = new XslTransform ();
-			t.Load ("Test/XmlFiles/xsl/stripspace.xsl");
+			t.Load (TestResourceHelper.GetFullPathOfResource ("Test/XmlFiles/xsl/stripspace.xsl"));
 			XPathDocument d = new XPathDocument (
-				"Test/XmlFiles/xsl/stripspace.xml", XmlSpace.Preserve);
+				TestResourceHelper.GetFullPathOfResource ("Test/XmlFiles/xsl/stripspace.xml"), XmlSpace.Preserve);
 			StringWriter sw = new StringWriter ();
 			t.Transform (d, null, sw);
 			Assert.AreEqual (result, sw.ToString ());
@@ -1926,13 +1926,13 @@ Services
 		public void CurrentInSelect ()
 		{
 			XslTransform t = new XslTransform ();
-			t.Load ("Test/XmlFiles/xsl/current-in-select.xsl");
+			t.Load (TestResourceHelper.GetFullPathOfResource ("Test/XmlFiles/xsl/current-in-select.xsl"));
 			XPathDocument d = new XPathDocument (
-				"Test/XmlFiles/xsl/current-in-select.xml", XmlSpace.Preserve);
+				TestResourceHelper.GetFullPathOfResource ("Test/XmlFiles/xsl/current-in-select.xml"), XmlSpace.Preserve);
 			StringWriter sw = new StringWriter ();
 			t.Transform (d, null, sw);
 			using (StreamReader sr = File.OpenText (
-			       "Test/XmlFiles/xsl/current-in-select.ref")) {
+			       TestResourceHelper.GetFullPathOfResource ("Test/XmlFiles/xsl/current-in-select.ref"))) {
 				string s = sr.ReadToEnd ();
 				Assert.AreEqual (s, sw.ToString ());
 			}
@@ -1961,9 +1961,9 @@ Services
 		public void Bug82493 ()
 		{
 			XslTransform transformationObj = new XslTransform ();
-			transformationObj.Load ("Test/XmlFiles/xsl/82493.xsl");
+			transformationObj.Load (TestResourceHelper.GetFullPathOfResource ("Test/XmlFiles/xsl/82493.xsl"));
 			XmlDocument doc = new XmlDocument ();
-			doc.Load ("Test/XmlFiles/xsl/82493.xml");
+			doc.Load (TestResourceHelper.GetFullPathOfResource ("Test/XmlFiles/xsl/82493.xml"));
 			XslExtensionClass GetCalculation = new XslExtensionClass ("a4");
 			XsltArgumentList xslExtensionArgList = new XsltArgumentList ();
 			xslExtensionArgList.AddExtensionObject ("urn:e", GetCalculation);
@@ -1976,9 +1976,9 @@ Services
 		public void BugNovell325482 ()
 		{
 			XmlDocument doc = new XmlDocument ();
-			doc.Load ("Test/XmlFiles/xsl/325482.xml");
+			doc.Load (TestResourceHelper.GetFullPathOfResource ("Test/XmlFiles/xsl/325482.xml"));
 			XslTransform xt = new XslTransform ();
-			xt.Load ("Test/XmlFiles/xsl/325482.xsl");
+			xt.Load (TestResourceHelper.GetFullPathOfResource ("Test/XmlFiles/xsl/325482.xsl"));
 			xt.Transform (doc, null, TextWriter.Null);
 		}
 
@@ -2169,8 +2169,8 @@ Services
 		[Test] // bug #316238
 		public void ResolveVariableInXsltArgumentList ()
 		{
-			ResolveVariableInXsltArgumentList ("Test/XmlFiles/xsl/316238-1.xsl");
-			ResolveVariableInXsltArgumentList ("Test/XmlFiles/xsl/316238-2.xsl");
+			ResolveVariableInXsltArgumentList (TestResourceHelper.GetFullPathOfResource ("Test/XmlFiles/xsl/316238-1.xsl"));
+			ResolveVariableInXsltArgumentList (TestResourceHelper.GetFullPathOfResource ("Test/XmlFiles/xsl/316238-2.xsl"));
 		}
 
 		void ResolveVariableInXsltArgumentList (string file)
@@ -2303,8 +2303,7 @@ NO
 		}
 
 		[Test] // reverse case of #349375
-		[Category ("NotWorking")]
-//		[Category ("NotDotNet")]
+	    [Category ("MobileNotWorking")]
 		public void PreserveWhitespace2 ()
 		{
 			XslCompiledTransform xslt = new XslCompiledTransform ();
@@ -2363,9 +2362,9 @@ NO
 			XslCompiledTransform xslTransform = new XslCompiledTransform ();
 			XsltSettings settings = new XsltSettings ();
 			settings.EnableDocumentFunction = true;
-			xslTransform.Load ("Test/XmlFiles/xsl/391424.xsl", settings, new XmlUrlResolver ());
+			xslTransform.Load (TestResourceHelper.GetFullPathOfResource ("Test/XmlFiles/xsl/391424.xsl"), settings, new XmlUrlResolver ());
 			StringWriter sw = new StringWriter ();
-			xslTransform.Transform (new XPathDocument ("Test/XmlFiles/xsl/391424.xml", XmlSpace.Preserve), null, sw);
+			xslTransform.Transform (new XPathDocument (TestResourceHelper.GetFullPathOfResource ("Test/XmlFiles/xsl/391424.xml"), XmlSpace.Preserve), null, sw);
 			Assert.AreEqual ("<?xml version=\"1.0\" encoding=\"utf-16\"?>Document found", sw.ToString ());
 		}
 
@@ -2388,7 +2387,7 @@ NO
 		}
 
 		[Test]
-		[Category ("NotWorking")] // FIXME: SRE related regression
+	    [Category ("MobileNotWorking")]
 		public void Bug487065 ()
 		{
 			using (XmlReader input = GetInput ()) {
