@@ -40,7 +40,7 @@ using System.Security.Permissions;
 
 namespace System.Security.Cryptography.X509Certificates {
 
-	public sealed class X509Store {
+	public sealed class X509Store : IDisposable {
 
 		private string _name;
 		private StoreLocation _location;
@@ -208,6 +208,11 @@ namespace System.Security.Cryptography.X509Certificates {
 				list.Clear ();
 		}
 
+		public void Dispose ()
+		{
+			Close ();
+		}
+
 		public void Open (OpenFlags flags)
 		{
 			if (String.IsNullOrEmpty (_name))
@@ -232,7 +237,7 @@ namespace System.Security.Cryptography.X509Certificates {
 
 			foreach (MX.X509Certificate x in store.Certificates) {
 				var cert2 = new X509Certificate2 (x.RawData);
-				cert2.PrivateKey = x.RSA;
+				cert2.Impl.PrivateKey = x.RSA;
 				Certificates.Add (cert2);
 			}
 		}
