@@ -138,22 +138,22 @@ namespace System
 			return m_serializationCtor;
 		}
 
-		internal Object CreateInstanceSlow(bool publicOnly, bool skipCheckThis, bool fillCache, ref StackCrawlMark stackMark)
+		internal Object CreateInstanceSlow(bool publicOnly, bool wrapExceptions, bool skipCheckThis, bool fillCache, ref StackCrawlMark stackMark)
 		{
-			bool bNeedSecurityCheck = true;
-			bool bCanBeCached = false;
-			bool bSecurityCheckOff = false;
+			//bool bNeedSecurityCheck = true;
+			//bool bCanBeCached = false;
+			//bool bSecurityCheckOff = false;
 
 			if (!skipCheckThis)
 				CreateInstanceCheckThis();
 
-			if (!fillCache)
-				bSecurityCheckOff = true;
+			//if (!fillCache)
+			//	bSecurityCheckOff = true;
 
-			return CreateInstanceMono (!publicOnly);
+			return CreateInstanceMono (!publicOnly, wrapExceptions);
 		}
 
-		object CreateInstanceMono (bool nonPublic)
+		object CreateInstanceMono (bool nonPublic, bool wrapExceptions)
 		{
 			var ctor = GetDefaultConstructor ();
 			if (!nonPublic && ctor != null && !ctor.IsPublic) {
@@ -176,7 +176,7 @@ namespace System
 				throw new MissingMethodException (Locale.GetText ("Cannot create an abstract class '{0}'.", FullName));
 			}
 
-			return ctor.InternalInvoke (null, null);
+			return ctor.InternalInvoke (null, null, wrapExceptions);
 		}
 
 		internal Object CheckValue (Object value, Binder binder, CultureInfo culture, BindingFlags invokeAttr)
@@ -465,20 +465,20 @@ namespace System
 		{
 			var gt = (RuntimeType) MakeGenericType (genericType, new Type [] { genericArgument });
 			var ctor = gt.GetDefaultConstructor ();
-			return ctor.InternalInvoke (null, null);
+			return ctor.InternalInvoke (null, null, wrapExceptions: true);
 		}
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		static extern Type MakeGenericType (Type gt, Type [] types);
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		internal extern IntPtr GetMethodsByName_native (IntPtr namePtr, BindingFlags bindingAttr, bool ignoreCase);
+		internal extern IntPtr GetMethodsByName_native (IntPtr namePtr, BindingFlags bindingAttr, MemberListType listType);
 
-		internal RuntimeMethodInfo[] GetMethodsByName (string name, BindingFlags bindingAttr, bool ignoreCase, RuntimeType reflectedType)
+		internal RuntimeMethodInfo[] GetMethodsByName (string name, BindingFlags bindingAttr, MemberListType listType, RuntimeType reflectedType)
 		{
 			var refh = new RuntimeTypeHandle (reflectedType);
 			using (var namePtr = new Mono.SafeStringMarshal (name))
-			using (var h = new Mono.SafeGPtrArrayHandle (GetMethodsByName_native (namePtr.Value, bindingAttr, ignoreCase))) {
+			using (var h = new Mono.SafeGPtrArrayHandle (GetMethodsByName_native (namePtr.Value, bindingAttr, listType))) {
 				var n = h.Length;
 				var a = new RuntimeMethodInfo [n];
 				for (int i = 0; i < n; i++) {
@@ -490,7 +490,7 @@ namespace System
 		}
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		extern IntPtr GetPropertiesByName_native (IntPtr name, BindingFlags bindingAttr, bool icase);		
+		extern IntPtr GetPropertiesByName_native (IntPtr name, BindingFlags bindingAttr, MemberListType listType);
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		extern IntPtr GetConstructors_native (BindingFlags bindingAttr);
@@ -509,16 +509,16 @@ namespace System
 			}
 		}
 
-		RuntimePropertyInfo[] GetPropertiesByName (string name, BindingFlags bindingAttr, bool icase, RuntimeType reflectedType)
+		RuntimePropertyInfo[] GetPropertiesByName (string name, BindingFlags bindingAttr, MemberListType listType, RuntimeType reflectedType)
 		{
 			var refh = new RuntimeTypeHandle (reflectedType);
 			using (var namePtr = new Mono.SafeStringMarshal (name))
-			using (var h = new Mono.SafeGPtrArrayHandle (GetPropertiesByName_native (namePtr.Value, bindingAttr, icase))) {
+			using (var h = new Mono.SafeGPtrArrayHandle (GetPropertiesByName_native (namePtr.Value, bindingAttr, listType))) {
 				var n = h.Length;
 				var a = new RuntimePropertyInfo [n];
 				for (int i = 0; i < n; i++) {
 					var ph = new Mono.RuntimePropertyHandle (h[i]);
-					a[i] = (RuntimePropertyInfo) PropertyInfo.GetPropertyFromHandle (ph, refh);
+					a[i] = (RuntimePropertyInfo) MonoProperty.GetPropertyFromHandle (ph, refh);
 				}
 				return a;
 			}
@@ -592,8 +592,10 @@ namespace System
 				{
 					AssemblyName assemblyname = new AssemblyName ();
 					assemblyname.Name = "GetTypeFromCLSIDDummyAssembly";
-					clsid_assemblybuilder = AppDomain.CurrentDomain.DefineDynamicAssembly (
-						assemblyname, AssemblyBuilderAccess.Run);
+					/* Dynamically created assembly is marked internal to corlib to allow
+					  __ComObject access for dynamic types. */
+					clsid_assemblybuilder = new AssemblyBuilder (assemblyname, null,
+						AssemblyBuilderAccess.Run, true);
 				}
 				ModuleBuilder modulebuilder = clsid_assemblybuilder.DefineDynamicModule (
 					clsid.ToString ());
@@ -671,16 +673,16 @@ namespace System
 		extern int GetGenericParameterPosition ();
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		extern IntPtr GetEvents_native (IntPtr name, BindingFlags bindingAttr);
+		extern IntPtr GetEvents_native (IntPtr name, BindingFlags bindingAttr,  MemberListType listType);
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		extern IntPtr GetFields_native (IntPtr name, BindingFlags bindingAttr);
+		extern IntPtr GetFields_native (IntPtr name, BindingFlags bindingAttr, MemberListType listType);
 
-		RuntimeFieldInfo[] GetFields_internal (string name, BindingFlags bindingAttr, RuntimeType reflectedType)
+		RuntimeFieldInfo[] GetFields_internal (string name, BindingFlags bindingAttr, MemberListType listType, RuntimeType reflectedType)
 		{
 			var refh = new RuntimeTypeHandle (reflectedType);
 			using (var namePtr = new Mono.SafeStringMarshal (name))
-			using (var h = new Mono.SafeGPtrArrayHandle (GetFields_native (namePtr.Value, bindingAttr))) {
+			using (var h = new Mono.SafeGPtrArrayHandle (GetFields_native (namePtr.Value, bindingAttr, listType))) {
 				int n = h.Length;
 				var a = new RuntimeFieldInfo[n];
 				for (int i = 0; i < n; i++) {
@@ -691,11 +693,11 @@ namespace System
 			}
 		}
 
-		RuntimeEventInfo[] GetEvents_internal (string name, BindingFlags bindingAttr, RuntimeType reflectedType)
+		RuntimeEventInfo[] GetEvents_internal (string name, BindingFlags bindingAttr, MemberListType listType, RuntimeType reflectedType)
 		{
 			var refh = new RuntimeTypeHandle (reflectedType);
 			using (var namePtr = new Mono.SafeStringMarshal (name))
-			using (var h = new Mono.SafeGPtrArrayHandle (GetEvents_native (namePtr.Value, bindingAttr))) {
+			using (var h = new Mono.SafeGPtrArrayHandle (GetEvents_native (namePtr.Value, bindingAttr, listType))) {
 				int n = h.Length;
 				var a = new RuntimeEventInfo[n];
 				for (int i = 0; i < n; i++) {
@@ -710,15 +712,15 @@ namespace System
 		public extern override Type[] GetInterfaces();
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		extern IntPtr GetNestedTypes_native (IntPtr name, BindingFlags bindingAttr);
+		extern IntPtr GetNestedTypes_native (IntPtr name, BindingFlags bindingAttr, MemberListType listType);
 
-		RuntimeType[] GetNestedTypes_internal (string displayName, BindingFlags bindingAttr)
+		RuntimeType[] GetNestedTypes_internal (string displayName, BindingFlags bindingAttr, MemberListType listType)
 		{
 			string internalName = null;
 			if (displayName != null)
 				internalName = TypeIdentifiers.FromDisplay (displayName).InternalName;
 			using (var namePtr = new Mono.SafeStringMarshal (internalName))
-			using (var h = new Mono.SafeGPtrArrayHandle (GetNestedTypes_native (namePtr.Value, bindingAttr))) {
+			using (var h = new Mono.SafeGPtrArrayHandle (GetNestedTypes_native (namePtr.Value, bindingAttr, listType))) {
 				int n = h.Length;
 				var a = new RuntimeType [n];
 				for (int i = 0; i < n; i++) {
@@ -783,6 +785,10 @@ namespace System
 
 		public override string FullName {
 			get {
+				// https://bugzilla.xamarin.com/show_bug.cgi?id=57938
+				if (IsGenericType && ContainsGenericParameters && !IsGenericTypeDefinition)
+					return null;
+
 				string fullName;
 				// This doesn't need locking
 				if (type_info == null)
@@ -794,10 +800,33 @@ namespace System
 			}
 		}
 
+		public sealed override bool HasSameMetadataDefinitionAs (MemberInfo other) => HasSameMetadataDefinitionAsCore<RuntimeType> (other);	
+
+		public override bool IsSZArray {
+			get {
+				// TODO: intrinsic
+				return IsArray && ReferenceEquals (this, GetElementType ().MakeArrayType ());
+			}
+		}
+
 		internal override bool IsUserType {
 			get {
 				return false;
 			}
 		}
+
+		[System.Runtime.InteropServices.ComVisible(true)]
+		[Pure]
+		public override bool IsSubclassOf(Type type)
+		{
+			if ((object)type == null)
+				throw new ArgumentNullException("type");
+			Contract.EndContractBlock();
+			RuntimeType rtType = type as RuntimeType;
+			if (rtType == null)
+				return false;
+
+			return RuntimeTypeHandle.IsSubclassOf (this, rtType);
+		}
 	}
 }
