@@ -28,6 +28,7 @@
 // WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 //
 
+using System.Collections;
 using System.Collections.Generic;
 using System.Globalization;
 using System.Reflection;
@@ -56,12 +57,13 @@ namespace System.Diagnostics {
         }
 
 		public const int METHODS_TO_SKIP = 0;
+		const string prefix = "  at ";
 
 		private StackFrame[] frames;
 		readonly StackTrace[] captured_traces;
+#pragma warning disable 414		
 		private bool debug_info;
-
-		private static Dictionary<string, Func<StackTrace, string>> metadataHandlers;
+#pragma warning restore
 
 		[MethodImplAttribute (MethodImplOptions.NoInlining)]
 		public StackTrace ()
@@ -179,7 +181,17 @@ namespace System.Diagnostics {
 		[ComVisibleAttribute (false)]
 		public virtual StackFrame[] GetFrames ()
 		{
-			return frames;
+			if (captured_traces == null)
+				return frames;
+
+			var accum = new List<StackFrame> ();
+			foreach (var t in captured_traces) {
+				for (int i = 0; i < t.FrameCount; i++)
+					accum.Add (t.GetFrame (i));
+			}
+
+			accum.AddRange (frames);
+			return accum.ToArray ();
 		}
 
 		static bool isAotidSet;
@@ -196,34 +208,28 @@ namespace System.Diagnostics {
 			return aotid;
 		}
 
-		bool AddFrames (StringBuilder sb)
+		bool AddFrames (StringBuilder sb, bool separator, out bool isAsync)
 		{
-			bool printOffset;
-			string debugInfo, indentation;
-			string unknown = Locale.GetText ("<unknown method>");
+			isAsync = false;
+			bool any_frame = false;
 
-			indentation = "  ";
-			debugInfo = Locale.GetText (" in {0}:{1} ");
-
-			var newline = String.Format ("{0}{1}{2} ", Environment.NewLine, indentation,
-					Locale.GetText ("at"));
-
-			int i;
-			for (i = 0; i < FrameCount; i++) {
+			for (int i = 0; i < FrameCount; i++) {
 				StackFrame frame = GetFrame (i);
-				if (i == 0)
-					sb.AppendFormat ("{0}{1} ", indentation, Locale.GetText ("at"));
-				else
-					sb.Append (newline);
 
 				if (frame.GetMethod () == null) {
+					if (any_frame || separator)
+						sb.Append (Environment.NewLine);
+					sb.Append (prefix);
+
 					string internal_name = frame.GetInternalMethodName ();
 					if (internal_name != null)
 						sb.Append (internal_name);
 					else
-						sb.AppendFormat ("<0x{0:x5} + 0x{1:x5}> {2}", frame.GetMethodAddress (), frame.GetNativeOffset (), unknown);
+						sb.AppendFormat ("<0x{0:x5} + 0x{1:x5}> <unknown method>", frame.GetMethodAddress (), frame.GetNativeOffset ());
 				} else {
-					GetFullNameForStackTrace (sb, frame.GetMethod ());
+					GetFullNameForStackTrace (sb, frame.GetMethod (), any_frame || separator, out var skipped, out isAsync);
+					if (skipped)
+						continue;
 
 					if (frame.GetILOffset () == -1) {
 						sb.AppendFormat (" <0x{0:x5} + 0x{1:x5}>", frame.GetMethodAddress (), frame.GetNativeOffset ());
@@ -244,34 +250,53 @@ namespace System.Diagnostics {
 						}
 					}
 
-					sb.AppendFormat (debugInfo, filename, frame.GetFileLineNumber ());
+					sb.AppendFormat (" in {0}:{1} ", filename, frame.GetFileLineNumber ());
 				}
+
+				any_frame = true;
 			}
 
-			return i != 0;
+			return any_frame;
 		}
 
-		internal void GetFullNameForStackTrace (StringBuilder sb, MethodBase mi)
+		void GetFullNameForStackTrace (StringBuilder sb, MethodBase mi, bool needsNewLine, out bool skipped, out bool isAsync)
 		{
-			var declaringType = mi.DeclaringType;
-			if (declaringType.IsGenericType && !declaringType.IsGenericTypeDefinition)
-				declaringType = declaringType.GetGenericTypeDefinition ();
+			Type declaringType = mi.DeclaringType;
 
 			// Get generic definition
-			const BindingFlags bindingflags = BindingFlags.Instance | BindingFlags.Static | BindingFlags.Public | BindingFlags.NonPublic;
-			foreach (var m in declaringType.GetMethods (bindingflags)) {
-				if (m.MetadataToken == mi.MetadataToken) {
-					mi = m;
-					break;
+			if (declaringType.IsGenericType && !declaringType.IsGenericTypeDefinition) {
+				declaringType = declaringType.GetGenericTypeDefinition ();
+
+				const BindingFlags bindingflags = BindingFlags.Instance | BindingFlags.Static |
+					BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.DeclaredOnly;
+				foreach (var m in declaringType.GetMethods (bindingflags)) {
+					if (m.MetadataToken == mi.MetadataToken) {
+						mi = m;
+						break;
+					}
 				}
 			}
 
+			isAsync = typeof (IAsyncStateMachine).IsAssignableFrom (declaringType);
+			skipped = mi.IsDefined (typeof (StackTraceHiddenAttribute)) || declaringType.IsDefined (typeof (StackTraceHiddenAttribute));
+			if (skipped)
+				return;
+
+			if (isAsync) {
+				ConvertAsyncStateMachineMethod (ref mi, ref declaringType);
+			}
+
+			if (needsNewLine)
+				sb.Append (Environment.NewLine);
+			sb.Append (prefix);
+
 			sb.Append (declaringType.ToString ());
 
 			sb.Append (".");
 			sb.Append (mi.Name);
 
 			if (mi.IsGenericMethod) {
+				mi = ((MethodInfo)mi).GetGenericMethodDefinition ();
 				Type[] gen_params = mi.GetGenericArguments ();
 				sb.Append ("[");
 				for (int j = 0; j < gen_params.Length; j++) {
@@ -301,7 +326,32 @@ namespace System.Diagnostics {
 				}
 			}
 			sb.Append (")");
-		}		
+		}
+        
+		static void ConvertAsyncStateMachineMethod (ref MethodBase method, ref Type declaringType)
+		{
+			Type parentType = declaringType.DeclaringType;
+			if (parentType == null)
+				return;
+
+			MethodInfo[] methods = parentType.GetMethods (BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Static | BindingFlags.Instance | BindingFlags.DeclaredOnly);
+			if (methods == null)
+				return;
+
+			foreach (MethodInfo candidateMethod in methods) {
+				var attributes = candidateMethod.GetCustomAttributes<AsyncStateMachineAttribute> ();
+				if (attributes == null)
+					continue;
+
+				foreach (var attr in attributes) {
+					if (attr.StateMachineType == declaringType) {
+						method = candidateMethod;
+						declaringType = candidateMethod.DeclaringType;
+						return;
+					}
+				}
+			}
+		}
 
 		public override string ToString ()
 		{
@@ -310,18 +360,22 @@ namespace System.Diagnostics {
 			//
 			// Add traces captured using ExceptionDispatchInfo
 			//
+			bool has_frames = false;
 			if (captured_traces != null) {
 				foreach (var t in captured_traces) {
-					if (!t.AddFrames (sb))
+					has_frames = t.AddFrames (sb, has_frames, out var isAsync);
+					if (!has_frames)
 						continue;
 
-					sb.Append (Environment.NewLine);
-					sb.Append ("--- End of stack trace from previous location where exception was thrown ---");
-					sb.Append (Environment.NewLine);
+					if (!isAsync) {
+						sb.Append (Environment.NewLine);
+						sb.Append ("--- End of stack trace from previous location where exception was thrown ---");
+						sb.Append (Environment.NewLine);
+					}
 				}
 			}
 
-			AddFrames (sb);
+			AddFrames (sb, has_frames, out _);
 
 			return sb.ToString ();
 		}
