@@ -46,6 +46,11 @@ namespace System.Reflection.Emit {
 	public class CustomAttributeBuilder : _CustomAttributeBuilder {
 		ConstructorInfo ctor;
 		byte[] data;
+		object [] args;
+		PropertyInfo [] namedProperties;
+		object [] propertyValues;
+		FieldInfo [] namedFields;
+		object [] fieldValues;
 
 		internal ConstructorInfo Ctor {
 			get {return ctor;}
@@ -57,10 +62,27 @@ namespace System.Reflection.Emit {
 		
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		static extern byte[] GetBlob(Assembly asmb, ConstructorInfo con, object[] constructorArgs, PropertyInfo[] namedProperties, object[] propertyValues, FieldInfo[] namedFields, object[] fieldValues);
-		
-		internal CustomAttributeBuilder( ConstructorInfo con, byte[] cdata) {
+
+		internal object Invoke ()
+		{
+			object result = ctor.Invoke (args);
+
+			for (int i=0; i < namedFields.Length; i++)
+				namedFields [i].SetValue (result, fieldValues [i]);
+
+			for (int i=0; i < namedProperties.Length; i++)
+				namedProperties [i].SetValue (result, propertyValues [i]);
+
+			return result;
+		}
+
+		internal CustomAttributeBuilder( ConstructorInfo con, byte[] binaryAttribute) {
+			if (con == null)
+				throw new ArgumentNullException ("con");
+			if (binaryAttribute == null)
+				throw new ArgumentNullException ("binaryAttribute");
 			ctor = con;
-			data = (byte[])cdata.Clone ();
+			data = (byte[])binaryAttribute.Clone ();
 			/* should we check that the user supplied data is correct? */
 		}
 		
@@ -136,6 +158,12 @@ namespace System.Reflection.Emit {
 				FieldInfo [] namedFields, object [] fieldValues)
 		{
 			ctor = con;
+			args = constructorArgs;
+			this.namedProperties = namedProperties;
+			this.propertyValues = propertyValues;
+			this.namedFields = namedFields;
+			this.fieldValues = fieldValues;
+
 			if (con == null)
 				throw new ArgumentNullException ("con");
 			if (constructorArgs == null)
@@ -249,11 +277,24 @@ namespace System.Reflection.Emit {
 			return System.Text.Encoding.UTF8.GetString(data, pos, len);
 		}
 
+		internal static string decode_string (byte [] data, int pos, out int rpos)
+		{
+			if (data [pos] == 0xff) {
+				rpos = pos + 1;
+				return null;
+			} else {
+				int len = decode_len (data, pos, out pos);
+				string s = string_from_bytes (data, pos, len);
+				pos += len;
+				rpos = pos;
+				return s;
+			}
+		}
+
 		internal string string_arg ()
 		{
 			int pos = 2;
-			int len = decode_len (data, pos, out pos);
-			return string_from_bytes (data, pos, len);
+			return decode_string (data, pos, out pos);
 		}			
 
 		internal static UnmanagedMarshal get_umarshal (CustomAttributeBuilder customBuilder, bool is_field) {
@@ -281,18 +322,14 @@ namespace System.Reflection.Emit {
 				int paramType; // What is this ?
 				
 				/* Skip field/property signature */
-				pos ++;
+				int fieldPropSig = (int)data [pos ++];
 				/* Read type */
 				paramType = ((int)data [pos++]);
 				if (paramType == 0x55) {
 					/* enums, the value is preceeded by the type */
-					int len2 = decode_len (data, pos, out pos);
-					string_from_bytes (data, pos, len2);
-					pos += len2;
+					decode_string (data, pos, out pos);
 				}
-				int len = decode_len (data, pos, out pos);
-				string named_name = string_from_bytes (data, pos, len);
-				pos += len;
+				string named_name = decode_string (data, pos, out pos);
 
 				switch (named_name) {
 				case "ArraySubType":
@@ -321,9 +358,7 @@ namespace System.Reflection.Emit {
 					pos += 4;
 					break;
 				case "SafeArrayUserDefinedSubType":
-					len = decode_len (data, pos, out pos);
-					string_from_bytes (data, pos, len);
-					pos += len;
+					decode_string (data, pos, out pos);
 					break;
 				case "SizeParamIndex":
 					value = (int)data [pos++];
@@ -332,20 +367,15 @@ namespace System.Reflection.Emit {
 					hasSize = true;
 					break;
 				case "MarshalType":
-					len = decode_len (data, pos, out pos);
-					marshalTypeName = string_from_bytes (data, pos, len);
-					pos += len;
+					marshalTypeName = decode_string (data, pos, out pos);
 					break;
 				case "MarshalTypeRef":
-					len = decode_len (data, pos, out pos);
-					marshalTypeName = string_from_bytes (data, pos, len);
-					marshalTypeRef = Type.GetType (marshalTypeName);
-					pos += len;
+					marshalTypeName = decode_string (data, pos, out pos);
+					if (marshalTypeName != null)
+						marshalTypeRef = Type.GetType (marshalTypeName);
 					break;
 				case "MarshalCookie":
-					len = decode_len (data, pos, out pos);
-					marshalCookie = string_from_bytes (data, pos, len);
-					pos += len;
+					marshalCookie = decode_string (data, pos, out pos);
 					break;
 				default:
 					throw new Exception ("Unknown MarshalAsAttribute field: " + named_name);
@@ -358,8 +388,10 @@ namespace System.Reflection.Emit {
 					return UnmanagedMarshal.DefineLPArrayInternal (subtype, sizeConst, sizeParamIndex);
 				else
 					return UnmanagedMarshal.DefineLPArray (subtype);
+#if FEATURE_COMINTEROP
 			case UnmanagedType.SafeArray:
 				return UnmanagedMarshal.DefineSafeArray (subtype);
+#endif
 			case UnmanagedType.ByValArray:
 				if (!is_field)
 					throw new ArgumentException ("Specified unmanaged type is only valid on fields");
@@ -367,8 +399,10 @@ namespace System.Reflection.Emit {
 				return UnmanagedMarshal.DefineByValArray (sizeConst);
 			case UnmanagedType.ByValTStr:
 				return UnmanagedMarshal.DefineByValTStr (sizeConst);
+#if FEATURE_COMINTEROP
 			case UnmanagedType.CustomMarshaler:
 				return UnmanagedMarshal.DefineCustom (marshalTypeRef, marshalCookie, marshalTypeName, Guid.Empty);
+#endif
 			default:
 				return UnmanagedMarshal.DefineUnmanagedMarshal ((UnmanagedType)utype);
 			}
