@@ -103,7 +103,7 @@ namespace System.Reflection.Emit
 				System.Array.Copy (parameterTypes, this.parameters, parameterTypes.Length);
 			}
 			type = tb;
-			table_idx = get_next_table_index (this, 0x06, true);
+			table_idx = get_next_table_index (this, 0x06, 1);
 
 			((ModuleBuilder)tb.Module).RegisterToken (this, GetToken ().Token);
 		}
@@ -138,6 +138,12 @@ namespace System.Reflection.Emit
 			}
 		}
 
+		internal RuntimeMethodHandle MethodHandleInternal {
+			get {
+				return mhandle;
+			}
+		}
+
 		public override Type ReturnType {
 			get { return rtype; }
 		}
@@ -231,7 +237,7 @@ namespace System.Reflection.Emit
 
 			ParameterInfo[] retval = new ParameterInfo [parameters.Length];
 			for (int i = 0; i < parameters.Length; i++) {
-				retval [i] = ParameterInfo.New (pinfo == null ? null : pinfo [i + 1], parameters [i], this, i + 1);
+				retval [i] = MonoParameterInfo.New (pinfo?[i + 1], parameters [i], this, i + 1);
 			}
 			return retval;
 		}
@@ -248,6 +254,10 @@ namespace System.Reflection.Emit
 			return parameters [pos];
 		}
 
+		internal MethodBase RuntimeResolve () {
+			return type.RuntimeResolve ().GetMethod (this);
+		}
+
 		public Module GetModule ()
 		{
 			return type.Module;
@@ -332,7 +342,7 @@ namespace System.Reflection.Emit
 			//
 			// Extension: Mono allows position == 0 for the return attribute
 			//
-			if ((position < 0) || (position > parameters.Length))
+			if ((position < 0) || parameters == null || (position > parameters.Length))
 				throw new ArgumentOutOfRangeException ("position");
 
 			ParameterBuilder pb = new ParameterBuilder (this, position, attributes, strParamName);
@@ -380,6 +390,11 @@ namespace System.Reflection.Emit
 			}
 		}
 
+		internal void FixupTokens (Dictionary<int, int> token_map, Dictionary<int, MemberInfo> member_map) {
+			if (ilgen != null)
+				ilgen.FixupTokens (token_map, member_map);
+		}
+
 		internal void GenerateDebugInfo (ISymbolWriter symbolWriter)
 		{
 			if (ilgen != null && ilgen.HasDebugInfo) {
@@ -551,9 +566,9 @@ namespace System.Reflection.Emit
 			return name.GetHashCode ();
 		}
 
-		internal override int get_next_table_index (object obj, int table, bool inc)
+		internal override int get_next_table_index (object obj, int table, int count)
 		{
-			return type.get_next_table_index (obj, table, inc);
+			return type.get_next_table_index (obj, table, count);
 		}
 
 		void ExtendArray<T> (ref T[] array, T elem) {
