@@ -40,6 +40,7 @@ using System.Runtime.CompilerServices;
 using System.Runtime.InteropServices;
 using System.Globalization;
 using System.Collections;
+using System.Collections.Generic;
 using System.Security;
 using System.Security.Permissions;
 using System.Diagnostics.SymbolStore;
@@ -78,6 +79,7 @@ namespace System.Reflection.Emit
 		private GenericTypeParameterBuilder[] generic_params;
 		private RefEmitPermissionSet[] permissions;
 		private TypeInfo created;
+		private int state;
 		#endregion
 #pragma warning restore 169		
 		
@@ -91,15 +93,6 @@ namespace System.Reflection.Emit
 		{
 			return attrs;
 		}
-		
-		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		private extern void setup_internal_class ();
-		
-		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		private extern void create_generic_class ();
-
-		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		private extern EventInfo get_event_info (EventBuilder eb);
 
 		internal TypeBuilder (ModuleBuilder mb, TypeAttributes attr, int table_idx)
 		{
@@ -111,7 +104,6 @@ namespace System.Reflection.Emit
 			this.nspace = String.Empty;
 			this.fullname = TypeIdentifiers.WithoutEscape(this.tname);
 			pmodule = mb;
-			setup_internal_class ();
 		}
 
 		internal TypeBuilder (ModuleBuilder mb, string name, TypeAttributes attr, Type parent, Type[] interfaces, PackingSize packing_size, int type_size, Type nesting_type)
@@ -146,8 +138,7 @@ namespace System.Reflection.Emit
 				this.parent = typeof (object);
 
 			// skip .<Module> ?
-			table_idx = mb.get_next_table_index (this, 0x02, true);
-			setup_internal_class ();
+			table_idx = mb.get_next_table_index (this, 0x02, 1);
 			fullname = GetFullName ();
 		}
 
@@ -764,8 +755,6 @@ namespace System.Reflection.Emit
 				SetParent (pmodule.assemblyb.corlib_object_type);
 			}
 
-			create_generic_class ();
-
 			// Fire TypeResolve events for fields whose type is an unfinished
 			// value type.
 			if (fields != null) {
@@ -873,6 +862,21 @@ namespace System.Reflection.Emit
 			}
 		}
 
+		internal void FixupTokens (Dictionary<int, int> token_map, Dictionary<int, MemberInfo> member_map) {
+			if (methods != null) {
+				for (int i = 0; i < num_methods; ++i)
+					methods[i].FixupTokens (token_map, member_map);
+			}
+			if (ctors != null) {
+				foreach (var cb in ctors)
+					cb.FixupTokens (token_map, member_map);
+			}
+			if (subtypes != null) {
+				foreach (var tb in subtypes)
+					tb.FixupTokens (token_map, member_map);
+			}
+		}
+
 		internal void GenerateDebugInfo (ISymbolWriter symbolWriter)
 		{
 			symbolWriter.OpenNamespace (this.Namespace);
@@ -968,53 +972,6 @@ namespace System.Reflection.Emit
 			throw new NotSupportedException ();
 		}
 
-		// This is only used from MonoGenericInst.initialize().
-		internal EventInfo[] GetEvents_internal (BindingFlags bindingAttr)
-		{
-			if (events == null)
-				return new EventInfo [0];
-			ArrayList l = new ArrayList ();
-			bool match;
-			MethodAttributes mattrs;
-			MethodInfo accessor;
-
-			foreach (EventBuilder eb in events) {
-				if (eb == null)
-					continue;
-				EventInfo c = get_event_info (eb);
-				match = false;
-				accessor = c.GetAddMethod (true);
-				if (accessor == null)
-					accessor = c.GetRemoveMethod (true);
-				if (accessor == null)
-					continue;
-				mattrs = accessor.Attributes;
-				if ((mattrs & MethodAttributes.MemberAccessMask) == MethodAttributes.Public) {
-					if ((bindingAttr & BindingFlags.Public) != 0)
-						match = true;
-				} else {
-					if ((bindingAttr & BindingFlags.NonPublic) != 0)
-						match = true;
-				}
-				if (!match)
-					continue;
-				match = false;
-				if ((mattrs & MethodAttributes.Static) != 0) {
-					if ((bindingAttr & BindingFlags.Static) != 0)
-						match = true;
-				} else {
-					if ((bindingAttr & BindingFlags.Instance) != 0)
-						match = true;
-				}
-				if (!match)
-					continue;
-				l.Add (c);
-			}
-			EventInfo[] result = new EventInfo [l.Count];
-			l.CopyTo (result);
-			return result;
-		}
-
 		public override FieldInfo GetField (string name, BindingFlags bindingAttr)
 		{
 			if (created != null)
@@ -1642,13 +1599,10 @@ namespace System.Reflection.Emit
 				this.parent = parent;
 			}
 			this.parent = ResolveUserType (this.parent);
-
-			// will just set the parent-related bits if called a second time
-			setup_internal_class ();
 		}
 
-		internal int get_next_table_index (object obj, int table, bool inc) {
-			return pmodule.get_next_table_index (obj, table, inc);
+		internal int get_next_table_index (object obj, int table, int count) {
+			return pmodule.get_next_table_index (obj, table, count);
 		}
 
 		[ComVisible (true)]
@@ -1666,6 +1620,12 @@ namespace System.Reflection.Emit
 			return created;
 		}
 
+		internal override Type RuntimeResolve ()
+		{
+			check_created ();
+			return created;
+		}
+
 		internal bool is_created {
 			get {
 				return createTypeCalled;
@@ -1764,9 +1724,10 @@ namespace System.Reflection.Emit
 			}
 		}
 
-		public extern override bool IsGenericParameter {
-			[MethodImplAttribute(MethodImplOptions.InternalCall)]
-			get;
+		public override bool IsGenericParameter {
+			get {
+				return false;
+			}
 		}
 
 		public override GenericParameterAttributes GenericParameterAttributes {
@@ -1843,7 +1804,7 @@ namespace System.Reflection.Emit
 
 		static bool IsValidGetMethodType (Type type)
 		{
-			if (type is TypeBuilder || type is MonoGenericClass)
+			if (type is TypeBuilder || type is TypeBuilderInstantiation)
 				return true;
 			/*GetMethod() must work with TypeBuilders after CreateType() was called.*/
 			if (type.Module is ModuleBuilder)
@@ -1942,6 +1903,125 @@ namespace System.Reflection.Emit
 		{
 			return base.IsAssignableFrom (typeInfo);
 		}
+
+		internal static bool SetConstantValue (Type destType, Object value, ref Object destValue)
+		{
+			// Mono: This is based on the CoreCLR
+			// TypeBuilder.SetConstantValue except it writes to an
+			// out argument instead of doing an icall, and it uses
+			// TypeCode instead of CorElementType (like
+			// MonoTypeEnum) which we don't have in our corlib and
+			// our member fields are different.
+
+			// This is a helper function that is used by ParameterBuilder, PropertyBuilder,
+			// and FieldBuilder to validate a default value and save it in the meta-data.
+
+			if (value != null) {
+				Type type = value.GetType ();
+
+				// We should allow setting a constant value on a ByRef parameter
+				if (destType.IsByRef)
+					destType = destType.GetElementType ();
+
+				// Convert nullable types to their underlying type.
+				// This is necessary for nullable enum types to pass the IsEnum check that's coming next.
+				destType = Nullable.GetUnderlyingType (destType) ?? destType;
+
+				if (destType.IsEnum)
+				{
+					//                                   |  UnderlyingSystemType     |  Enum.GetUnderlyingType() |  IsEnum
+					// ----------------------------------|---------------------------|---------------------------|---------
+					// runtime Enum Type                 |  self                     |  underlying type of enum  |  TRUE
+					// EnumBuilder                       |  underlying type of enum  |  underlying type of enum* |  TRUE
+					// TypeBuilder of enum types**       |  underlying type of enum  |  Exception                |  TRUE
+					// TypeBuilder of enum types (baked) |  runtime enum type        |  Exception                |  TRUE
+
+					//  *: the behavior of Enum.GetUnderlyingType(EnumBuilder) might change in the future
+					//     so let's not depend on it.
+					// **: created with System.Enum as the parent type.
+
+					// The above behaviors might not be the most consistent but we have to live with them.
+
+					Type underlyingType;
+					EnumBuilder enumBldr;
+					TypeBuilder typeBldr;
+					if ((enumBldr = destType as EnumBuilder) != null) {
+						underlyingType = enumBldr.GetEnumUnderlyingType ();
+
+						// The constant value supplied should match either the baked enum type or its underlying type
+						// we don't need to compare it with the EnumBuilder itself because you can never have an object of that type
+						if (!((enumBldr.GetTypeBuilder ().is_created && type == enumBldr.GetTypeBuilder ().created) ||
+						      type == underlyingType))
+							throw_argument_ConstantDoesntMatch ();
+					} else if ((typeBldr = destType as TypeBuilder) != null) {
+						underlyingType = typeBldr.underlying_type;
+
+						// The constant value supplied should match either the baked enum type or its underlying type
+						// typeBldr.m_enumUnderlyingType is null if the user hasn't created a "value__" field on the enum
+						if (underlyingType == null || (type != typeBldr.UnderlyingSystemType && type != underlyingType))
+							throw_argument_ConstantDoesntMatch ();
+					} else {
+						// must be a runtime Enum Type
+
+						// Debug.Assert(destType is RuntimeType, "destType is not a runtime type, an EnumBuilder, or a TypeBuilder.");
+
+						underlyingType = Enum.GetUnderlyingType (destType);
+
+						// The constant value supplied should match either the enum itself or its underlying type
+						if (type != destType && type != underlyingType)
+							throw_argument_ConstantDoesntMatch ();
+					}
+
+					type = underlyingType;
+				} else {
+					// Note that it is non CLS compliant if destType != type. But RefEmit never guarantees CLS-Compliance.
+					if (!destType.IsAssignableFrom (type))
+						throw_argument_ConstantDoesntMatch ();
+				}
+
+				TypeCode corType = Type.GetTypeCode (type);
+
+				switch (corType)
+				{
+					case TypeCode.Byte:
+					case TypeCode.SByte:
+					case TypeCode.Boolean:
+					case TypeCode.Int16:
+					case TypeCode.UInt16:
+					case TypeCode.Char:
+					case TypeCode.Int32:
+					case TypeCode.UInt32:
+					case TypeCode.Single:
+					case TypeCode.Int64:
+					case TypeCode.UInt64:
+					case TypeCode.Double:
+						destValue = value;
+						return true;
+					case TypeCode.String:
+						destValue = value;
+						return true;
+					case TypeCode.DateTime:
+						//date is a I8 representation
+						long ticks = ((DateTime)value).Ticks;
+						destValue = ticks;
+						return true;
+					default:
+						throw new ArgumentException(type.ToString() + " is not a supported constant type.");
+				}
+			} else {
+				// A null default value in metadata is permissible even for non-nullable value types.
+				// (See ECMA-335 II.15.4.1.4 "The .param directive" and II.22.9 "Constant" for details.)
+				// This is how the Roslyn compilers generally encode `default(TValueType)` default values.
+
+				destValue = null;
+				return true;
+			}
+		}		
+
+		private static void throw_argument_ConstantDoesntMatch ()
+		{
+			throw new ArgumentException("Constant does not match the defined type.");
+		}
 	}
 }
 #endif
