@@ -37,8 +37,8 @@ using System.Reflection;
 using System.Runtime.CompilerServices;
 using System.Runtime.InteropServices;
 using System.Security.Permissions;
+using System.Security.Cryptography;
 using System.Security.Cryptography.X509Certificates;
-using Mono.Security.Authenticode;
 
 namespace System.Security.Policy {
 
@@ -150,6 +150,12 @@ namespace System.Security.Policy {
 				assemblyEvidenceList.Clear ();
 		}
 
+        [ComVisible(false)]
+        public Evidence Clone ()
+        {
+            return new Evidence(this);
+        }		
+
 		[Obsolete]
 		public void CopyTo (Array array, int index) 
 		{
@@ -219,7 +225,7 @@ namespace System.Security.Policy {
 		// "possible" presence of an Authenticode signature
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		static extern bool IsAuthenticodePresent (Assembly a);
-#if NET_2_1
+#if MOBILE
 		static internal Evidence GetDefaultHostEvidence (Assembly a)
 		{
 			return new Evidence ();
@@ -245,7 +251,7 @@ namespace System.Security.Policy {
 			}
 
 			// strongnamed assemblies gets a StrongName evidence
-			AssemblyName an = a.UnprotectedGetName ();
+			AssemblyName an = a.GetName ();
 			byte[] pk = an.GetPublicKey ();
 			if ((pk != null) && (pk.Length > 0)) {
 				StrongNamePublicKeyBlob blob = new StrongNamePublicKeyBlob (pk);
@@ -254,16 +260,14 @@ namespace System.Security.Policy {
 
 			// Authenticode(r) signed assemblies get a Publisher evidence
 			if (IsAuthenticodePresent (a)) {
-				// Note: The certificate is part of the evidences even if it is not trusted!
-				// so we can't call X509Certificate.CreateFromSignedFile
-				AuthenticodeDeformatter ad = new AuthenticodeDeformatter (a.Location);
-				if (ad.SigningCertificate != null) {
-					X509Certificate x509 = new X509Certificate (ad.SigningCertificate.RawData);
-					if (x509.GetHashCode () != 0) {
-						e.AddHost (new Publisher (x509));
-					}
+				try {
+					X509Certificate x509 = X509Certificate.CreateFromSignedFile (a.Location);
+					e.AddHost (new Publisher (x509));
+				}
+				catch (CryptographicException) {
 				}
 			}
+
 			// assemblies loaded from the GAC also get a Gac evidence (new in Fx 2.0)
 			if (a.GlobalAssemblyCache) {
 				e.AddHost (new GacInstalled ());
@@ -281,7 +285,7 @@ namespace System.Security.Policy {
 			return e;
 		}
 
-#endif // NET_2_1
+#endif // MOBILE
 
 		private class EvidenceEnumerator : IEnumerator {
 			
