@@ -49,10 +49,9 @@ namespace System.Threading {
 		#region Sync with metadata/object-internals.h
 		int lock_thread_id;
 		// stores a thread handle
-		internal IntPtr system_thread_handle;
-
-		/* Note this is an opaque object (an array), not a CultureInfo */
-		private object cached_culture_info;
+		IntPtr handle;
+		IntPtr native_handle; // used only on Win32
+		IntPtr unused3;
 		/* accessed only from unmanaged code */
 		private IntPtr name;
 		private int name_len; 
@@ -61,11 +60,7 @@ namespace System.Threading {
 		private int abort_state_handle;
 		/* thread_id is only accessed from unmanaged code */
 		internal Int64 thread_id;
-		
-		/* start_notify is used by the runtime to signal that Start()
-		 * is ok to return
-		 */
-		private IntPtr stack_ptr;
+		private IntPtr debugger_thread; // FIXME switch to bool as soon as CI testing with corlib version bump works
 		private UIntPtr static_data; /* GC-tracked */
 		private IntPtr runtime_thread_info;
 		/* current System.Runtime.Remoting.Contexts.Context instance
@@ -76,7 +71,7 @@ namespace System.Threading {
 		internal int _serialized_principal_version;
 		private IntPtr appdomain_refs;
 		private int interruption_requested;
-		private IntPtr synch_cs;
+		private IntPtr longlived;
 		internal bool threadpool_thread;
 		private bool thread_interrupt_requested;
 		/* These are used from managed code */
@@ -86,10 +81,14 @@ namespace System.Threading {
 		internal int managed_id;
 		private int small_id;
 		private IntPtr manage_callback;
-		private IntPtr interrupt_on_stop;
+		private IntPtr unused4;
 		private IntPtr flags;
 		private IntPtr thread_pinning_ref;
 		private IntPtr abort_protected_block_count;
+		private int priority = (int) ThreadPriority.Normal;
+		private IntPtr owned_mutex;
+		private IntPtr suspended_event;
+		private int self_suspended;
 		/* 
 		 * These fields are used to avoid having to increment corlib versions
 		 * when a new field is added to the unmanaged MonoThread structure.
@@ -107,11 +106,11 @@ namespace System.Threading {
 
 		// Closes the system thread handle
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		private extern void Thread_free_internal(IntPtr handle);
+		private extern void Thread_free_internal();
 
 		[ReliabilityContract (Consistency.WillNotCorruptState, Cer.Success)]
 		~InternalThread() {
-			Thread_free_internal(system_thread_handle);
+			Thread_free_internal();
 		}
 	}
 
@@ -122,7 +121,6 @@ namespace System.Threading {
 		private InternalThread internal_thread;
 		object m_ThreadStartArg;
 		object pending_exception;
-		int priority = (int) ThreadPriority.Normal;
 		#endregion
 #pragma warning restore 414
 
@@ -247,6 +245,10 @@ namespace System.Threading {
 			get {
 				Thread th = CurrentThread;
 
+				var logicalPrincipal = th.GetExecutionContextReader().LogicalCallContext.Principal;
+				if (logicalPrincipal != null)
+					return logicalPrincipal;
+
 				if (th.principal_version != th.Internal._serialized_principal_version)
 					th.principal = null;
 
@@ -268,6 +270,8 @@ namespace System.Threading {
 			set {
 				Thread th = CurrentThread;
 
+				th.GetMutableExecutionContext().LogicalCallContext.Principal = value;
+
 				if (value != GetDomain ().DefaultPrincipal) {
 					++th.Internal._serialized_principal_version;
 					try {
@@ -284,17 +288,17 @@ namespace System.Threading {
 			}
 		}
 
-		// Looks up the object associated with the current thread
-		// this is called by the JIT directly, too
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		private extern static InternalThread CurrentInternalThread_internal();
+		private extern static Thread GetCurrentThread ();
 
 		public static Thread CurrentThread {
 			[ReliabilityContract (Consistency.WillNotCorruptState, Cer.MayFail)]
 			get {
-				if (current_thread == null)
-					current_thread = new Thread (CurrentInternalThread_internal ());
-				return current_thread;
+				Thread current = current_thread;
+				if (current != null)
+					return current;
+				// This will set the current_thread tls variable
+				return GetCurrentThread ();
 			}
 		}
 
@@ -313,7 +317,7 @@ namespace System.Threading {
 
 		// Returns the system thread handle
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		private extern IntPtr Thread_internal (MulticastDelegate start);
+		private extern bool Thread_internal (MulticastDelegate start);
 
 		private Thread (InternalThread it) {
 			internal_thread = it;
@@ -328,13 +332,12 @@ namespace System.Threading {
 		[Obsolete ("Deprecated in favor of GetApartmentState, SetApartmentState and TrySetApartmentState.")]
 		public ApartmentState ApartmentState {
 			get {
-				if ((ThreadState & ThreadState.Stopped) != 0)
-					throw new ThreadStateException ("Thread is dead; state can not be accessed.");
-
+				ValidateThreadState ();
 				return (ApartmentState)Internal.apartment_state;
 			}
 
 			set {
+				ValidateThreadState ();
 				TrySetApartmentState (value);
 			}
 		}
@@ -370,14 +373,12 @@ namespace System.Threading {
 
 		public bool IsBackground {
 			get {
-				ThreadState thread_state = GetState (Internal);
-				if ((thread_state & ThreadState.Stopped) != 0)
-					throw new ThreadStateException ("Thread is dead; state can not be accessed.");
-
-				return (thread_state & ThreadState.Background) != 0;
+				var state = ValidateThreadState ();
+				return (state & ThreadState.Background) != 0;
 			}
 			
 			set {
+				ValidateThreadState ();
 				if (value) {
 					SetState (Internal, ThreadState.Background);
 				} else {
@@ -489,7 +490,7 @@ namespace System.Threading {
 #endif
 
 			// Thread_internal creates and starts the new thread, 
-			if (Thread_internal(m_Delegate) == IntPtr.Zero)
+			if (!Thread_internal(m_Delegate))
 				throw new SystemException ("Thread creation failed.");
 
 			m_ThreadStartArg = null;
@@ -651,6 +652,7 @@ namespace System.Threading {
 
 		public ApartmentState GetApartmentState ()
 		{
+			ValidateThreadState ();
 			return (ApartmentState)Internal.apartment_state;
 		}
 
@@ -714,5 +716,13 @@ namespace System.Threading {
 		{
 			throw new PlatformNotSupportedException ();
 		}
+
+		ThreadState ValidateThreadState ()
+		{
+			var state = GetState (Internal);
+			if ((state & ThreadState.Stopped) != 0)
+				throw new ThreadStateException ("Thread is dead; state can not be accessed.");
+			return state;
+		}
 	}
 }
