@@ -116,7 +116,7 @@ namespace System {
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private extern AppDomainSetup getSetup ();
 
-#if NET_2_1
+#if MOBILE
 		internal
 #endif
 		AppDomainSetup SetupInformationNoCopy {
@@ -130,7 +130,7 @@ namespace System {
 			}
 		}
 
-#if !NET_2_1
+#if !MOBILE
 		[MonoTODO]
 		public ApplicationTrust ApplicationTrust {
 			get { throw new NotImplementedException (); }
@@ -139,7 +139,7 @@ namespace System {
 		public string BaseDirectory {
 			get {
 				string path = SetupInformationNoCopy.ApplicationBase;
-#if !NET_2_1
+#if !MOBILE
 				if (SecurityManager.SecurityEnabled && (path != null) && (path.Length > 0)) {
 					// we cannot divulge local file informations
 					new FileIOPermission (FileIOPermissionAccess.PathDiscovery, path).Demand ();
@@ -152,7 +152,7 @@ namespace System {
 		public string RelativeSearchPath {
 			get {
 				string path = SetupInformationNoCopy.PrivateBinPath;
-#if !NET_2_1
+#if !MOBILE
 				if (SecurityManager.SecurityEnabled && (path != null) && (path.Length > 0)) {
 					// we cannot divulge local file informations
 					new FileIOPermission (FileIOPermissionAccess.PathDiscovery, path).Demand ();
@@ -169,7 +169,7 @@ namespace System {
 					return null;
 
 				string path = Path.Combine (setup.DynamicBase, setup.ApplicationName);
-#if !NET_2_1
+#if !MOBILE
 				if (SecurityManager.SecurityEnabled && (path != null) && (path.Length > 0)) {
 					// we cannot divulge local file informations
 					new FileIOPermission (FileIOPermissionAccess.PathDiscovery, path).Demand ();
@@ -311,7 +311,7 @@ namespace System {
 			SetupInformationNoCopy.ShadowCopyDirectories = String.Empty;
 		}
 
-#if !NET_2_1
+#if !MOBILE
 		public ObjectHandle CreateComInstanceFrom (string assemblyName, string typeName)
 		{
 			return Activator.CreateComInstanceFrom (assemblyName, typeName);
@@ -324,6 +324,46 @@ namespace System {
 		}
 #endif
 
+		internal ObjectHandle InternalCreateInstanceWithNoSecurity (string assemblyName, string typeName)
+		{
+			return CreateInstance(assemblyName, typeName);
+		}
+
+		internal ObjectHandle InternalCreateInstanceWithNoSecurity (string assemblyName, 
+                                                                    string typeName,
+                                                                    bool ignoreCase,
+                                                                    BindingFlags bindingAttr,
+                                                                    Binder binder,
+                                                                    Object[] args,
+                                                                    CultureInfo culture,
+                                                                    Object[] activationAttributes,
+                                                                    Evidence securityAttributes)
+		{
+#pragma warning disable 618
+		return CreateInstance(assemblyName, typeName, ignoreCase, bindingAttr, binder, args, culture, activationAttributes, securityAttributes);
+#pragma warning restore 618
+		}
+
+		internal ObjectHandle InternalCreateInstanceFromWithNoSecurity (string assemblyName, string typeName)
+		{
+			return CreateInstanceFrom(assemblyName, typeName);
+		}
+
+		internal ObjectHandle InternalCreateInstanceFromWithNoSecurity (string assemblyName, 
+                                                                        string typeName,
+                                                                        bool ignoreCase,
+                                                                        BindingFlags bindingAttr,
+                                                                        Binder binder,
+                                                                        Object[] args,
+                                                                        CultureInfo culture,
+                                                                        Object[] activationAttributes,
+                                                                        Evidence securityAttributes)
+		{
+#pragma warning disable 618
+			return CreateInstanceFrom(assemblyName, typeName, ignoreCase, bindingAttr, binder, args, culture, activationAttributes, securityAttributes);
+#pragma warning restore 618
+		}
+
 		public ObjectHandle CreateInstance (string assemblyName, string typeName)
 		{
 			if (assemblyName == null)
@@ -943,6 +983,7 @@ namespace System {
 			return _process_guid;
 		}
 
+#if MONO_FEATURE_MULTIPLE_APPDOMAINS
 		public static AppDomain CreateDomain (string friendlyName)
 		{
 			return CreateDomain (friendlyName, null, null);
@@ -986,7 +1027,7 @@ namespace System {
 			} else if (info.ConfigurationFile == null)
 				info.ConfigurationFile = "[I don't have a config file]";
 
-#if !NET_2_1
+#if !MOBILE
 			if (info.AppDomainInitializer != null) {
 				if (!info.AppDomainInitializer.Method.IsStatic)
 					throw new ArgumentException ("Non-static methods cannot be invoked as an appdomain initializer");
@@ -1006,7 +1047,7 @@ namespace System {
 			else
 				ad._evidence = new Evidence (securityInfo);	// copy
 
-#if !NET_2_1
+#if !MOBILE
 			if (info.AppDomainInitializer != null) {
 				Loader loader = new Loader (
 					info.AppDomainInitializer.Method.DeclaringType.Assembly.Location);
@@ -1021,8 +1062,27 @@ namespace System {
 
 			return ad;
 		}
+#else
+		[Obsolete ("AppDomain.CreateDomain is not supported on the current platform.", true)]
+		public static AppDomain CreateDomain (string friendlyName)
+		{
+			throw new PlatformNotSupportedException ("AppDomain.CreateDomain is not supported on the current platform.");
+		}
+		
+		[Obsolete ("AppDomain.CreateDomain is not supported on the current platform.", true)]
+		public static AppDomain CreateDomain (string friendlyName, Evidence securityInfo)
+		{
+			throw new PlatformNotSupportedException ("AppDomain.CreateDomain is not supported on the current platform.");
+		}
 
-#if !NET_2_1
+		[Obsolete ("AppDomain.CreateDomain is not supported on the current platform.", true)]
+		public static AppDomain CreateDomain (string friendlyName, Evidence securityInfo, AppDomainSetup info)
+		{
+			throw new PlatformNotSupportedException ("AppDomain.CreateDomain is not supported on the current platform.");
+		}
+#endif // MONO_FEATURE_MULTIPLE_APPDOMAINS
+
+#if !MOBILE
 		[Serializable]
 		class Loader {
 
@@ -1058,13 +1118,23 @@ namespace System {
 		}
 #endif
 
+#if MONO_FEATURE_MULTIPLE_APPDOMAINS
 		public static AppDomain CreateDomain (string friendlyName, Evidence securityInfo,string appBasePath,
 		                                      string appRelativeSearchPath, bool shadowCopyFiles)
 		{
 			return CreateDomain (friendlyName, securityInfo, CreateDomainSetup (appBasePath, appRelativeSearchPath, shadowCopyFiles));
 		}
+#else
+		[Obsolete ("AppDomain.CreateDomain is not supported on the current platform.", true)]
+		public static AppDomain CreateDomain (string friendlyName, Evidence securityInfo,string appBasePath,
+		                                      string appRelativeSearchPath, bool shadowCopyFiles)
+		{
+			throw new PlatformNotSupportedException ("AppDomain.CreateDomain is not supported on the current platform.");
+		}
+#endif // MONO_FEATURE_MULTIPLE_APPDOMAINS
 		
-#if !NET_2_1
+#if !MOBILE
+#if MONO_FEATURE_MULTIPLE_APPDOMAINS
 		public static AppDomain CreateDomain (string friendlyName, Evidence securityInfo, AppDomainSetup info,
 		                                      PermissionSet grantSet, params StrongName [] fullTrustAssemblies)
 		{
@@ -1074,8 +1144,17 @@ namespace System {
 			info.ApplicationTrust = new ApplicationTrust (grantSet, fullTrustAssemblies ?? EmptyArray<StrongName>.Value);
 			return CreateDomain (friendlyName, securityInfo, info);		
 		}
+#else
+		[Obsolete ("AppDomain.CreateDomain is not supported on the current platform.", true)]
+		public static AppDomain CreateDomain (string friendlyName, Evidence securityInfo, AppDomainSetup info,
+		                                      PermissionSet grantSet, params StrongName [] fullTrustAssemblies)
+		{
+			throw new PlatformNotSupportedException ("AppDomain.CreateDomain is not supported on the current platform.");
+		}
+#endif // MONO_FEATURE_MULTIPLE_APPDOMAINS
 #endif
 
+#if MONO_FEATURE_MULTIPLE_APPDOMAINS
 		static AppDomainSetup CreateDomainSetup (string appBasePath, string appRelativeSearchPath, bool shadowCopyFiles)
 		{
 			AppDomainSetup info = new AppDomainSetup ();
@@ -1090,6 +1169,7 @@ namespace System {
 
 			return info;
 		}
+#endif // MONO_FEATURE_MULTIPLE_APPDOMAINS
 		
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private static extern bool InternalIsFinalizingForUnload (int domain_id);
@@ -1109,6 +1189,7 @@ namespace System {
 			return Thread.GetDomainID ();
 		}
 
+#if MONO_FEATURE_MULTIPLE_APPDOMAINS
 		[SecurityPermission (SecurityAction.Demand, ControlAppDomain = true)]
 		[ReliabilityContractAttribute (Consistency.MayCorruptAppDomain, Cer.MayFail)]
 		public static void Unload (AppDomain domain)
@@ -1118,6 +1199,13 @@ namespace System {
 
 			InternalUnload (domain.getDomainID());
 		}
+#else
+		[Obsolete ("AppDomain.Unload is not supported on the current platform.", true)]
+		public static void Unload (AppDomain domain)
+		{
+			throw new PlatformNotSupportedException ("AppDomain.Unload is not supported on the current platform.");
+		}
+#endif // MONO_FEATURE_MULTIPLE_APPDOMAINS
 
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		[SecurityPermission (SecurityAction.LinkDemand, ControlAppDomain = true)]
@@ -1129,14 +1217,16 @@ namespace System {
 			SetData (name, data);
 		}
 
-#if !NET_2_1
 		[Obsolete ("Use AppDomainSetup.DynamicBase")]
 		[SecurityPermission (SecurityAction.LinkDemand, ControlAppDomain = true)]
 		public void SetDynamicBase (string path)
 		{
+#if MOBILE
+			throw new PlatformNotSupportedException ();
+#else
 			SetupInformationNoCopy.DynamicBase = path;
+#endif // MOBILE
 		}
-#endif // !NET_2_1
 
 		[Obsolete ("AppDomain.GetCurrentThreadId has been deprecated"
 			+ " because it does not provide a stable Id when managed"
@@ -1198,14 +1288,11 @@ namespace System {
 		private Assembly DoAssemblyResolve (string name, Assembly requestingAssembly, bool refonly)
 		{
 			ResolveEventHandler del;
-#if !NET_2_1
 			if (refonly)
 				del = ReflectionOnlyAssemblyResolve;
 			else
 				del = AssemblyResolve;
-#else
-			del = AssemblyResolve;
-#endif
+
 			if (del == null)
 				return null;
 			
@@ -1299,11 +1386,16 @@ namespace System {
 			return null;
 		}
 
+#if MONO_FEATURE_MULTIPLE_APPDOMAINS
 		private void DoDomainUnload ()
 		{
 			if (DomainUnload != null)
 				DomainUnload(this, null);
 		}
+#endif // MONO_FEATURE_MULTIPLE_APPDOMAINS
+
+		[MethodImplAttribute(MethodImplOptions.InternalCall)]
+		internal extern void DoUnhandledException (Exception e);
 
 		internal void DoUnhandledException (UnhandledExceptionEventArgs args) {
 			if (UnhandledException != null)
@@ -1346,7 +1438,11 @@ namespace System {
 		[method: SecurityPermission (SecurityAction.LinkDemand, ControlAppDomain = true)]
 		public event ResolveEventHandler AssemblyResolve;
 
+#if MONO_FEATURE_MULTIPLE_APPDOMAINS
 		[method: SecurityPermission (SecurityAction.LinkDemand, ControlAppDomain = true)]
+#else
+		[Obsolete ("AppDomain.DomainUnload is not supported on the current platform.", true)]
+#endif // MONO_FEATURE_MULTIPLE_APPDOMAINS
 		public event EventHandler DomainUnload;
 
 		[method: SecurityPermission (SecurityAction.LinkDemand, ControlAppDomain = true)]
@@ -1381,14 +1477,19 @@ namespace System {
 #endif
         #pragma warning restore 649
 
+#if MONO_FEATURE_MULTIPLE_APPDOMAINS
 		// default is null
 		public AppDomainManager DomainManager {
 			get { return (AppDomainManager)_domain_manager; }
 		}
+#else
+		[Obsolete ("AppDomain.DomainManager is not supported on this platform.", true)]
+		public AppDomainManager DomainManager {
+			get { throw new PlatformNotSupportedException ("AppDomain.DomainManager is not supported on this platform."); }
+		}
+#endif // MONO_FEATURE_MULTIPLE_APPDOMAINS
 
-#if !MOBILE
 		public event ResolveEventHandler ReflectionOnlyAssemblyResolve;
-#endif
 
         #pragma warning disable 649
 #if MOBILE
@@ -1430,6 +1531,7 @@ namespace System {
 
 		// static methods
 
+#if MONO_FEATURE_MULTIPLE_APPDOMAINS
 		public static AppDomain CreateDomain (string friendlyName, Evidence securityInfo, string appBasePath,
 			string appRelativeSearchPath, bool shadowCopyFiles, AppDomainInitializer adInit, string[] adInitArgs)
 		{
@@ -1440,6 +1542,14 @@ namespace System {
 
 			return CreateDomain (friendlyName, securityInfo, info);
 		}
+#else
+		[Obsolete ("AppDomain.CreateDomain is not supported on the current platform.", true)]
+		public static AppDomain CreateDomain (string friendlyName, Evidence securityInfo, string appBasePath,
+			string appRelativeSearchPath, bool shadowCopyFiles, AppDomainInitializer adInit, string[] adInitArgs)
+		{
+			throw new PlatformNotSupportedException ("AppDomain.CreateDomain is not supported on the current platform.");
+		}
+#endif // MONO_FEATURE_MULTIPLE_APPDOMAINS
 
 		public int ExecuteAssemblyByName (string assemblyName)
 		{
