@@ -14,6 +14,8 @@ using System;
 using System.IO;
 using System.Globalization;
 using System.Threading;
+using System.Runtime.InteropServices;
+
 
 using NUnit.Framework;
 
@@ -23,13 +25,16 @@ namespace MonoTests.System.IO
 	public class FileTest
 	{
 		CultureInfo old_culture;
-		static string TempFolder = Path.Combine (Path.GetTempPath (), "MonoTests.System.IO.Tests");
+		string tmpFolder;
 
 		[SetUp]
 		public void SetUp ()
 		{
-			DeleteDirectory (TempFolder);
-			Directory.CreateDirectory (TempFolder);
+			tmpFolder = Path.GetTempFileName ();
+			if (File.Exists (tmpFolder))
+				File.Delete (tmpFolder);
+			DeleteDirectory (tmpFolder);
+			Directory.CreateDirectory (tmpFolder);
 			old_culture = Thread.CurrentThread.CurrentCulture;
 			Thread.CurrentThread.CurrentCulture = new CultureInfo ("en-US", false);
 		}
@@ -37,7 +42,7 @@ namespace MonoTests.System.IO
 		[TearDown]
 		public void TearDown ()
 		{
-			DeleteDirectory (TempFolder);
+			DeleteDirectory (tmpFolder);
 			Thread.CurrentThread.CurrentCulture = old_culture;
 		}
 
@@ -63,7 +68,7 @@ namespace MonoTests.System.IO
 		public void TestExists ()
 		{
 			FileStream s = null;
-			string path = TempFolder + Path.DirectorySeparatorChar + "AFile.txt";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "AFile.txt";
 			try {
 				Assert.IsFalse (File.Exists (null), "#1");
 				Assert.IsFalse (File.Exists (string.Empty), "#2");
@@ -72,7 +77,7 @@ namespace MonoTests.System.IO
 				s = File.Create (path);
 				s.Close ();
 				Assert.IsTrue (File.Exists (path), "#4");
-				Assert.IsFalse (File.Exists (TempFolder + Path.DirectorySeparatorChar + "doesnotexist"), "#5");
+				Assert.IsFalse (File.Exists (tmpFolder + Path.DirectorySeparatorChar + "doesnotexist"), "#5");
 			} finally {
 				if (s != null)
 					s.Close ();
@@ -110,7 +115,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void Create_Path_Directory ()
 		{
-			string path = Path.Combine (TempFolder, "foo");
+			string path = Path.Combine (tmpFolder, "foo");
 			Directory.CreateDirectory (path);
 			try {
 				File.Create (path);
@@ -141,10 +146,27 @@ namespace MonoTests.System.IO
 			}
 		}
 
+		internal const string LIBC = "libc";
+		// geteuid(2)
+		//    uid_t geteuid(void);
+		[DllImport (LIBC, SetLastError=true)]
+		static extern uint geteuid ();
+
+		static bool RunningAsRoot // FIXME?
+		{
+			get {
+				//return RunningOnUnix && System.Security.WindowsIdentity.GetCurrentToken () == IntPtr.Zero;
+				return RunningOnUnix && geteuid () == 0;
+			}
+		}
+
 		[Test]
 		public void Create_Path_ReadOnly ()
 		{
-			string path = Path.Combine (TempFolder, "foo");
+			if (RunningAsRoot) // FIXME?
+				Assert.Ignore ("Setting readonly in Mono does not block root writes.");
+
+			string path = Path.Combine (tmpFolder, "foo");
 			File.Create (path).Close ();
 			File.SetAttributes (path, FileAttributes.ReadOnly);
 			try {
@@ -164,6 +186,8 @@ namespace MonoTests.System.IO
 		[Test]
 		public void Create_Path_Whitespace ()
 		{
+			// FIXME? This is valid on Unix and can work on Windows.
+			// This test and Mono have in mind an incorrect low common denominator.
 			try {
 				File.Create (" ");
 				Assert.Fail ("#1");
@@ -180,7 +204,7 @@ namespace MonoTests.System.IO
 		public void Create_Directory_DoesNotExist ()
 		{
 			FileStream stream = null;
-			string path = TempFolder + Path.DirectorySeparatorChar + "directory_does_not_exist" + Path.DirectorySeparatorChar + "foo";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "directory_does_not_exist" + Path.DirectorySeparatorChar + "foo";
 			
 			try {
 				stream = File.Create (path);
@@ -205,7 +229,7 @@ namespace MonoTests.System.IO
 			string path = null;
 
 			/* positive test: create resources/foo */
-			path = TempFolder + Path.DirectorySeparatorChar + "foo";
+			path = tmpFolder + Path.DirectorySeparatorChar + "foo";
 			try {
 
 				stream = File.Create (path);
@@ -220,7 +244,7 @@ namespace MonoTests.System.IO
 			stream = null;
 
 			/* positive test: repeat test above again to test for overwriting file */
-			path = TempFolder + Path.DirectorySeparatorChar + "foo";
+			path = tmpFolder + Path.DirectorySeparatorChar + "foo";
 			try {
 				stream = File.Create (path);
 				Assert.IsTrue (File.Exists (path), "#2");
@@ -337,8 +361,8 @@ namespace MonoTests.System.IO
 		[Test]
 		public void Copy_DestFileName_AlreadyExists ()
 		{
-			string source = TempFolder + Path.DirectorySeparatorChar + "AFile.txt";
-			string dest = TempFolder + Path.DirectorySeparatorChar + "bar";
+			string source = tmpFolder + Path.DirectorySeparatorChar + "AFile.txt";
+			string dest = tmpFolder + Path.DirectorySeparatorChar + "bar";
 			DeleteFile (source);
 			DeleteFile (dest);
 			try {
@@ -363,7 +387,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void Copy_SourceFileName_DestFileName_Same ()
 		{
-			string source = TempFolder + Path.DirectorySeparatorChar + "SameFile.txt";
+			string source = tmpFolder + Path.DirectorySeparatorChar + "SameFile.txt";
 			DeleteFile (source);
 			try {
 				// new empty file
@@ -384,8 +408,8 @@ namespace MonoTests.System.IO
 		[Test]
 		public void Copy ()
 		{
-			string path1 = TempFolder + Path.DirectorySeparatorChar + "bar";
-			string path2 = TempFolder + Path.DirectorySeparatorChar + "AFile.txt";
+			string path1 = tmpFolder + Path.DirectorySeparatorChar + "bar";
+			string path2 = tmpFolder + Path.DirectorySeparatorChar + "AFile.txt";
 			/* positive test: copy resources/AFile.txt to resources/bar */
 			try {
 				DeleteFile (path1);
@@ -453,7 +477,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void Delete_Directory_DoesNotExist ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "directory_does_not_exist" + Path.DirectorySeparatorChar + "foo";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "directory_does_not_exist" + Path.DirectorySeparatorChar + "foo";
 			if (Directory.Exists (path))
 				Directory.Delete (path, true);
 
@@ -472,7 +496,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void Delete ()
 		{
-			string foopath = TempFolder + Path.DirectorySeparatorChar + "foo";
+			string foopath = tmpFolder + Path.DirectorySeparatorChar + "foo";
 			DeleteFile (foopath);
 			try {
 				File.Create (foopath).Close ();
@@ -488,7 +512,7 @@ namespace MonoTests.System.IO
 		[Category ("NotWorking")]
 		public void Delete_FileLock ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "DeleteOpenStreamException";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "DeleteOpenStreamException";
 			DeleteFile (path);
 			FileStream stream = null;
 			try {
@@ -518,7 +542,7 @@ namespace MonoTests.System.IO
 			if (RunningOnUnix)
 				Assert.Ignore ("ReadOnly files can be deleted on unix since fdef50957f508627928c7876a905d5584da45748.");
 
-			string path = TempFolder + Path.DirectorySeparatorChar + "DeleteReadOnly";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "DeleteReadOnly";
 			DeleteFile (path);
 			try {
 				File.Create (path).Close ();
@@ -530,6 +554,12 @@ namespace MonoTests.System.IO
 			}
 		}
 
+		[Test]
+		public void Delete_NonExisting_NoException ()
+		{
+			File.Delete (Path.Combine (Directory.GetDirectoryRoot (Directory.GetCurrentDirectory ()), "monononexistingfile.dat"));
+		}
+
 		[Test]
 		public void GetAttributes_Archive ()
 		{
@@ -538,7 +568,7 @@ namespace MonoTests.System.IO
 
 			FileAttributes attrs;
 
-			string path = Path.Combine (TempFolder, "GetAttributes.tmp");
+			string path = Path.Combine (tmpFolder, "GetAttributes.tmp");
 			File.Create (path).Close ();
 
 			attrs = File.GetAttributes (path);
@@ -557,7 +587,7 @@ namespace MonoTests.System.IO
 			if (RunningOnUnix)
 				Assert.Ignore ("bug #325181: FileAttributes.Archive has no effect on Unix.");
 
-			string path = Path.Combine (TempFolder, "GetAttributes.tmp");
+			string path = Path.Combine (tmpFolder, "GetAttributes.tmp");
 			File.Create (path).Close ();
 
 			FileAttributes attrs = File.GetAttributes (path);
@@ -573,7 +603,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void GetAttributes_Default_Directory ()
 		{
-			FileAttributes attrs = File.GetAttributes (TempFolder);
+			FileAttributes attrs = File.GetAttributes (tmpFolder);
 
 			Assert.IsFalse ((attrs & FileAttributes.Archive) != 0, "#1");
 			Assert.IsTrue ((attrs & FileAttributes.Directory) != 0, "#2");
@@ -586,16 +616,16 @@ namespace MonoTests.System.IO
 		[Test]
 		public void GetAttributes_Directory ()
 		{
-			FileAttributes attrs = File.GetAttributes (TempFolder);
+			FileAttributes attrs = File.GetAttributes (tmpFolder);
 
 			Assert.IsTrue ((attrs & FileAttributes.Directory) != 0, "#1");
 
 			attrs &= ~FileAttributes.Directory;
-			File.SetAttributes (TempFolder, attrs);
+			File.SetAttributes (tmpFolder, attrs);
 
 			Assert.IsFalse ((attrs & FileAttributes.Directory) != 0, "#2");
 
-			string path = Path.Combine (TempFolder, "GetAttributes.tmp");
+			string path = Path.Combine (tmpFolder, "GetAttributes.tmp");
 			File.Create (path).Close ();
 
 			attrs = File.GetAttributes (path);
@@ -608,9 +638,12 @@ namespace MonoTests.System.IO
 		[Test]
 		public void GetAttributes_ReadOnly ()
 		{
+			if (RunningAsRoot) // FIXME?
+				Assert.Ignore ("Setting readonly in Mono does not block root writes.");
+
 			FileAttributes attrs;
 
-			string path = Path.Combine (TempFolder, "GetAttributes.tmp");
+			string path = Path.Combine (tmpFolder, "GetAttributes.tmp");
 			File.Create (path).Close ();
 
 			attrs = File.GetAttributes (path);
@@ -635,7 +668,7 @@ namespace MonoTests.System.IO
 
 			FileAttributes attrs;
 
-			string path = Path.Combine (TempFolder, "GetAttributes.tmp");
+			string path = Path.Combine (tmpFolder, "GetAttributes.tmp");
 			File.Create (path).Close ();
 
 			attrs = File.GetAttributes (path);
@@ -651,7 +684,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void GetAttributes_Path_DoesNotExist ()
 		{
-			string path = Path.Combine (TempFolder, "GetAttributes.tmp");
+			string path = Path.Combine (tmpFolder, "GetAttributes.tmp");
 			try {
 				File.GetAttributes (path);
 				Assert.Fail ("#1");
@@ -783,7 +816,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void Move_SourceFileName_DoesNotExist ()
 		{
-			string file = TempFolder + Path.DirectorySeparatorChar + "doesnotexist";
+			string file = tmpFolder + Path.DirectorySeparatorChar + "doesnotexist";
 			DeleteFile (file);
 			try {
 				File.Move (file, "b");
@@ -799,8 +832,8 @@ namespace MonoTests.System.IO
 		[Test]
 		public void Move_DestFileName_DirectoryDoesNotExist ()
 		{
-			string sourceFile = TempFolder + Path.DirectorySeparatorChar + "foo";
-			string destFile = Path.Combine (Path.Combine (TempFolder, "doesnotexist"), "b");
+			string sourceFile = tmpFolder + Path.DirectorySeparatorChar + "foo";
+			string destFile = Path.Combine (Path.Combine (tmpFolder, "doesnotexist"), "b");
 			DeleteFile (sourceFile);
 			try {
 				File.Create (sourceFile).Close ();
@@ -822,13 +855,13 @@ namespace MonoTests.System.IO
 		[Test]
 		public void Move_DestFileName_AlreadyExists ()
 		{
-			string sourceFile = TempFolder + Path.DirectorySeparatorChar + "foo";
+			string sourceFile = tmpFolder + Path.DirectorySeparatorChar + "foo";
 			string destFile;
 
 			// move to same directory
 			File.Create (sourceFile).Close ();
 			try {
-				File.Move (sourceFile, TempFolder);
+				File.Move (sourceFile, tmpFolder);
 				Assert.Fail ("#A1");
 			} catch (IOException ex) {
 				// Cannot create a file when that file already exists
@@ -836,14 +869,14 @@ namespace MonoTests.System.IO
 				Assert.IsNull (ex.InnerException, "#A3");
 				Assert.IsNotNull (ex.Message, "#A4");
 				Assert.IsFalse (ex.Message.IndexOf (sourceFile) != -1, "#A5");
-				Assert.IsFalse (ex.Message.IndexOf (TempFolder) != -1, "#A6");
+				Assert.IsFalse (ex.Message.IndexOf (tmpFolder) != -1, "#A6");
 			} finally {
 				DeleteFile (sourceFile);
 			}
 
 			// move to exist file
 			File.Create (sourceFile).Close ();
-			destFile = TempFolder + Path.DirectorySeparatorChar + "bar";
+			destFile = tmpFolder + Path.DirectorySeparatorChar + "bar";
 			File.Create (destFile).Close ();
 			try {
 				File.Move (sourceFile, destFile);
@@ -862,7 +895,7 @@ namespace MonoTests.System.IO
 
 			// move to existing directory
 			File.Create (sourceFile).Close ();
-			destFile = TempFolder + Path.DirectorySeparatorChar + "bar";
+			destFile = tmpFolder + Path.DirectorySeparatorChar + "bar";
 			Directory.CreateDirectory (destFile);
 			try {
 				File.Move (sourceFile, destFile);
@@ -883,8 +916,8 @@ namespace MonoTests.System.IO
 		[Test]
 		public void Move ()
 		{
-			string bar = TempFolder + Path.DirectorySeparatorChar + "bar";
-			string baz = TempFolder + Path.DirectorySeparatorChar + "baz";
+			string bar = tmpFolder + Path.DirectorySeparatorChar + "bar";
+			string baz = tmpFolder + Path.DirectorySeparatorChar + "baz";
 			if (!File.Exists (bar)) {
 				FileStream f = File.Create(bar);
 				f.Close();
@@ -896,8 +929,8 @@ namespace MonoTests.System.IO
 			Assert.IsTrue (File.Exists (baz), "#3");
 
 			// Test moving of directories
-			string dir = Path.Combine (TempFolder, "dir");
-			string dir2 = Path.Combine (TempFolder, "dir2");
+			string dir = Path.Combine (tmpFolder, "dir");
+			string dir2 = Path.Combine (tmpFolder, "dir2");
 			string dir_foo = Path.Combine (dir, "foo");
 			string dir2_foo = Path.Combine (dir2, "foo");
 
@@ -957,7 +990,7 @@ namespace MonoTests.System.IO
 			string path = null;
 			FileStream stream = null;
 
-			path = TempFolder + Path.DirectorySeparatorChar + "AFile.txt";
+			path = tmpFolder + Path.DirectorySeparatorChar + "AFile.txt";
 			try {
 				if (!File.Exists (path))
 					stream = File.Create (path);
@@ -1001,7 +1034,7 @@ namespace MonoTests.System.IO
 			stream = null;
 
 			/* Exception tests */
-			path = TempFolder + Path.DirectorySeparatorChar + "filedoesnotexist";
+			path = tmpFolder + Path.DirectorySeparatorChar + "filedoesnotexist";
 			try {
 				stream = File.Open (path, FileMode.Open);
 				Assert.Fail ("#D1");
@@ -1020,10 +1053,10 @@ namespace MonoTests.System.IO
 		[Test]
 		public void Open_CreateNewMode_ReadAccess ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "AFile.txt";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "AFile.txt";
 			FileStream stream = null;
 			try {
-				stream = File.Open (TempFolder + Path.DirectorySeparatorChar + "AFile.txt", FileMode.CreateNew, FileAccess.Read);
+				stream = File.Open (tmpFolder + Path.DirectorySeparatorChar + "AFile.txt", FileMode.CreateNew, FileAccess.Read);
 				Assert.Fail ("#1");
 			} catch (ArgumentException ex) {
 				// Combining FileMode: CreateNew with FileAccess: Read is invalid
@@ -1041,7 +1074,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void Open_AppendMode_ReadAccess ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "AFile.txt";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "AFile.txt";
 			FileStream s = null;
 			if (!File.Exists (path))
 				File.Create (path).Close ();
@@ -1064,7 +1097,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void OpenRead ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "AFile.txt";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "AFile.txt";
 			if (!File.Exists (path))
 				File.Create (path).Close ();
 			FileStream stream = null;
@@ -1084,7 +1117,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void OpenWrite ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "AFile.txt";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "AFile.txt";
 			if (!File.Exists (path))
 				File.Create (path).Close ();
 			FileStream stream = null;
@@ -1105,7 +1138,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void TestGetCreationTime ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "baz";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "baz";
 			DeleteFile (path);
 
 			try {
@@ -1162,7 +1195,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void LastAccessTime ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "lastAccessTime";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "lastAccessTime";
 			if (File.Exists (path))
 				File.Delete (path);
 			FileStream stream = null;
@@ -1209,7 +1242,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void LastWriteTime ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "lastWriteTime";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "lastWriteTime";
 			if (File.Exists (path))
 				File.Delete (path);
 			FileStream stream = null;
@@ -1253,6 +1286,61 @@ namespace MonoTests.System.IO
 			}
 		}
 
+		/* regression test for https://github.com/mono/mono/issues/7646 */
+		[Test]
+		public void LastWriteTimeSubMs ()
+		{
+			string path = tmpFolder + Path.DirectorySeparatorChar + "lastWriteTimeSubMs";
+			if (File.Exists (path))
+				File.Delete (path);
+			try {
+				var fmt = "HH:mm:ss:fffffff";
+				for (var i = 0; i < 200; i++) {
+					File.WriteAllText (path, "");
+					var untouched = File.GetLastWriteTimeUtc (path);
+					var now = DateTime.UtcNow;
+					var diff = now - untouched;
+					// sanity check
+					Assert.IsTrue (diff.TotalSeconds >= 0 && diff.TotalSeconds < 2.0, $"Iteration #{i} failed, diff.TotalSeconds: {diff.TotalSeconds}, untouched: {untouched.ToString (fmt)}, now: {now.ToString (fmt)}");
+					File.SetLastWriteTimeUtc (path, now);
+					var touched = File.GetLastWriteTimeUtc (path);
+
+					Assert.IsTrue (touched >= untouched, $"Iteration #{i} failed, untouched: {untouched.ToString (fmt)} touched: {touched.ToString (fmt)}");
+				}
+			} finally {
+				DeleteFile (path);
+			}
+		}
+
+		/* regression test from https://github.com/xamarin/xamarin-macios/issues/3929 */
+		[Test]
+		public void LastWriteTimeSubMsCopy ()
+		{
+			string path = tmpFolder + Path.DirectorySeparatorChar + "lastWriteTimeSubMs";
+			if (File.Exists (path))
+				File.Delete (path);
+
+			string path_copy = tmpFolder + Path.DirectorySeparatorChar + "LastWriteTimeSubMs_copy";
+			if (File.Exists (path_copy))
+				File.Delete (path_copy);
+
+			try {
+				var fmt = "HH:mm:ss:fffffff";
+				for (var i = 0; i < 200; i++) {
+					File.WriteAllText (path, "");
+					var untouched = File.GetLastWriteTimeUtc (path);
+					File.Copy (path, path_copy);
+					var copy_touched = File.GetLastWriteTimeUtc (path_copy);
+
+					Assert.IsTrue (copy_touched >= untouched, $"Iteration #{i} failed, untouched: {untouched.ToString (fmt)} copy_touched: {copy_touched.ToString (fmt)}");
+					File.Delete (path_copy);
+				}
+			} finally {
+				DeleteFile (path);
+				DeleteFile (path_copy);
+			}
+		}
+
 		[Test]
 		public void GetCreationTime_Path_Null ()
 		{
@@ -1285,7 +1373,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void GetCreationTime_Path_DoesNotExist ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "GetCreationTimeException3";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "GetCreationTimeException3";
 			DeleteFile (path);
 
 			DateTime time = File.GetCreationTime (path);
@@ -1360,7 +1448,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void GetCreationTimeUtc_Path_DoesNotExist ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "GetCreationTimeUtcException3";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "GetCreationTimeUtcException3";
 			DeleteFile (path);
 
 			DateTime time = File.GetCreationTimeUtc (path);
@@ -1434,7 +1522,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void GetLastAccessTime_Path_DoesNotExist ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "GetLastAccessTimeException3";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "GetLastAccessTimeException3";
 			DeleteFile (path);
 
 			DateTime time = File.GetLastAccessTime (path);
@@ -1509,7 +1597,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void GetLastAccessTimeUtc_Path_DoesNotExist ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "GetLastAccessTimeUtcException3";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "GetLastAccessTimeUtcException3";
 			DeleteFile (path);
 
 			DateTime time = File.GetLastAccessTimeUtc (path);
@@ -1583,7 +1671,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void GetLastWriteTime_Path_DoesNotExist ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "GetLastAccessTimeUtcException3";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "GetLastAccessTimeUtcException3";
 			DeleteFile (path);
 
 			DateTime time = File.GetLastWriteTime (path);
@@ -1658,7 +1746,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void GetLastWriteTimeUtc_Path_DoesNotExist ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "GetLastWriteTimeUtcException3";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "GetLastWriteTimeUtcException3";
 			DeleteFile (path);
 
 			DateTime time = File.GetLastWriteTimeUtc (path);
@@ -1703,7 +1791,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void FileStreamClose ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "FileStreamClose";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "FileStreamClose";
 			FileStream stream = null;
 			try {
 				stream = File.Create (path);
@@ -1784,7 +1872,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void SetCreationTime_Path_DoesNotExist ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "SetCreationTimeFileNotFoundException1";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "SetCreationTimeFileNotFoundException1";
 			DeleteFile (path);
 			
 			try {
@@ -1802,7 +1890,7 @@ namespace MonoTests.System.IO
 //		[ExpectedException(typeof (ArgumentOutOfRangeException))]
 //		public void SetCreationTimeArgumentOutOfRangeException1 ()
 //		{
-//			string path = TempFolder + Path.DirectorySeparatorChar + "SetCreationTimeArgumentOutOfRangeException1";
+//			string path = tmpFolder + Path.DirectorySeparatorChar + "SetCreationTimeArgumentOutOfRangeException1";
 //			FileStream stream = null;
 //			DeleteFile (path);
 //			try {
@@ -1819,7 +1907,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void SetCreationTime_FileLock ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "CreationTimeIOException1";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "CreationTimeIOException1";
 			DeleteFile (path);
 			FileStream stream = null;
 			try {
@@ -1908,7 +1996,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void SetCreationTimeUtc_Path_DoesNotExist ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "SetCreationTimeUtcFileNotFoundException1";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "SetCreationTimeUtcFileNotFoundException1";
 			DeleteFile (path);
 
 			try {
@@ -1926,7 +2014,7 @@ namespace MonoTests.System.IO
 //		[ExpectedException(typeof (ArgumentOutOfRangeException))]
 //		public void SetCreationTimeUtcArgumentOutOfRangeException1 ()
 //		{
-//			string path = TempFolder + Path.DirectorySeparatorChar + "SetCreationTimeUtcArgumentOutOfRangeException1";
+//			string path = tmpFolder + Path.DirectorySeparatorChar + "SetCreationTimeUtcArgumentOutOfRangeException1";
 //			DeleteFile (path);
 //			FileStream stream = null;
 //			try {
@@ -1943,7 +2031,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void SetCreationTimeUtc_FileLock ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "SetCreationTimeUtcIOException1";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "SetCreationTimeUtcIOException1";
 			DeleteFile (path);
 			FileStream stream = null;
 			try {
@@ -2034,7 +2122,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void SetLastAccessTime_Path_DoesNotExist ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "SetLastAccessTimeFileNotFoundException1";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "SetLastAccessTimeFileNotFoundException1";
 			DeleteFile (path);
 
 			try {
@@ -2052,7 +2140,7 @@ namespace MonoTests.System.IO
 //		[ExpectedException(typeof (ArgumentOutOfRangeException))]
 //		public void SetLastAccessTimeArgumentOutOfRangeException1 ()
 //		{
-//			string path = TempFolder + Path.DirectorySeparatorChar + "SetLastTimeArgumentOutOfRangeException1";
+//			string path = tmpFolder + Path.DirectorySeparatorChar + "SetLastTimeArgumentOutOfRangeException1";
 //			DeleteFile (path);
 //			FileStream stream = null;
 //			try {
@@ -2069,7 +2157,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void SetLastAccessTime_FileLock ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "LastAccessIOException1";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "LastAccessIOException1";
 			DeleteFile (path);
 			FileStream stream = null;
 			try {
@@ -2158,7 +2246,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void SetLastAccessTimeUtc_Path_DoesNotExist ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "SetLastAccessTimeUtcFileNotFoundException1";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "SetLastAccessTimeUtcFileNotFoundException1";
 			DeleteFile (path);
 
 			try {
@@ -2176,7 +2264,7 @@ namespace MonoTests.System.IO
 //		[ExpectedException(typeof (ArgumentOutOfRangeException))]
 //		public void SetLastAccessTimeUtcArgumentOutOfRangeException1 ()
 //		{
-//			string path = TempFolder + Path.DirectorySeparatorChar + "SetLastAccessTimeUtcArgumentOutOfRangeException1";
+//			string path = tmpFolder + Path.DirectorySeparatorChar + "SetLastAccessTimeUtcArgumentOutOfRangeException1";
 //			DeleteFile (path);
 //			FileStream stream = null;
 //			try {
@@ -2193,7 +2281,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void SetLastAccessTimeUtc_FileLock ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "SetLastAccessTimeUtcIOException1";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "SetLastAccessTimeUtcIOException1";
 			DeleteFile (path);
 			FileStream stream = null;
 			try {
@@ -2284,7 +2372,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void SetLastWriteTime_Path_DoesNotExist ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "SetLastWriteTimeFileNotFoundException1";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "SetLastWriteTimeFileNotFoundException1";
 			DeleteFile (path);
 
 			try {
@@ -2302,7 +2390,7 @@ namespace MonoTests.System.IO
 //		[ExpectedException(typeof (ArgumentOutOfRangeException))]
 //		public void SetLastWriteTimeArgumentOutOfRangeException1 ()
 //		{
-//			string path = TempFolder + Path.DirectorySeparatorChar + "SetLastWriteTimeArgumentOutOfRangeException1";
+//			string path = tmpFolder + Path.DirectorySeparatorChar + "SetLastWriteTimeArgumentOutOfRangeException1";
 //			DeleteFile (path);
 //			FileStream stream = null;
 //			try {
@@ -2319,7 +2407,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void SetLastWriteTime_FileLock ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "LastWriteTimeIOException1";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "LastWriteTimeIOException1";
 			DeleteFile (path);
 			FileStream stream = null;
 			try {
@@ -2408,7 +2496,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void SetLastWriteTimeUtc_Path_DoesNotExist ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "SetLastWriteTimeUtcFileNotFoundException1";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "SetLastWriteTimeUtcFileNotFoundException1";
 			DeleteFile (path);
 
 			try {
@@ -2426,7 +2514,7 @@ namespace MonoTests.System.IO
 //		[ExpectedException(typeof (ArgumentOutOfRangeException))]
 //		public void SetLastWriteTimeUtcArgumentOutOfRangeException1 ()
 //		{
-//			string path = TempFolder + Path.DirectorySeparatorChar + "SetLastWriteTimeUtcArgumentOutOfRangeException1";
+//			string path = tmpFolder + Path.DirectorySeparatorChar + "SetLastWriteTimeUtcArgumentOutOfRangeException1";
 //			DeleteFile (path);
 //			FileStream stream = null;
 //			try {
@@ -2443,7 +2531,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void SetLastWriteTimeUtc_FileLock ()
 		{
-			string path = TempFolder + Path.DirectorySeparatorChar + "SetLastWriteTimeUtcIOException1";
+			string path = tmpFolder + Path.DirectorySeparatorChar + "SetLastWriteTimeUtcIOException1";
 			DeleteFile (path);
 			FileStream stream = null;
 			try {
@@ -2491,6 +2579,7 @@ namespace MonoTests.System.IO
 		}
 
 		[Test]
+		[Category ("LargeFileSupport")]
 		public void Position_Large ()
 		{
 			// fails if HAVE_LARGE_FILE_SUPPORT is not enabled in device builds
@@ -2513,6 +2602,7 @@ namespace MonoTests.System.IO
 		}
 
 		[Test]
+		[Category ("LargeFileSupport")]
 		public void Seek_Large ()
 		{
 			// fails if HAVE_LARGE_FILE_SUPPORT is not enabled in device builds
@@ -2537,6 +2627,7 @@ namespace MonoTests.System.IO
 		}
 
 		[Test]
+		[Category ("LargeFileSupport")]
 		public void Lock_Large ()
 		{
 			// note: already worked without HAVE_LARGE_FILE_SUPPORT
@@ -2569,7 +2660,7 @@ namespace MonoTests.System.IO
 		[Test]
 		public void ReplaceTest ()
 		{
-			string tmp = Path.Combine (TempFolder, "ReplaceTest");
+			string tmp = Path.Combine (tmpFolder, "ReplaceTest");
 			Directory.CreateDirectory (tmp);
 			string origFile = Path.Combine (tmp, "origFile");
 			string replaceFile = Path.Combine (tmp, "replaceFile");
@@ -2677,5 +2768,58 @@ namespace MonoTests.System.IO
 			MoveTest (FileAccess.ReadWrite, FileShare.Write | FileShare.Delete, true);
 			MoveTest (FileAccess.ReadWrite, FileShare.ReadWrite | FileShare.Delete, true);
 		}
+
+
+		[DllImport ("libc", SetLastError=true)]
+		public static extern int symlink (string oldpath, string newpath);
+
+		[Test]
+#if MONOTOUCH_TV
+		[Ignore ("See bug #59239")]
+#endif
+		public void SymLinkLoop ()
+		{
+			if (!RunningOnUnix)
+				Assert.Ignore ("Symlink are hard on windows");
+
+			var name1 = Path.GetRandomFileName ();
+			var name2 = Path.GetRandomFileName ();
+
+			var path1 = Path.Combine (Path.GetTempPath (), name1);
+			var path2 = Path.Combine (Path.GetTempPath (), name2);
+
+			File.Delete (path1);
+			File.Delete (path2);
+
+			try {
+				if (symlink (path1, path2) != 0)
+					Assert.Fail ("symlink #1 failed with errno={0}", Marshal.GetLastWin32Error ());
+				if (symlink (path2, path1) != 0)
+					Assert.Fail ("symlink #2 failed with errno={0}", Marshal.GetLastWin32Error ());
+
+				Assert.IsTrue (File.Exists (path1), "File.Exists must return true for path1 symlink loop");
+				Assert.IsTrue (File.Exists (path2), "File.Exists must return true for path2 symlink loop");
+
+				try {
+					using (var f = File.Open (path1, FileMode.Open, FileAccess.Read)) {
+						Assert.Fail ("File.Open must throw for symlink loops");
+					}
+				} catch (IOException ex) {
+					Assert.AreEqual (0x80070781u, (uint)ex.HResult, "Ensure HRESULT is correct");
+				}
+
+				File.Delete (path1); //Delete must not throw and must work
+				Assert.IsFalse (File.Exists (path1), "File.Delete must delete symlink loops");
+
+			} finally {
+				try {
+					File.Delete (path1);
+					File.Delete (path2);
+				} catch (IOException) {
+					//Don't double fault any exception from the tests.
+				}
+
+			}
+		}
 	}
 }
