@@ -31,7 +31,7 @@ using System;
 using System.Reflection;
 using System.Runtime.InteropServices;
 using System.Threading;
-#if !MONOTOUCH && !MOBILE_STATIC
+#if !MONOTOUCH && !FULL_AOT_RUNTIME
 using System.Reflection.Emit;
 #endif
 using System.IO;
@@ -355,7 +355,7 @@ namespace MonoTests.System.Reflection
 				get { return 99; }
 			}
 		}
-#if !MONOTOUCH && !MOBILE_STATIC
+#if !MONOTOUCH && !FULL_AOT_RUNTIME
 		[Test]
 		public void ConstantValue () {
 			/*This test looks scary because we can't generate a default value with C# */
@@ -560,5 +560,12 @@ namespace MonoTests.System.Reflection
 			Assert.AreEqual ("param", defaultParam.Name, "#1");
 			Assert.AreEqual ("test", defaultParam.DefaultValue, "#2");
 		}
+
+		[Test]
+		public void MetadataToken ()
+		{
+			PropertyInfo property = typeof (Base).GetProperty ("P");
+			Assert.IsTrue ((int)property.MetadataToken > 0);
+		}
 	}
 }
