@@ -14,6 +14,7 @@ using System;
 using System.IO;
 using System.Threading;
 using System.Globalization;
+using System.Reflection;
 using System.Runtime.Serialization.Formatters.Binary;
 
 namespace MonoTests.System {
@@ -23,17 +24,17 @@ public class TimeZoneTest {
 
 	private void CET (TimeZone t1) 
 	{
-		Assert.AreEqual("CET", t1.StandardName, "A01");
-		Assert.AreEqual("CEST", t1.DaylightName, "A02");
+		Assert.IsTrue("CET" == t1.StandardName || "W. Europe Standard Time" == t1.StandardName, "A01");
+		Assert.IsTrue("CEST" == t1.DaylightName || "W. Europe Daylight Time" == t1.DaylightName, "A02");
 	
 		DaylightTime d1 = t1.GetDaylightChanges (2002);
-		Assert.AreEqual("03/31/2002 02:00:00", d1.Start.ToString ("G"), "A03");
-		Assert.AreEqual("10/27/2002 03:00:00", d1.End.ToString ("G"), "A04");
+		Assert.AreEqual("03/31/2002 02:00:00", d1.Start.ToString ("G", CultureInfo.InvariantCulture), "A03");
+		Assert.AreEqual("10/27/2002 03:00:00", d1.End.ToString ("G", CultureInfo.InvariantCulture), "A04");
 		Assert.AreEqual(36000000000L, d1.Delta.Ticks, "A05");
 	
 		DaylightTime d2 = t1.GetDaylightChanges (1996);
-		Assert.AreEqual("03/31/1996 02:00:00", d2.Start.ToString ("G"), "A06");
-		Assert.AreEqual("10/27/1996 03:00:00", d2.End.ToString ("G"), "A07");
+		Assert.AreEqual("03/31/1996 02:00:00", d2.Start.ToString ("G", CultureInfo.InvariantCulture), "A06");
+		Assert.AreEqual("10/27/1996 03:00:00", d2.End.ToString ("G", CultureInfo.InvariantCulture), "A07");
 		Assert.AreEqual(36000000000L, d2.Delta.Ticks, "A08");
 	
 		DateTime d3 = new DateTime (2002,2,25);
@@ -46,34 +47,44 @@ public class TimeZoneTest {
 		Assert.AreEqual(36000000000L, t1.GetUtcOffset (d3).Ticks, "A12");
 		Assert.AreEqual(72000000000L, t1.GetUtcOffset (d4).Ticks, "A13");
 		Assert.AreEqual(36000000000L, t1.GetUtcOffset (d5).Ticks, "A14");
+
+		// Test TimeZone methods with UTC DateTime in DST.
+		DateTime d6 = d4.ToUniversalTime ();
+		Assert.AreEqual(false, t1.IsDaylightSavingTime (d6), "A15");
+		Assert.AreEqual(0, t1.GetUtcOffset (d6).Ticks, "A16");
 	}
 
 	private void EST (TimeZone t1) 
 	{
-		// It could be EST though...
-		//Assert.AreEqual("Eastern Standard Time", t1.StandardName, "B01");
-		//Assert.AreEqual("Eastern Daylight Time", t1.DaylightName, "B02");
+		Assert.IsTrue("EST" == t1.StandardName || "Eastern Standard Time" == t1.StandardName, "B01");
+		Assert.IsTrue("EDT" == t1.DaylightName || "Eastern Daylight Time" == t1.DaylightName, "B02");
 
 		DaylightTime d1 = t1.GetDaylightChanges (2002);
-		Assert.AreEqual("04/07/2002 02:00:00", d1.Start.ToString ("G"), "B03");
-		Assert.AreEqual("10/27/2002 02:00:00", d1.End.ToString ("G"), "B04");
+		Assert.AreEqual("04/07/2002 02:00:00", d1.Start.ToString ("G", CultureInfo.InvariantCulture), "B03");
+		Assert.AreEqual("10/27/2002 02:00:00", d1.End.ToString ("G", CultureInfo.InvariantCulture), "B04");
 		Assert.AreEqual(36000000000L, d1.Delta.Ticks, "B05");
 
 		DaylightTime d2 = t1.GetDaylightChanges (1996);
-		Assert.AreEqual("04/07/1996 02:00:00", d2.Start.ToString ("G"), "B06");
-		Assert.AreEqual("10/27/1996 02:00:00", d2.End.ToString ("G"), "B07");
+		Assert.AreEqual("04/07/1996 02:00:00", d2.Start.ToString ("G", CultureInfo.InvariantCulture), "B06");
+		Assert.AreEqual("10/27/1996 02:00:00", d2.End.ToString ("G", CultureInfo.InvariantCulture), "B07");
 		Assert.AreEqual(36000000000L, d2.Delta.Ticks, "B08");
 
 		DateTime d3 = new DateTime (2002,2,25);
 		Assert.AreEqual(false, t1.IsDaylightSavingTime (d3), "B09");
 		DateTime d4 = new DateTime (2002,4,8);
 		Assert.AreEqual(true, t1.IsDaylightSavingTime (d4), "B10");
+		
 		DateTime d5 = new DateTime (2002,11,4);
 		Assert.AreEqual(false, t1.IsDaylightSavingTime (d5), "B11");
 
 		Assert.AreEqual(-180000000000L, t1.GetUtcOffset (d3).Ticks, "B12");
 		Assert.AreEqual(-144000000000L, t1.GetUtcOffset (d4).Ticks, "B13");
 		Assert.AreEqual(-180000000000L, t1.GetUtcOffset (d5).Ticks, "B14");
+
+		// Test TimeZone methods with UTC DateTime in DST.
+		DateTime d6 = d4.ToUniversalTime ();
+		Assert.AreEqual(false, t1.IsDaylightSavingTime (d6), "B15");
+		Assert.AreEqual(0, t1.GetUtcOffset (d6).Ticks, "B16");
 	}
 
 	private void TST (TimeZone t1) 
@@ -101,13 +112,13 @@ public class TimeZoneTest {
 		Assert.IsTrue("BST" == t1.DaylightName || "IST" == t1.DaylightName, "D02");
 	
 		DaylightTime d1 = t1.GetDaylightChanges (2002);
-		Assert.AreEqual("03/31/2002 01:00:00", d1.Start.ToString ("G"), "D03");
-		Assert.AreEqual("10/27/2002 02:00:00", d1.End.ToString ("G"), "D04");
+		Assert.AreEqual("03/31/2002 01:00:00", d1.Start.ToString ("G", CultureInfo.InvariantCulture), "D03");
+		Assert.AreEqual("10/27/2002 02:00:00", d1.End.ToString ("G", CultureInfo.InvariantCulture), "D04");
 		Assert.AreEqual(36000000000L, d1.Delta.Ticks, "D05");
 	
 		DaylightTime d2 = t1.GetDaylightChanges (1996);
-		Assert.AreEqual("03/31/1996 01:00:00", d2.Start.ToString ("G"), "D06");
-		Assert.AreEqual("10/27/1996 02:00:00", d2.End.ToString ("G"), "D07");
+		Assert.AreEqual("03/31/1996 01:00:00", d2.Start.ToString ("G", CultureInfo.InvariantCulture), "D06");
+		Assert.AreEqual("10/27/1996 02:00:00", d2.End.ToString ("G", CultureInfo.InvariantCulture), "D07");
 		Assert.AreEqual(36000000000L, d2.Delta.Ticks, "D08");
 	
 		DateTime d3 = new DateTime (2002,2,25);
@@ -120,6 +131,11 @@ public class TimeZoneTest {
 		Assert.AreEqual(0L, t1.GetUtcOffset (d3).Ticks, "D12");
 		Assert.AreEqual(36000000000L, t1.GetUtcOffset (d4).Ticks, "D13");
 		Assert.AreEqual(0L, t1.GetUtcOffset (d5).Ticks, "D14");
+
+		// Test TimeZone methods with UTC DateTime in DST.
+		DateTime d6 = d4.ToUniversalTime ();
+		Assert.AreEqual(false, t1.IsDaylightSavingTime (d6), "D15");
+		Assert.AreEqual(0, t1.GetUtcOffset (d6).Ticks, "D16");
 	}
 
 	private void NZST(TimeZone t1) {
@@ -127,13 +143,13 @@ public class TimeZoneTest {
 		Assert.AreEqual("NZDT", t1.DaylightName, "E02");
 
 		DaylightTime d1 = t1.GetDaylightChanges (2013);
-		Assert.AreEqual("09/29/2013 02:00:00", d1.Start.ToString ("G"), "E03");
-		Assert.AreEqual("04/07/2013 03:00:00", d1.End.ToString ("G"), "E04");
+		Assert.AreEqual("09/29/2013 02:00:00", d1.Start.ToString ("G", CultureInfo.InvariantCulture), "E03");
+		Assert.AreEqual("04/07/2013 03:00:00", d1.End.ToString ("G", CultureInfo.InvariantCulture), "E04");
 		Assert.AreEqual(36000000000L, d1.Delta.Ticks, "E05");
 
 		DaylightTime d2 = t1.GetDaylightChanges (2001);
-		Assert.AreEqual("10/07/2001 02:00:00", d2.Start.ToString ("G"), "E06");
-		Assert.AreEqual("03/18/2001 03:00:00", d2.End.ToString ("G"), "E07");
+		Assert.AreEqual("10/07/2001 02:00:00", d2.Start.ToString ("G", CultureInfo.InvariantCulture), "E06");
+		Assert.AreEqual("03/18/2001 03:00:00", d2.End.ToString ("G", CultureInfo.InvariantCulture), "E07");
 		Assert.AreEqual(36000000000L, d2.Delta.Ticks, "E08");
 
 		DateTime d3 = new DateTime(2013,02,15);
@@ -146,14 +162,19 @@ public class TimeZoneTest {
 		Assert.AreEqual(36000000000L /*hour*/ * 13L, t1.GetUtcOffset (d3).Ticks, "E12");
 		Assert.AreEqual(36000000000L /*hour*/ * 12L, t1.GetUtcOffset (d4).Ticks, "E13");
 		Assert.AreEqual(36000000000L /*hour*/ * 13L, t1.GetUtcOffset (d5).Ticks, "E14");
+
+		// Test TimeZone methods with UTC DateTime in DST.
+		DateTime d6 = d5.ToUniversalTime ();
+		Assert.AreEqual(false, t1.IsDaylightSavingTime (d6), "E15");
+		Assert.AreEqual(0, t1.GetUtcOffset (d6).Ticks, "E16");
 	}
 
 	[Test]
-	[Culture ("")]
 	public void TestCtors ()
 	{
 		TimeZone t1 = TimeZone.CurrentTimeZone;
 		switch (t1.StandardName) {
+			case "W. Europe Standard Time":
 			case "CET":
 				CET (t1);
 				break;
@@ -165,6 +186,10 @@ public class TimeZoneTest {
 				TST (t1);
 				break;
 			case "GMT":
+#if MONOTOUCH
+				if (string.IsNullOrEmpty (t1.DaylightName))
+					Assert.Ignore ("This test may fail due to: http://www.openradar.me/38174449");
+#endif
 				GMT (t1);
 				break;
 			case "NZST":
@@ -250,6 +275,47 @@ public class TimeZoneTest {
 		Assert.IsTrue (tz.ToLocalTime (dst_start_utc.Add (new TimeSpan (1, 0, 0))) < tz.ToLocalTime (dst_start_utc.Add (new TimeSpan (1, 1, 0))), "0:4:00 < 0:4:01");
 	}
 
+		[Test]
+		public void GetUTCNowAtDSTBoundaries ()
+		{
+			TimeZoneInfo.TransitionTime startTransition = TimeZoneInfo.TransitionTime.CreateFloatingDateRule(new DateTime(1, 1, 1, 2, 0, 0), 3, 5, DayOfWeek.Sunday);
+
+			TimeZoneInfo.TransitionTime endTransition = TimeZoneInfo.TransitionTime.CreateFloatingDateRule(new DateTime(1, 1, 1, 3, 0, 0), 10, 5, DayOfWeek.Sunday);
+
+			TimeSpan delta = TimeSpan.FromMinutes(60.0);
+			TimeZoneInfo.AdjustmentRule adjustment = TimeZoneInfo.AdjustmentRule.CreateAdjustmentRule(new DateTime(1970, 1, 1), DateTime.MaxValue.Date, delta, startTransition, endTransition);
+			TimeZoneInfo.TransitionTime startTrans = adjustment.DaylightTransitionStart;
+			TimeZoneInfo.TransitionTime endTrans = adjustment.DaylightTransitionEnd;
+			TimeZoneInfo.AdjustmentRule[] adjustments = { adjustment };
+
+			TimeZoneInfo tzInfo = TimeZoneInfo.CreateCustomTimeZone("MY Standard Time", TimeSpan.Zero, "MST", "MST", "MDT", adjustments);
+
+			// There is no .NET API to set timezone. Use reflection to assign time zone to the TimeZoneInfo.local field.
+			FieldInfo localTimeZone = typeof(TimeZoneInfo).GetField("local", BindingFlags.NonPublic | BindingFlags.Static | BindingFlags.Instance);
+			localTimeZone.SetValue(null, tzInfo);
+
+			DateTime st = new DateTime(2016, 3, 27, 1, 0, 0, DateTimeKind.Local);
+			Assert.IsTrue (!tzInfo.IsDaylightSavingTime(st));	
+			Assert.IsTrue (!tzInfo.IsAmbiguousTime(st));
+			Assert.IsTrue ((TimeZoneInfo.ConvertTimeToUtc(st).Hour == 1));
+			st = new DateTime(2016, 3, 27, 3, 0, 0, DateTimeKind.Local);
+			Assert.IsTrue (tzInfo.IsDaylightSavingTime(st));	
+			Assert.IsTrue (!tzInfo.IsAmbiguousTime(st));
+			Assert.IsTrue ((TimeZoneInfo.ConvertTimeToUtc(st).Hour == 2));
+			st = new DateTime(2016, 10, 30, 2, 0, 0, DateTimeKind.Local);
+			Assert.IsTrue (tzInfo.IsDaylightSavingTime(st));	
+			Assert.IsTrue (!tzInfo.IsAmbiguousTime(st));
+			Assert.IsTrue ((TimeZoneInfo.ConvertTimeToUtc(st).Hour == 1));
+			st = new DateTime(2016, 10, 30, 3, 0, 0, DateTimeKind.Local);
+			Assert.IsTrue (!tzInfo.IsDaylightSavingTime(st));	
+			Assert.IsTrue (tzInfo.IsAmbiguousTime(st));
+			Assert.IsTrue ((TimeZoneInfo.ConvertTimeToUtc(st).Hour == 3));
+			st = new DateTime(2016, 10, 30, 4, 0, 0, DateTimeKind.Local);
+			Assert.IsTrue (!tzInfo.IsDaylightSavingTime(st));	
+			Assert.IsTrue (!tzInfo.IsAmbiguousTime(st));
+			Assert.IsTrue ((TimeZoneInfo.ConvertTimeToUtc(st).Hour == 4));
+		}
+
 		[Test]
 		public void GetUtcOffsetAtDSTBoundary ()
 		{
@@ -282,11 +348,12 @@ public class TimeZoneTest {
 				Assert.Ignore (tz.StandardName + " did not observe daylight saving time during " + year + ".");
 
 			var standardOffset = tz.GetUtcOffset(daylightChanges.Start.AddMinutes(-1));
+			var dstOffset = tz.GetUtcOffset(daylightChanges.Start.AddMinutes(61));
 
-			Assert.AreEqual(standardOffset, tz.GetUtcOffset (dst_end));
-			Assert.AreEqual(standardOffset, tz.GetUtcOffset (dst_end.Add (daylightChanges.Delta.Negate ().Add (TimeSpan.FromSeconds(1)))));
-			Assert.AreEqual(standardOffset, tz.GetUtcOffset (dst_end.Add(daylightChanges.Delta.Negate ())));
-			Assert.AreNotEqual(standardOffset, tz.GetUtcOffset (dst_end.Add(daylightChanges.Delta.Negate ().Add (TimeSpan.FromSeconds(-1)))));
+//			Assert.AreEqual(standardOffset, tz.GetUtcOffset (dst_end));
+			Assert.AreEqual(dstOffset, tz.GetUtcOffset (dst_end.Add (daylightChanges.Delta.Negate ().Add (TimeSpan.FromSeconds(1)))));
+			Assert.AreEqual(dstOffset, tz.GetUtcOffset (dst_end.Add(daylightChanges.Delta.Negate ())));
+			Assert.AreEqual(dstOffset, tz.GetUtcOffset (dst_end.Add(daylightChanges.Delta.Negate ().Add (TimeSpan.FromSeconds(-1)))));
 		}
 
 
@@ -300,42 +367,33 @@ public class TimeZoneTest {
 		[Test]
 		public void FindSystemTimeZoneById ()
 		{
-			TimeZoneInfo tzi = TimeZoneInfo.FindSystemTimeZoneById ("Canada/Eastern");
-			Assert.AreEqual ("EDT", tzi.DaylightName, "DaylightName");
-			Assert.AreEqual ("EST", tzi.StandardName, "StandardName");
+			TimeZoneInfo tzi = TimeZoneInfo.FindSystemTimeZoneById (TimeZoneInfoTest.MapTimeZoneId ("Canada/Eastern"));
+			Assert.IsTrue ("EDT" == tzi.DaylightName || "Eastern Daylight Time" == tzi.DaylightName, "DaylightName");
+			Assert.IsTrue ("EST" == tzi.StandardName || "Eastern Standard Time" == tzi.StandardName, "StandardName");
 			Assert.IsTrue (tzi.SupportsDaylightSavingTime, "SupportsDaylightSavingTime");
 		}
 
-#if MOBILE
+		[Test]
+		public void OldEraToLocalTime ()
+		{
+			TimeSpan offset = TimeSpan.Zero;
+			var dto = new DateTimeOffset (new DateTime (1900, 1, 1).Ticks, offset);
+
+			// Should never throw
+			dto.ToLocalTime ();
+		}
+
+#if !WIN_PLATFORM
 		// On device we cannot read the OS file system to look for /etc/localtime
 		// and /usr/share/zoneinfo - so we must initialize the BCL TimeZoneInfo
 		// from NSTimeZoneInfo. The tests here check the code paths between the
 		// two types - if they break then TimeZoneInfo work work at all
 		// ref: http://bugzilla.xamarin.com/show_bug.cgi?id=1790
 		
-		bool incomplete_data_on_simulator_only_bug;
-
 		[Test]
 		public void GetSystemTimeZones ()
 		{
-			// if test is executed a second time then it report less than 400 (about 127) items available
-			if (incomplete_data_on_simulator_only_bug)
-				Assert.Ignore ("known to fail on some iOS simulator versions - see source comments");
-			
-			try {
-				Assert.That (TimeZoneInfo.GetSystemTimeZones ().Count, Is.GreaterThan (400), "GetSystemTimeZones");
-			} catch (NullReferenceException) {
-				// that's a weird one. It failed on iOS 5.1 *beta* simulator (on Lion) but it worked on *final*
-				// now it fails on Snow Leopard the same way (incomplete data) with iOS5 simulator (OS update ?)
-				// but it *never*ever* failed on devices
-				incomplete_data_on_simulator_only_bug = true;
-#if XAMCORE_2_0
-				if (ObjCRuntime.Runtime.Arch == ObjCRuntime.Arch.SIMULATOR)
-#else
-				if (MonoTouch.ObjCRuntime.Runtime.Arch == MonoTouch.ObjCRuntime.Arch.SIMULATOR)
-#endif
-					Assert.Ignore ("known to fail on some iOS simulator versions - see source comments");
-			}
+			Assert.That (TimeZoneInfo.GetSystemTimeZones ().Count, Is.GreaterThan (400), "GetSystemTimeZones");
 		}
 #endif
 	}
