@@ -20,11 +20,8 @@
 #include "mono/metadata/class.h"
 #include "mono/metadata/class-internals.h"
 #include "mono/utils/mono-compiler.h"
-
-#if defined(__native_client__) && defined(__GLIBC__)
-volatile int __nacl_thread_suspension_needed = 0;
-void __nacl_suspend_thread_if_needed() {}
-#endif
+#include "mono/utils/mono-error-internals.h"
+#include "mono/utils/mono-math.h"
 
 void
 dump_table_assembly (MonoImage *m)
@@ -288,8 +285,7 @@ dump_table_memberref (MonoImage *m)
 			 x ? x : "",
 			 sig);
 
-		if (x)
-			g_free (x);
+		g_free (x);
 		g_free (sig);
 	}
 }
@@ -565,7 +561,7 @@ dump_table_method (MonoImage *m)
 	current_type = 1;
 	last_m = first_m = 1;
 	for (i = 1; i <= t->rows; i++){
-		MonoError error;
+		ERROR_DECL (error);
 		guint32 cols [MONO_METHOD_SIZE];
 		char *sig, *impl_flags;
 		const char *sigblob;
@@ -582,23 +578,29 @@ dump_table_method (MonoImage *m)
 				mono_metadata_string_heap (m, mono_metadata_decode_row_col (td, current_type - 2, MONO_TYPEDEF_NAMESPACE)),
 				mono_metadata_string_heap (m, mono_metadata_decode_row_col (td, current_type - 2, MONO_TYPEDEF_NAME)));
 			first_m = last_m;
-			type_container = mono_metadata_load_generic_params (m, MONO_TOKEN_TYPE_DEF | (current_type - 1), NULL);
+			type_container = mono_metadata_load_generic_params (m, MONO_TOKEN_TYPE_DEF | (current_type - 1), NULL, NULL);
 			if (type_container) {
-				mono_metadata_load_generic_param_constraints_checked (m, MONO_TOKEN_TYPE_DEF | (current_type - 1), type_container, &error);
-				g_assert (mono_error_ok (&error)); /*FIXME don't swallow the error message*/
+				mono_metadata_load_generic_param_constraints_checked (m, MONO_TOKEN_TYPE_DEF | (current_type - 1), type_container, error);
+				g_assert (mono_error_ok (error)); /*FIXME don't swallow the error message*/
 			}
 		}
 
-		method_container = mono_metadata_load_generic_params (m, MONO_TOKEN_METHOD_DEF | i, type_container);
+		method_container = mono_metadata_load_generic_params (m, MONO_TOKEN_METHOD_DEF | i, type_container, NULL);
 		if (method_container) {
-			mono_metadata_load_generic_param_constraints_checked (m, MONO_TOKEN_METHOD_DEF | i, method_container, &error);
-			g_assert (mono_error_ok (&error)); /*FIXME don't swallow the error message*/
+			mono_metadata_load_generic_param_constraints_checked (m, MONO_TOKEN_METHOD_DEF | i, method_container, error);
+			g_assert (mono_error_ok (error)); /*FIXME don't swallow the error message*/
 		}
 		mono_metadata_decode_table_row (m, MONO_TABLE_METHOD, i - 1, cols, MONO_METHOD_SIZE);
 		sigblob = mono_metadata_blob_heap (m, cols [MONO_METHOD_SIGNATURE]);
 		mono_metadata_decode_blob_size (sigblob, &sigblob);
-		method = mono_metadata_parse_method_signature_full (m, method_container ? method_container : type_container, i, sigblob, &sigblob, &error);
-		g_assert (mono_error_ok (&error)); /*FIXME don't swallow the error message*/
+		method = mono_metadata_parse_method_signature_full (m, method_container ? method_container : type_container, i, sigblob, &sigblob, error);
+		if (!mono_error_ok (error)) {
+			fprintf (output,"%d: failed to parse due to %s\n", i, mono_error_get_message (error));
+			mono_error_cleanup (error);
+			continue;
+		}
+
+		g_assert (mono_error_ok (error)); /*FIXME don't swallow the error message*/
 		sig = dis_stringify_method_signature (m, method, i, method_container ? method_container : type_container, FALSE);
                 impl_flags = get_method_impl_flags (cols [MONO_METHOD_IMPLFLAGS]);
 		fprintf (output, "%d: %s (param: %d impl_flags: %s)\n", i, sig, cols [MONO_METHOD_PARAMLIST], impl_flags);
@@ -865,14 +867,13 @@ handle_enum:
 			break;
 		case MONO_TYPE_R4: {
 			float val;
-			int inf;
 			readr4 (p, &val);
-			inf = dis_isinf (val);
+			const int inf = mono_isinf (val);
 			if (inf == -1) 
 				g_string_append_printf (res, "(00 00 80 ff)"); /* negative infinity */
 			else if (inf == 1)
 				g_string_append_printf (res, "(00 00 80 7f)"); /* positive infinity */
-			else if (dis_isnan (val))
+			else if (mono_isnan (val))
 				g_string_append_printf (res, "(00 00 c0 ff)"); /* NaN */
 			else
 				g_string_append_printf (res, "%g", val);
@@ -881,15 +882,13 @@ handle_enum:
 		}
 		case MONO_TYPE_R8: {
 			double val;
-			int inf;
-			
 			readr8 (p, &val);
-			inf = dis_isinf (val);
+			const int inf = mono_isinf (val);
 			if (inf == -1) 
 				g_string_append_printf (res, "(00 00 00 00 00 00 f0 ff)"); /* negative infinity */
 			else if (inf == 1)
 				g_string_append_printf (res, "(00 00 00 00 00 00 f0 7f)"); /* positive infinity */
-			else if (isnan (val))
+			else if (mono_isnan (val))
 				g_string_append_printf (res, "(00 00 00 00 00 00 f8 ff)"); /* NaN */
 			else
 				g_string_append_printf (res, "%g", val);
@@ -897,8 +896,8 @@ handle_enum:
 			break;
 		}
 		case MONO_TYPE_VALUETYPE:
-			if (mono_class_is_enum (sig->params [i]->data.klass)) {
-				type = mono_class_enum_basetype (sig->params [i]->data.klass)->type;
+			if (m_class_is_enumtype (sig->params [i]->data.klass)) {
+				type = mono_class_enum_basetype_internal (sig->params [i]->data.klass)->type;
 				goto handle_enum;
 			} else {
 				g_warning ("generic valutype not handled in custom attr value decoding");
@@ -944,6 +943,7 @@ dump_table_customattr (MonoImage *m)
 
 	fprintf (output, "Custom Attributes Table (1..%d)\n", t->rows);
 	for (i = 1; i <= t->rows; i++) {
+		ERROR_DECL (error);
 		guint32 cols [MONO_CUSTOM_ATTR_SIZE];
 		guint32 mtoken;
 		char * desc;
@@ -966,12 +966,18 @@ dump_table_customattr (MonoImage *m)
 			break;
 		}
 		method = get_method (m, mtoken, NULL);
-		meth = mono_get_method (m, mtoken, NULL);
-		params = custom_attr_params (m, mono_method_signature (meth), mono_metadata_blob_heap (m, cols [MONO_CUSTOM_ATTR_VALUE]));
-		fprintf (output, "%d: %s: %s [%s]\n", i, desc, method, params);
+		meth = mono_get_method_checked (m, mtoken, NULL, NULL, error);
+		if (meth) {
+			params = custom_attr_params (m, mono_method_signature (meth), mono_metadata_blob_heap (m, cols [MONO_CUSTOM_ATTR_VALUE]));
+			fprintf (output, "%d: %s: %s [%s]\n", i, desc, method, params);
+			g_free (params);
+		} else {
+			fprintf (output, "Could not decode method due to %s", mono_error_get_message (error));
+			mono_error_cleanup (error);
+		}
+
 		g_free (desc);
 		g_free (method);
-		g_free (params);
 	}
 }
 
