@@ -1,5 +1,5 @@
-/*
- * debug-helpers.c:
+/**
+ * \file
  *
  * Author:
  *	Mono Project (http://www.mono-project.com)
@@ -8,6 +8,7 @@
  * Licensed under the MIT license. See LICENSE file in the project root for full license information.
  */
 
+#include <config.h>
 #include <string.h>
 #include "mono/metadata/tokentype.h"
 #include "mono/metadata/opcodes.h"
@@ -18,6 +19,18 @@
 #include "mono/metadata/debug-helpers.h"
 #include "mono/metadata/tabledefs.h"
 #include "mono/metadata/appdomain.h"
+#include "mono/metadata/abi-details.h"
+#ifdef MONO_CLASS_DEF_PRIVATE
+/* Rationale: we want the functions in this file to work even when everything
+ * is broken.  They may be called from a debugger session, for example.  If
+ * MonoClass getters include assertions or trigger class loading, we don't want
+ * that kicked off by a call to one of the functions in here.
+ */
+#define REALLY_INCLUDE_CLASS_DEF 1
+#include <mono/metadata/class-private-definition.h>
+#undef REALLY_INCLUDE_CLASS_DEF
+#endif
+
 
 struct MonoMethodDesc {
 	char *name_space;
@@ -41,13 +54,13 @@ static const struct msgstr_t {
 #undef WRAPPER
 };
 static const gint16 opidx [] = {
-#define WRAPPER(a,b) [MONO_WRAPPER_ ## a] = offsetof (struct msgstr_t, MSGSTRFIELD(__LINE__)),
+#define WRAPPER(a,b) offsetof (struct msgstr_t, MSGSTRFIELD(__LINE__)),
 #include "wrapper-types.h"
 #undef WRAPPER
 };
 
-static const char*
-wrapper_type_to_str (guint32 wrapper_type)
+const char*
+mono_wrapper_type_to_str (guint32 wrapper_type)
 {
 	g_assert (wrapper_type < MONO_WRAPPER_NUM);
 
@@ -62,8 +75,8 @@ wrapper_type_names [MONO_WRAPPER_NUM + 1] = {
 	NULL
 };
 
-static const char*
-wrapper_type_to_str (guint32 wrapper_type)
+const char*
+mono_wrapper_type_to_str (guint32 wrapper_type)
 {
 	g_assert (wrapper_type < MONO_WRAPPER_NUM);
 
@@ -162,14 +175,14 @@ mono_type_get_desc (GString *res, MonoType *type, gboolean include_namespace)
 		g_string_append_c (res, '*');
 		break;
 	case MONO_TYPE_ARRAY:
-		mono_type_get_desc (res, &type->data.array->eklass->byval_arg, include_namespace);
+		mono_type_get_desc (res, &type->data.array->eklass->_byval_arg, include_namespace);
 		g_string_append_c (res, '[');
 		for (i = 1; i < type->data.array->rank; ++i)
 			g_string_append_c (res, ',');
 		g_string_append_c (res, ']');
 		break;
 	case MONO_TYPE_SZARRAY:
-		mono_type_get_desc (res, &type->data.klass->byval_arg, include_namespace);
+		mono_type_get_desc (res, &type->data.klass->_byval_arg, include_namespace);
 		g_string_append (res, "[]");
 		break;
 	case MONO_TYPE_CLASS:
@@ -179,7 +192,7 @@ mono_type_get_desc (GString *res, MonoType *type, gboolean include_namespace)
 	case MONO_TYPE_GENERICINST: {
 		MonoGenericContext *context;
 
-		mono_type_get_desc (res, &type->data.generic_class->container_class->byval_arg, include_namespace);
+		mono_type_get_desc (res, &type->data.generic_class->container_class->_byval_arg, include_namespace);
 		g_string_append (res, "<");
 		context = &type->data.generic_class->context;
 		if (context->class_inst) {
@@ -223,6 +236,9 @@ mono_type_get_desc (GString *res, MonoType *type, gboolean include_namespace)
 		g_string_append_c (res, '&');
 }
 
+/**
+ * mono_type_full_name:
+ */
 char*
 mono_type_full_name (MonoType *type)
 {
@@ -233,6 +249,9 @@ mono_type_full_name (MonoType *type)
 	return g_string_free (str, FALSE);
 }
 
+/**
+ * mono_signature_get_desc:
+ */
 char*
 mono_signature_get_desc (MonoMethodSignature *sig, gboolean include_namespace)
 {
@@ -280,8 +299,8 @@ mono_signature_full_name (MonoMethodSignature *sig)
 	return result;
 }
 
-static void
-ginst_get_desc (GString *str, MonoGenericInst *ginst)
+void
+mono_ginst_get_desc (GString *str, MonoGenericInst *ginst)
 {
 	int i;
 
@@ -302,11 +321,11 @@ mono_context_get_desc (MonoGenericContext *context)
 	g_string_append (str, "<");
 
 	if (context->class_inst)
-		ginst_get_desc (str, context->class_inst);
+		mono_ginst_get_desc (str, context->class_inst);
 	if (context->method_inst) {
 		if (context->class_inst)
 			g_string_append (str, "; ");
-		ginst_get_desc (str, context->method_inst);
+		mono_ginst_get_desc (str, context->method_inst);
 	}
 
 	g_string_append (str, ">");
@@ -317,19 +336,19 @@ mono_context_get_desc (MonoGenericContext *context)
 
 /**
  * mono_method_desc_new:
- * @name: the method name.
- * @include_namespace: whether the name includes a namespace or not.
+ * \param name the method name.
+ * \param include_namespace whether the name includes a namespace or not.
  *
- * Creates a method description for @name, which conforms to the following
+ * Creates a method description for \p name, which conforms to the following
  * specification:
  *
- * [namespace.]classname:methodname[(args...)]
+ * <code>[namespace.]classname:methodname[(args...)]</code>
  *
  * in all the loaded assemblies.
  *
- * Both classname and methodname can contain '*' which matches anything.
+ * Both classname and methodname can contain <code>*</code> which matches anything.
  *
- * Returns: a parsed representation of the method description.
+ * \returns a parsed representation of the method description.
  */
 MonoMethodDesc*
 mono_method_desc_new (const char *name, gboolean include_namespace)
@@ -400,6 +419,9 @@ mono_method_desc_new (const char *name, gboolean include_namespace)
 	return result;
 }
 
+/**
+ * mono_method_desc_from_method:
+ */
 MonoMethodDesc*
 mono_method_desc_from_method (MonoMethod *method)
 {
@@ -416,9 +438,8 @@ mono_method_desc_from_method (MonoMethod *method)
 
 /**
  * mono_method_desc_free:
- * @desc: method description to be released
- *
- * Releases the MonoMethodDesc object @desc.
+ * \param desc method description to be released
+ * Releases the \c MonoMethodDesc object \p desc.
  */
 void
 mono_method_desc_free (MonoMethodDesc *desc)
@@ -431,14 +452,14 @@ mono_method_desc_free (MonoMethodDesc *desc)
 }
 
 /**
- * mono_method_descr_match:
- * @desc: MonoMethoDescription
- * @method: MonoMethod to test
+ * mono_method_desc_match:
+ * \param desc \c MonoMethoDescription
+ * \param method \c MonoMethod to test
  *
- * Determines whether the specified @method matches the provided @desc description.
+ * Determines whether the specified \p method matches the provided \p desc description.
  *
  * namespace and class are supposed to match already if this function is used.
- * Returns: True if the method matches the description, false otherwise.
+ * \returns TRUE if the method matches the description, FALSE otherwise.
  */
 gboolean
 mono_method_desc_match (MonoMethodDesc *desc, MonoMethod *method)
@@ -482,6 +503,7 @@ static gboolean
 match_class (MonoMethodDesc *desc, int pos, MonoClass *klass)
 {
 	const char *p;
+	gboolean is_terminal = TRUE;
 
 	if (desc->klass_glob && !strcmp (desc->klass, "*"))
 		return TRUE;
@@ -489,9 +511,14 @@ match_class (MonoMethodDesc *desc, int pos, MonoClass *klass)
 	if (desc->klass_glob && g_pattern_match_simple (desc->klass, klass->name))
 		return TRUE;
 #endif
+	if (desc->klass[pos] == '/')
+		is_terminal = FALSE;
+
 	p = my_strrchr (desc->klass, '/', &pos);
 	if (!p) {
-		if (strncmp (desc->klass, klass->name, pos))
+		if (is_terminal && strcmp (desc->klass, klass->name))
+			return FALSE;
+		if (!is_terminal && strncmp (desc->klass, klass->name, pos))
 			return FALSE;
 		if (desc->name_space && strcmp (desc->name_space, klass->name_space))
 			return FALSE;
@@ -506,9 +533,30 @@ match_class (MonoMethodDesc *desc, int pos, MonoClass *klass)
 	return match_class (desc, pos, klass->nested_in);
 }
 
+/**
+ * mono_method_desc_is_full:
+ */
+gboolean
+mono_method_desc_is_full (MonoMethodDesc *desc)
+{
+	return desc->klass && desc->klass[0] != '\0';
+}
+
+/**
+ * mono_method_desc_full_match:
+ * \param desc A method description that you created with mono_method_desc_new
+ * \param method a MonoMethod instance that you want to match against
+ *
+ * This method is used to check whether the method matches the provided
+ * description, by making sure that the method matches both the class and the method parameters.
+ *
+ * \returns TRUE if the specified method matches the specified description, FALSE otherwise.
+ */
 gboolean
 mono_method_desc_full_match (MonoMethodDesc *desc, MonoMethod *method)
 {
+	if (!desc)
+		return FALSE;
 	if (!desc->klass)
 		return FALSE;
 	if (!match_class (desc, strlen (desc->klass), method->klass))
@@ -517,6 +565,9 @@ mono_method_desc_full_match (MonoMethodDesc *desc, MonoMethod *method)
 	return mono_method_desc_match (desc, method);
 }
 
+/**
+ * mono_method_desc_search_in_class:
+ */
 MonoMethod*
 mono_method_desc_search_in_class (MonoMethodDesc *desc, MonoClass *klass)
 {
@@ -529,6 +580,9 @@ mono_method_desc_search_in_class (MonoMethodDesc *desc, MonoClass *klass)
 	return NULL;
 }
 
+/**
+ * mono_method_desc_search_in_image:
+ */
 MonoMethod*
 mono_method_desc_search_in_image (MonoMethodDesc *desc, MonoImage *image)
 {
@@ -555,15 +609,15 @@ mono_method_desc_search_in_image (MonoMethodDesc *desc, MonoImage *image)
 	mono_image_get_table_info (image, MONO_TABLE_TYPEDEF);
 	methods = mono_image_get_table_info (image, MONO_TABLE_METHOD);
 	for (i = 0; i < mono_table_info_get_rows (methods); ++i) {
-		MonoError error;
+		ERROR_DECL (error);
 		guint32 token = mono_metadata_decode_row_col (methods, i, MONO_METHOD_NAME);
 		const char *n = mono_metadata_string_heap (image, token);
 
 		if (strcmp (n, desc->name))
 			continue;
-		method = mono_get_method_checked (image, MONO_TOKEN_METHOD_DEF | (i + 1), NULL, NULL, &error);
+		method = mono_get_method_checked (image, MONO_TOKEN_METHOD_DEF | (i + 1), NULL, NULL, error);
 		if (!method) {
-			mono_error_cleanup (&error);
+			mono_error_cleanup (error);
 			continue;
 		}
 		if (mono_method_desc_full_match (desc, method))
@@ -575,8 +629,8 @@ mono_method_desc_search_in_image (MonoMethodDesc *desc, MonoImage *image)
 static const unsigned char*
 dis_one (GString *str, MonoDisHelper *dh, MonoMethod *method, const unsigned char *ip, const unsigned char *end)
 {
-	MonoError error;
-	MonoMethodHeader *header = mono_method_get_header_checked (method, &error);
+	ERROR_DECL (error);
+	MonoMethodHeader *header = mono_method_get_header_checked (method, error);
 	const MonoOpcode *opcode;
 	guint32 label, token;
 	gint32 sval;
@@ -585,8 +639,8 @@ dis_one (GString *str, MonoDisHelper *dh, MonoMethod *method, const unsigned cha
 	const unsigned char* il_code;
 
 	if (!header) {
-		g_string_append_printf (str, "could not disassemble, bad header due to %s", mono_error_get_message (&error));
-		mono_error_cleanup (&error);
+		g_string_append_printf (str, "could not disassemble, bad header due to %s", mono_error_get_message (error));
+		mono_error_cleanup (error);
 		return end;
 	}
 	il_code = mono_method_header_get_code (header, NULL, NULL);
@@ -754,6 +808,9 @@ default_dh = {
 	NULL  /* user data */
 };
 
+/**
+ * mono_disasm_code_one:
+ */
 char*
 mono_disasm_code_one (MonoDisHelper *dh, MonoMethod *method, const guchar *ip, const guchar **endp)
 {
@@ -772,6 +829,9 @@ mono_disasm_code_one (MonoDisHelper *dh, MonoMethod *method, const guchar *ip, c
 	return result;
 }
 
+/**
+ * mono_disasm_code:
+ */
 char*
 mono_disasm_code (MonoDisHelper *dh, MonoMethod *method, const guchar *ip, const guchar* end)
 {
@@ -791,9 +851,8 @@ mono_disasm_code (MonoDisHelper *dh, MonoMethod *method, const guchar *ip, const
 
 /**
  * mono_field_full_name:
- * @field: field to retrieve information for
- *
- * Returns: the full name for the field, made up of the namespace, type name and the field name.
+ * \param field field to retrieve information for
+ * \returns the full name for the field, made up of the namespace, type name and the field name.
  */
 char *
 mono_field_full_name (MonoClassField *field)
@@ -814,11 +873,19 @@ mono_method_get_name_full (MonoMethod *method, gboolean signature, gboolean ret,
 	char wrapper [64];
 	char *klass_desc;
 	char *inst_desc = NULL;
+	ERROR_DECL (error);
+
+	const char *class_method_separator = ":";
+	const char *method_sig_space = " ";
+	if (format == MONO_TYPE_NAME_FORMAT_REFLECTION) {
+		class_method_separator = ".";
+		method_sig_space = "";
+	}
 
 	if (format == MONO_TYPE_NAME_FORMAT_IL)
-		klass_desc = mono_type_full_name (&method->klass->byval_arg);
+		klass_desc = mono_type_full_name (&method->klass->_byval_arg);
 	else
-		klass_desc = mono_type_get_name_full (&method->klass->byval_arg, format);
+		klass_desc = mono_type_get_name_full (&method->klass->_byval_arg, format);
 
 	if (method->is_inflated && ((MonoMethodInflated*)method)->context.method_inst) {
 		GString *str = g_string_new ("");
@@ -826,7 +893,7 @@ mono_method_get_name_full (MonoMethod *method, gboolean signature, gboolean ret,
 			g_string_append (str, "<");
 		else
 			g_string_append (str, "[");
-		ginst_get_desc (str, ((MonoMethodInflated*)method)->context.method_inst);
+		mono_ginst_get_desc (str, ((MonoMethodInflated*)method)->context.method_inst);
 		if (format == MONO_TYPE_NAME_FORMAT_IL)
 			g_string_append_c (str, '>');
 		else
@@ -842,7 +909,7 @@ mono_method_get_name_full (MonoMethod *method, gboolean signature, gboolean ret,
 			g_string_append (str, "<");
 		else
 			g_string_append (str, "[");
-		ginst_get_desc (str, container->context.method_inst);
+		mono_ginst_get_desc (str, container->context.method_inst);
 		if (format == MONO_TYPE_NAME_FORMAT_IL)
 			g_string_append_c (str, '>');
 		else
@@ -853,29 +920,40 @@ mono_method_get_name_full (MonoMethod *method, gboolean signature, gboolean ret,
 	}
 
 	if (method->wrapper_type != MONO_WRAPPER_NONE)
-		sprintf (wrapper, "(wrapper %s) ", wrapper_type_to_str (method->wrapper_type));
+		sprintf (wrapper, "(wrapper %s) ", mono_wrapper_type_to_str (method->wrapper_type));
 	else
 		strcpy (wrapper, "");
 
 	if (signature) {
-		char *tmpsig = mono_signature_get_desc (mono_method_signature (method), TRUE);
+		MonoMethodSignature *sig = mono_method_signature_checked (method, error);
+		char *tmpsig;
+
+		if (!is_ok (error)) {
+			tmpsig = g_strdup_printf ("<unable to load signature>");
+			mono_error_cleanup (error);
+		} else {
+			tmpsig = mono_signature_get_desc (sig, TRUE);
+		}
 
 		if (method->wrapper_type != MONO_WRAPPER_NONE)
-			sprintf (wrapper, "(wrapper %s) ", wrapper_type_to_str (method->wrapper_type));
+			sprintf (wrapper, "(wrapper %s) ", mono_wrapper_type_to_str (method->wrapper_type));
 		else
 			strcpy (wrapper, "");
-		if (ret) {
-			char *ret_str = mono_type_full_name (mono_method_signature (method)->ret);
-			res = g_strdup_printf ("%s%s %s:%s%s (%s)", wrapper, ret_str, klass_desc,
-								   method->name, inst_desc ? inst_desc : "", tmpsig);
+		if (ret && sig) {
+			char *ret_str = mono_type_full_name (sig->ret);
+			res = g_strdup_printf ("%s%s %s%s%s%s%s(%s)", wrapper, ret_str, klass_desc,
+								   class_method_separator,
+								   method->name, inst_desc ? inst_desc : "", method_sig_space, tmpsig);
 			g_free (ret_str);
 		} else {
-			res = g_strdup_printf ("%s%s:%s%s (%s)", wrapper, klass_desc,
-								   method->name, inst_desc ? inst_desc : "", tmpsig);
+			res = g_strdup_printf ("%s%s%s%s%s%s(%s)", wrapper, klass_desc,
+								   class_method_separator,
+								   method->name, inst_desc ? inst_desc : "", method_sig_space, tmpsig);
 		}
 		g_free (tmpsig);
 	} else {
-		res = g_strdup_printf ("%s%s:%s%s", wrapper, klass_desc,
+		res = g_strdup_printf ("%s%s%s%s%s", wrapper, klass_desc,
+							   class_method_separator,
 							   method->name, inst_desc ? inst_desc : "");
 	}
 
@@ -885,10 +963,17 @@ mono_method_get_name_full (MonoMethod *method, gboolean signature, gboolean ret,
 	return res;
 }
 
+/**
+ * mono_method_full_name:
+ */
 char *
 mono_method_full_name (MonoMethod *method, gboolean signature)
 {
-	return mono_method_get_name_full (method, signature, FALSE, MONO_TYPE_NAME_FORMAT_IL);
+	char *res;
+	MONO_ENTER_GC_UNSAFE;
+	res = mono_method_get_name_full (method, signature, FALSE, MONO_TYPE_NAME_FORMAT_IL);
+	MONO_EXIT_GC_UNSAFE;
+	return res;
 }
 
 char *
@@ -897,6 +982,17 @@ mono_method_get_full_name (MonoMethod *method)
 	return mono_method_get_name_full (method, TRUE, TRUE, MONO_TYPE_NAME_FORMAT_IL);
 }
 
+/**
+ * mono_method_get_reflection_name:
+ *
+ * Returns the name of the method, including signature, using the same formating as reflection.
+ */
+char *
+mono_method_get_reflection_name (MonoMethod *method)
+{
+	return mono_method_get_name_full (method, TRUE, FALSE, MONO_TYPE_NAME_FORMAT_REFLECTION);
+}
+
 static const char*
 print_name_space (MonoClass *klass)
 {
@@ -915,13 +1011,13 @@ print_name_space (MonoClass *klass)
 /**
  * mono_object_describe:
  *
- * Prints to stdout a small description of the object @obj.
+ * Prints to stdout a small description of the object \p obj.
  * For use in a debugger.
  */
 void
 mono_object_describe (MonoObject *obj)
 {
-	MonoError error;
+	ERROR_DECL (error);
 	MonoClass* klass;
 	const char* sep;
 	if (!obj) {
@@ -930,8 +1026,8 @@ mono_object_describe (MonoObject *obj)
 	}
 	klass = mono_object_class (obj);
 	if (klass == mono_defaults.string_class) {
-		char *utf8 = mono_string_to_utf8_checked ((MonoString*)obj, &error);
-		mono_error_cleanup (&error); /* FIXME don't swallow the error */
+		char *utf8 = mono_string_to_utf8_checked ((MonoString*)obj, error);
+		mono_error_cleanup (error); /* FIXME don't swallow the error */
 		if (utf8 && strlen (utf8) > 60) {
 			utf8 [57] = '.';
 			utf8 [58] = '.';
@@ -1041,12 +1137,12 @@ objval_describe (MonoClass *klass, const char *addr)
 	gssize type_offset = 0;
 
 	if (klass->valuetype)
-		type_offset = -sizeof (MonoObject);
+		type_offset = - MONO_ABI_SIZEOF (MonoObject);
 
 	for (p = klass; p != NULL; p = p->parent) {
 		gpointer iter = NULL;
 		int printed_header = FALSE;
-		while ((field = mono_class_get_fields (p, &iter))) {
+		while ((field = mono_class_get_fields_internal (p, &iter))) {
 			if (field->type->attrs & (FIELD_ATTRIBUTE_STATIC | FIELD_ATTRIBUTE_HAS_FIELD_RVA))
 				continue;
 
@@ -1067,7 +1163,7 @@ objval_describe (MonoClass *klass, const char *addr)
 /**
  * mono_object_describe_fields:
  *
- * Prints to stdout a small description of each field of the object @obj.
+ * Prints to stdout a small description of each field of the object \p obj.
  * For use in a debugger.
  */
 void
@@ -1081,7 +1177,7 @@ mono_object_describe_fields (MonoObject *obj)
  * mono_value_describe_fields:
  *
  * Prints to stdout a small description of each field of the value type
- * stored at @addr of type @klass.
+ * stored at \p addr of type \p klass.
  * For use in a debugger.
  */
 void
@@ -1093,22 +1189,22 @@ mono_value_describe_fields (MonoClass* klass, const char* addr)
 /**
  * mono_class_describe_statics:
  *
- * Prints to stdout a small description of each static field of the type @klass
+ * Prints to stdout a small description of each static field of the type \p klass
  * in the current application domain.
  * For use in a debugger.
  */
 void
 mono_class_describe_statics (MonoClass* klass)
 {
-	MonoError error;
+	ERROR_DECL (error);
 	MonoClassField *field;
 	MonoClass *p;
 	const char *field_ptr;
-	MonoVTable *vtable = mono_class_vtable_full (mono_domain_get (), klass, &error);
+	MonoVTable *vtable = mono_class_vtable_checked (mono_domain_get (), klass, error);
 	const char *addr;
 
-	if (!vtable || !is_ok (&error)) {
-		mono_error_cleanup (&error);
+	if (!vtable || !is_ok (error)) {
+		mono_error_cleanup (error);
 		return;
 	}
 
@@ -1117,7 +1213,7 @@ mono_class_describe_statics (MonoClass* klass)
 
 	for (p = klass; p != NULL; p = p->parent) {
 		gpointer iter = NULL;
-		while ((field = mono_class_get_fields (p, &iter))) {
+		while ((field = mono_class_get_fields_internal (p, &iter))) {
 			if (field->type->attrs & FIELD_ATTRIBUTE_LITERAL)
 				continue;
 			if (!(field->type->attrs & (FIELD_ATTRIBUTE_STATIC | FIELD_ATTRIBUTE_HAS_FIELD_RVA)))
@@ -1132,7 +1228,7 @@ mono_class_describe_statics (MonoClass* klass)
 
 /**
  * mono_print_method_code
- * @MonoMethod: a pointer to the method
+ * \param method: a pointer to the method
  *
  * This method is used from a debugger to print the code of the method.
  *
@@ -1141,12 +1237,12 @@ mono_class_describe_statics (MonoClass* klass)
 void
 mono_method_print_code (MonoMethod *method)
 {
-	MonoError error;
+	ERROR_DECL (error);
 	char *code;
-	MonoMethodHeader *header = mono_method_get_header_checked (method, &error);
+	MonoMethodHeader *header = mono_method_get_header_checked (method, error);
 	if (!header) {
-		printf ("METHOD HEADER NOT FOUND DUE TO: %s\n", mono_error_get_message (&error));
-		mono_error_cleanup (&error);
+		printf ("METHOD HEADER NOT FOUND DUE TO: %s\n", mono_error_get_message (error));
+		mono_error_cleanup (error);
 		return;
 	}
 	code = mono_disasm_code (0, method, header->code, header->code + header->code_size);
