@@ -1,5 +1,6 @@
-/*
- * null-gc.c: GC implementation using malloc: will leak everything, just for testing.
+/**
+ * \file
+ * GC implementation using malloc: will leak everything, just for testing.
  *
  * Copyright 2001-2003 Ximian, Inc (http://www.ximian.com)
  * Copyright 2004-2011 Novell, Inc (http://www.novell.com)
@@ -12,17 +13,22 @@
 #include <mono/metadata/mono-gc.h>
 #include <mono/metadata/gc-internals.h>
 #include <mono/metadata/runtime.h>
+#include <mono/metadata/w32handle.h>
+#include <mono/metadata/abi-details.h>
 #include <mono/utils/atomic.h>
 #include <mono/utils/mono-threads.h>
 #include <mono/utils/mono-counters.h>
+#include <mono/metadata/null-gc-handles.h>
 
 #ifdef HAVE_NULL_GC
 
+static gboolean gc_inited = FALSE;
+
 void
 mono_gc_base_init (void)
 {
-	MonoThreadInfoCallbacks cb;
-	int dummy;
+	if (gc_inited)
+		return;
 
 	mono_counters_init ();
 
@@ -30,15 +36,14 @@ mono_gc_base_init (void)
 	mono_w32handle_init ();
 #endif
 
-	memset (&cb, 0, sizeof (cb));
-	/* TODO: This casts away an incompatible pointer type warning in the same
-	         manner that boehm-gc does it. This is probably worth investigating
-	         more carefully. */
-	cb.mono_method_is_critical = (gpointer)mono_runtime_is_critical_method;
+	mono_thread_callbacks_init ();
+	mono_thread_info_init (sizeof (MonoThreadInfo));
 
-	mono_threads_init (&cb, sizeof (MonoThreadInfo));
+	mono_thread_info_attach ();
 
-	mono_thread_info_attach (&dummy);
+	null_gc_handles_init ();
+
+	gc_inited = TRUE;
 }
 
 void
@@ -93,12 +98,6 @@ mono_gc_is_gc_thread (void)
 	return TRUE;
 }
 
-gboolean
-mono_gc_register_thread (void *baseptr)
-{
-	return TRUE;
-}
-
 int
 mono_gc_walk_heap (int flags, MonoGCReferences callback, void *data)
 {
@@ -112,34 +111,23 @@ mono_object_is_alive (MonoObject* o)
 }
 
 int
-mono_gc_register_root (char *start, size_t size, void *descr, MonoGCRootSource source, const char *msg)
+mono_gc_register_root (char *start, size_t size, void *descr, MonoGCRootSource source, void *key, const char *msg)
 {
 	return TRUE;
 }
 
+int
+mono_gc_register_root_wbarrier (char *start, size_t size, MonoGCDescriptor descr, MonoGCRootSource source, void *key, const char *msg)
+{
+	return TRUE;
+}
+
+
 void
 mono_gc_deregister_root (char* addr)
 {
 }
 
-void
-mono_gc_weak_link_add (void **link_addr, MonoObject *obj, gboolean track)
-{
-	*link_addr = obj;
-}
-
-void
-mono_gc_weak_link_remove (void **link_addr, gboolean track)
-{
-	*link_addr = NULL;
-}
-
-MonoObject*
-mono_gc_weak_link_get (void **link_addr)
-{
-	return *link_addr;
-}
-
 void*
 mono_gc_make_descr_for_string (gsize *bitmap, int numbits)
 {
@@ -165,15 +153,27 @@ mono_gc_make_descr_from_bitmap (gsize *bitmap, int numbits)
 }
 
 void*
-mono_gc_make_root_descr_all_refs (int numbits)
+mono_gc_make_vector_descr (void)
 {
 	return NULL;
 }
 
 void*
-mono_gc_alloc_fixed (size_t size, void *descr, MonoGCRootSource source, const char *msg)
+mono_gc_make_root_descr_all_refs (int numbits)
 {
-	return g_malloc0 (size);
+	return NULL;
+}
+
+MonoObject*
+mono_gc_alloc_fixed (size_t size, void *descr, MonoGCRootSource source, void *key, const char *msg)
+{
+	return (MonoObject*)g_malloc0 (size);
+}
+
+MonoObject*
+mono_gc_alloc_fixed_no_descriptor (size_t size, MonoGCRootSource source, void *key, const char *msg)
+{
+	return mono_gc_alloc_fixed (size, NULL, source, key, msg);
 }
 
 void
@@ -182,7 +182,7 @@ mono_gc_free_fixed (void* addr)
 	g_free (addr);
 }
 
-void *
+MonoObject*
 mono_gc_alloc_obj (MonoVTable *vtable, size_t size)
 {
 	MonoObject *obj = g_calloc (1, size);
@@ -192,7 +192,7 @@ mono_gc_alloc_obj (MonoVTable *vtable, size_t size)
 	return obj;
 }
 
-void *
+MonoArray*
 mono_gc_alloc_vector (MonoVTable *vtable, size_t size, uintptr_t max_length)
 {
 	MonoArray *obj = g_calloc (1, size);
@@ -203,7 +203,7 @@ mono_gc_alloc_vector (MonoVTable *vtable, size_t size, uintptr_t max_length)
 	return obj;
 }
 
-void *
+MonoArray*
 mono_gc_alloc_array (MonoVTable *vtable, size_t size, uintptr_t max_length, uintptr_t bounds_size)
 {
 	MonoArray *obj = g_calloc (1, size);
@@ -217,7 +217,7 @@ mono_gc_alloc_array (MonoVTable *vtable, size_t size, uintptr_t max_length, uint
 	return obj;
 }
 
-void *
+MonoString*
 mono_gc_alloc_string (MonoVTable *vtable, size_t size, gint32 len)
 {
 	MonoString *obj = g_calloc (1, size);
@@ -229,13 +229,13 @@ mono_gc_alloc_string (MonoVTable *vtable, size_t size, gint32 len)
 	return obj;
 }
 
-void*
+MonoObject*
 mono_gc_alloc_mature (MonoVTable *vtable, size_t size)
 {
 	return mono_gc_alloc_obj (vtable, size);
 }
 
-void*
+MonoObject*
 mono_gc_alloc_pinned_obj (MonoVTable *vtable, size_t size)
 {
 	return mono_gc_alloc_obj (vtable, size);
@@ -268,7 +268,7 @@ mono_gc_wbarrier_generic_store (gpointer ptr, MonoObject* value)
 void
 mono_gc_wbarrier_generic_store_atomic (gpointer ptr, MonoObject *value)
 {
-	InterlockedWritePointer (ptr, value);
+	mono_atomic_store_ptr (ptr, value);
 }
 
 void
@@ -286,8 +286,8 @@ void
 mono_gc_wbarrier_object_copy (MonoObject* obj, MonoObject *src)
 {
 	/* do not copy the sync state */
-	mono_gc_memmove_aligned ((char*)obj + sizeof (MonoObject), (char*)src + sizeof (MonoObject),
-			mono_object_class (obj)->instance_size - sizeof (MonoObject));
+	mono_gc_memmove_aligned (mono_object_get_data (obj), (char*)src + MONO_ABI_SIZEOF (MonoObject),
+			mono_object_class (obj)->instance_size - MONO_ABI_SIZEOF (MonoObject));
 }
 
 gboolean
@@ -296,6 +296,25 @@ mono_gc_is_critical_method (MonoMethod *method)
 	return FALSE;
 }
 
+gpointer
+mono_gc_thread_attach (MonoThreadInfo* info)
+{
+	info->handle_stack = mono_handle_stack_alloc ();
+	return info;
+}
+
+void
+mono_gc_thread_detach_with_lock (MonoThreadInfo *p)
+{
+	mono_handle_stack_free (p->handle_stack);
+}
+
+gboolean
+mono_gc_thread_in_critical_region (MonoThreadInfo *info)
+{
+	return FALSE;
+}
+
 int
 mono_gc_get_aligned_size_for_allocator (int size)
 {
@@ -332,27 +351,6 @@ mono_gc_get_gc_name (void)
 	return "null";
 }
 
-void
-mono_gc_add_weak_track_handle (MonoObject *obj, guint32 gchandle)
-{
-}
-
-void
-mono_gc_change_weak_track_handle (MonoObject *old_obj, MonoObject *obj, guint32 gchandle)
-{
-}
-
-void
-mono_gc_remove_weak_track_handle (guint32 gchandle)
-{
-}
-
-GSList*
-mono_gc_remove_weak_track_object (MonoDomain *domain, MonoObject *obj)
-{
-	return NULL;
-}
-
 void
 mono_gc_clear_domain (MonoDomain *domain)
 {
@@ -419,11 +417,17 @@ mono_gc_is_disabled (void)
 }
 
 void
-mono_gc_wbarrier_value_copy_bitmap (gpointer _dest, gpointer _src, int size, unsigned bitmap)
+mono_gc_wbarrier_range_copy (gpointer _dest, gconstpointer _src, int size)
 {
 	g_assert_not_reached ();
 }
 
+MonoRangeCopyFunction
+mono_gc_get_range_copy_func (void)
+{
+	return &mono_gc_wbarrier_range_copy;
+}
+
 guint8*
 mono_gc_get_card_table (int *shift_bits, gpointer *card_mask)
 {
@@ -431,6 +435,14 @@ mono_gc_get_card_table (int *shift_bits, gpointer *card_mask)
 	return NULL;
 }
 
+guint8*
+mono_gc_get_target_card_table (int *shift_bits, target_mgreg_t *card_mask)
+{
+	*shift_bits = 0;
+	*card_mask = 0;
+	return NULL;
+}
+
 gboolean
 mono_gc_card_table_nursery_check (void)
 {
@@ -444,11 +456,6 @@ mono_gc_get_nursery (int *shift_bits, size_t *size)
 	return NULL;
 }
 
-void
-mono_gc_set_current_thread_appdomain (MonoDomain *domain)
-{
-}
-
 gboolean
 mono_gc_precise_stack_mark_enabled (void)
 {
@@ -461,6 +468,16 @@ mono_gc_get_logfile (void)
 	return NULL;
 }
 
+void
+mono_gc_params_set (const char* options)
+{
+}
+
+void
+mono_gc_debug_set (const char* options)
+{
+}
+
 void
 mono_gc_conservatively_scan_area (void *start, void *end)
 {
@@ -518,7 +535,14 @@ mono_gc_pthread_create (pthread_t *new_thread, const pthread_attr_t *attr, void
 }
 #endif
 
-void mono_gc_set_skip_thread (gboolean value)
+void
+mono_gc_skip_thread_changing (gboolean skip)
+{
+	// No STW, nothing needs to be done.
+}
+
+void
+mono_gc_skip_thread_changed (gboolean skip)
 {
 }
 
@@ -529,6 +553,13 @@ BOOL APIENTRY mono_gc_dllmain (HMODULE module_handle, DWORD reason, LPVOID reser
 }
 #endif
 
+MonoVTable *
+mono_gc_get_vtable (MonoObject *obj)
+{
+	// No pointer tagging.
+	return obj->vtable;
+}
+
 guint
 mono_gc_get_vtable_bits (MonoClass *klass)
 {
@@ -545,9 +576,26 @@ mono_gc_is_null (void)
 {
 	return TRUE;
 }
+
+int
+mono_gc_invoke_finalizers (void)
+{
+	return 0;
+}
+
+MonoBoolean
+mono_gc_pending_finalizers (void)
+{
+	return FALSE;
+}
+
+gboolean
+mono_gc_ephemeron_array_add (MonoObject *obj)
+{
+	return TRUE;
+}
+
 #else
-	#ifdef _MSC_VER
-		// Quiet Visual Studio linker warning, LNK4221, in cases when this source file intentional ends up empty.
-		void __mono_win32_null_gc_quiet_lnk4221(void) {}
-	#endif
+
+MONO_EMPTY_SOURCE_FILE (null_gc);
 #endif /* HAVE_NULL_GC */
