@@ -1,5 +1,6 @@
-/*
- * sgen-bridge.c: Simple generational GC.
+/**
+ * \file
+ * Simple generational GC.
  *
  * Copyright 2011 Novell, Inc (http://www.novell.com)
  * Copyright 2011 Xamarin Inc (http://www.xamarin.com)
@@ -21,6 +22,8 @@
 #include "sgen/sgen-qsort.h"
 #include "utils/mono-logger-internals.h"
 
+#ifndef DISABLE_SGEN_GC_BRIDGE
+
 typedef enum {
 	BRIDGE_PROCESSOR_INVALID,
 	BRIDGE_PROCESSOR_OLD,
@@ -36,20 +39,23 @@ static MonoGCBridgeCallbacks pending_bridge_callbacks;
 // Configuration to be passed to bridge processor after init
 static SgenBridgeProcessorConfig bridge_processor_config;
 // Currently-in-use callbacks
-MonoGCBridgeCallbacks bridge_callbacks;
+MonoGCBridgeCallbacks mono_bridge_callbacks;
 
 // Bridge processor state
 static SgenBridgeProcessor bridge_processor;
 // This is used for a special debug feature
 static SgenBridgeProcessor compare_to_bridge_processor;
 
-volatile gboolean bridge_processing_in_progress = FALSE;
+volatile gboolean mono_bridge_processing_in_progress = FALSE;
 
 // FIXME: The current usage pattern for this function is unsafe. Bridge processing could start immediately after unlock
+/**
+ * mono_gc_wait_for_bridge_processing:
+ */
 void
 mono_gc_wait_for_bridge_processing (void)
 {
-	if (!bridge_processing_in_progress)
+	if (!mono_bridge_processing_in_progress)
 		return;
 
 	mono_trace (G_LOG_LEVEL_INFO, MONO_TRACE_GC, "GC_BRIDGE waiting for bridge processing to finish");
@@ -58,6 +64,9 @@ mono_gc_wait_for_bridge_processing (void)
 	sgen_gc_unlock ();
 }
 
+/**
+ * mono_gc_register_bridge_callbacks:
+ */
 void
 mono_gc_register_bridge_callbacks (MonoGCBridgeCallbacks *callbacks)
 {
@@ -87,7 +96,7 @@ bridge_processor_name (const char *name)
 }
 
 static gboolean
-bridge_processor_started ()
+bridge_processor_started (void)
 {
 	return bridge_processor.reset_data != NULL;
 }
@@ -132,16 +141,16 @@ init_bridge_processor (SgenBridgeProcessor *processor, BridgeProcessorSelection
  * is done initing. Actual initialization then only occurs if it is ready.
  */
 void
-sgen_init_bridge ()
+sgen_init_bridge (void)
 {
 	if (sgen_gc_initialized ()) {
 		// This lock is not initialized until the GC is
 		sgen_gc_lock ();
 
-		bridge_callbacks = pending_bridge_callbacks;
+		mono_bridge_callbacks = pending_bridge_callbacks;
 
 		// If a bridge was registered but there is no bridge processor yet
-		if (bridge_callbacks.cross_references && !bridge_processor_started ()) {
+		if (mono_bridge_callbacks.cross_references && !bridge_processor_started ()) {
 			init_bridge_processor (&bridge_processor, bridge_processor_selection);
 
 			if (bridge_processor.set_config)
@@ -174,13 +183,13 @@ sgen_is_bridge_object (GCObject *obj)
 {
 	if ((obj->vtable->gc_bits & SGEN_GC_BIT_BRIDGE_OBJECT) != SGEN_GC_BIT_BRIDGE_OBJECT)
 		return FALSE;
-	return bridge_callbacks.is_bridge_object (obj);
+	return mono_bridge_callbacks.is_bridge_object (obj);
 }
 
 gboolean
 sgen_need_bridge_processing (void)
 {
-	return bridge_callbacks.cross_references != NULL;
+	return mono_bridge_callbacks.cross_references != NULL;
 }
 
 static gboolean
@@ -202,10 +211,10 @@ void
 sgen_bridge_processing_stw_step (void)
 {
 	/*
-	 * bridge_processing_in_progress must be set with the world
+	 * mono_bridge_processing_in_progress must be set with the world
 	 * stopped.  If not there would be race conditions.
 	 */
-	bridge_processing_in_progress = TRUE;
+	mono_bridge_processing_in_progress = TRUE;
 
 	bridge_processor.processing_stw_step ();
 	if (compare_bridge_processors ())
@@ -450,7 +459,7 @@ sgen_bridge_processing_finish (int generation)
 		goto after_callback;
 	}
 
-	bridge_callbacks.cross_references (bridge_processor.num_sccs, bridge_processor.api_sccs,
+	mono_bridge_callbacks.cross_references (bridge_processor.num_sccs, bridge_processor.api_sccs,
 			bridge_processor.num_xrefs, bridge_processor.api_xrefs);
 
 	if (compare_bridge_processors ())
@@ -469,7 +478,7 @@ sgen_bridge_processing_finish (int generation)
 
 	mono_trace (G_LOG_LEVEL_INFO, MONO_TRACE_GC, "GC_BRIDGE: Complete, was running for %.2fms", mono_time_since_last_stw () / 10000.0f);
 
-	bridge_processing_in_progress = FALSE;
+	mono_bridge_processing_in_progress = FALSE;
 }
 
 MonoGCBridgeObjectKind
@@ -507,7 +516,7 @@ static const char *bridge_class;
 static MonoGCBridgeObjectKind
 bridge_test_bridge_class_kind (MonoClass *klass)
 {
-	if (!strcmp (bridge_class, klass->name))
+	if (!strcmp (bridge_class, m_class_get_name (klass)))
 		return GC_BRIDGE_TRANSPARENT_BRIDGE_CLASS;
 	return GC_BRIDGE_TRANSPARENT_CLASS;
 }
@@ -574,7 +583,7 @@ bridge_test_cross_reference2 (int num_sccs, MonoGCBridgeSCC **sccs, int num_xref
 	gboolean modified;
 
 	if (!mono_bridge_test_field) {
-		mono_bridge_test_field = mono_class_get_field_from_name (mono_object_get_class (sccs[0]->objs [0]), "__test");
+		mono_bridge_test_field = mono_class_get_field_from_name_full (mono_object_class (sccs[0]->objs [0]), "__test", NULL);
 		g_assert (mono_bridge_test_field);
 	}
 
@@ -624,7 +633,7 @@ bridge_test_positive_status (int num_sccs, MonoGCBridgeSCC **sccs, int num_xrefs
 	int i;
 
 	if (!mono_bridge_test_field) {
-		mono_bridge_test_field = mono_class_get_field_from_name (mono_object_get_class (sccs[0]->objs [0]), "__test");
+		mono_bridge_test_field = mono_class_get_field_from_name_full (mono_object_class (sccs[0]->objs [0]), "__test", NULL);
 		g_assert (mono_bridge_test_field);
 	}
 
@@ -690,7 +699,7 @@ sgen_bridge_handle_gc_debug (const char *opt)
 	} else if (!strcmp (opt, "enable-bridge-accounting")) {
 		bridge_processor_config.accounting = TRUE;
 	} else if (g_str_has_prefix (opt, "bridge-dump=")) {
-		char *prefix = strchr (opt, '=') + 1;
+		const char* prefix = strchr (opt, '=') + 1;
 		set_dump_prefix(prefix);
 	} else if (g_str_has_prefix (opt, "bridge-compare-to=")) {
 		const char *name = strchr (opt, '=') + 1;
@@ -717,4 +726,87 @@ sgen_bridge_print_gc_debug_usage (void)
 	fprintf (stderr, "  bridge-compare-to=<implementation>\n");
 }
 
+#else
+
+//UG
+volatile gboolean mono_bridge_processing_in_progress = FALSE;
+
+void
+mono_gc_wait_for_bridge_processing (void)
+{
+}
+
+MonoGCBridgeObjectKind
+sgen_bridge_class_kind (MonoClass *klass)
+{
+	return GC_BRIDGE_TRANSPARENT_CLASS;
+}
+
+void
+sgen_bridge_describe_pointer (GCObject *obj)
+{
+}
+
+gboolean
+sgen_bridge_handle_gc_debug (const char *opt)
+{
+	return FALSE;
+}
+
+gboolean
+sgen_bridge_handle_gc_param (const char *opt)
+{
+	return FALSE;
+}
+
+void
+sgen_bridge_print_gc_debug_usage (void)
+{
+}
+
+void
+sgen_bridge_processing_finish (int generation)
+{
+}
+
+void
+sgen_bridge_processing_stw_step (void)
+{
+}
+
+void
+sgen_bridge_register_finalized_object (GCObject *obj)
+{
+}
+
+void
+sgen_bridge_reset_data (void)
+{
+}
+
+void
+sgen_init_bridge (void)
+{
+}
+
+gboolean
+sgen_is_bridge_object (GCObject *obj)
+{
+	return FALSE;
+}
+
+gboolean
+sgen_need_bridge_processing (void)
+{
+	return FALSE;
+}
+
+void
+sgen_set_bridge_implementation (const char *name)
+{
+	g_error ("Sgen bridge disabled");
+}
+
+#endif
+
 #endif
