@@ -1,5 +1,6 @@
-/*
- * cfold.c: Constant folding support
+/**
+ * \file
+ * Constant folding support
  *
  * Author:
  *   Paolo Molaro (lupus@ximian.com)
@@ -7,9 +8,12 @@
  *
  * (C) 2003 Ximian, Inc.  http://www.ximian.com
  */
+#include <config.h>
+
 #include "mini.h"
 #include "ir-emit.h"
 
+/* WTF is this doing here?!?!? */
 int
 mono_is_power_of_two (guint32 val)
 {
@@ -66,6 +70,8 @@ mono_is_power_of_two (guint32 val)
     } \
 } while (0)
 
+#ifndef DISABLE_JIT
+
 /**
  * mono_constant_fold_ins:
  *
@@ -215,6 +221,13 @@ mono_constant_fold_ins (MonoCompile *cfg, MonoInst *ins, MonoInst *arg1, MonoIns
 			MONO_INST_NULLIFY_SREGS (dest);
 		}
 		break;
+	case OP_SEXT_I4:
+		if (arg1->opcode == OP_ICONST && arg1->inst_c0 >= 0 && arg1->inst_c0 < (1 << 16) && overwrite) {
+			dest->opcode = OP_ICONST;
+			dest->sreg1 = -1;
+			dest->inst_c0 = arg1->inst_c0;
+		}
+		break;
 	case OP_MOVE:
 #if SIZEOF_REGISTER == 8
 		if ((arg1->opcode == OP_ICONST) || (arg1->opcode == OP_I8CONST)) {
@@ -330,6 +343,19 @@ mono_constant_fold_ins (MonoCompile *cfg, MonoInst *ins, MonoInst *arg1, MonoIns
 					dest->inst_c0 = res;
 				}
 				break;
+			case OP_COND_EXC_EQ:
+				res = arg1->inst_c0 == arg2->inst_c0;
+				if (!res) {
+					if (overwrite) {
+						NULLIFY_INS (ins);
+						NULLIFY_INS (next);
+					} else {
+						ALLOC_DEST (cfg, dest, ins);
+						dest->opcode = OP_ICONST;
+						dest->inst_c0 = res;
+					}
+				}
+				break;
 			case OP_NOP:
 			case OP_BR:
 				/* This happens when a conditional branch is eliminated */
@@ -343,6 +369,24 @@ mono_constant_fold_ins (MonoCompile *cfg, MonoInst *ins, MonoInst *arg1, MonoIns
 				return NULL;
 			}
 		}
+		if ((arg1->opcode == OP_PCONST) && (arg2->opcode == OP_PCONST) && ins->next) {
+			MonoInst *next = ins->next;
+
+			if (next->opcode == OP_LCEQ) {
+				gboolean res = arg1->inst_p0 == arg2->inst_p0;
+				if (overwrite) {
+					NULLIFY_INS (ins);
+					next->opcode = OP_ICONST;
+					next->inst_c0 = res;
+					MONO_INST_NULLIFY_SREGS (next);
+				} else {
+					ALLOC_DEST (cfg, dest, ins);
+					dest->opcode = OP_ICONST;
+					dest->inst_c0 = res;
+				}
+				break;
+			}
+		}
 		break;
 	}
 	case OP_FMOVE:
@@ -366,3 +410,6 @@ mono_constant_fold_ins (MonoCompile *cfg, MonoInst *ins, MonoInst *arg1, MonoIns
 		
     return dest;
 }	
+
+
+#endif /* DISABLE_JIT */
