@@ -1,14 +1,17 @@
-/*
- * liveness.c: liveness analysis
+/**
+ * \file
+ * liveness analysis
  *
  * Author:
  *   Dietmar Maurer (dietmar@ximian.com)
  *
  * (C) 2002 Ximian, Inc.
  * Copyright 2011 Xamarin, Inc (http://www.xamarin.com)
+ * Licensed under the MIT license. See LICENSE file in the project root for full license information.
  */
 
 #include <config.h>
+#include <mono/utils/mono-compiler.h>
 
 #ifndef DISABLE_JIT
 
@@ -34,6 +37,65 @@
 static void mono_analyze_liveness2 (MonoCompile *cfg);
 #endif
 
+
+#define INLINE_SIZE 16
+
+typedef struct {
+	int capacity;
+	union {
+		gpointer data [INLINE_SIZE];
+		GHashTable *hashtable;
+	};
+} MonoPtrSet;
+
+static void
+mono_ptrset_init (MonoPtrSet *set)
+{
+	set->capacity = 0;
+}
+
+static void
+mono_ptrset_destroy (MonoPtrSet *set)
+{
+	if (set->capacity > INLINE_SIZE)
+		g_hash_table_destroy (set->hashtable);
+}
+
+static void
+mono_ptrset_add (MonoPtrSet *set, gpointer val)
+{
+	//switch to hashtable
+	if (set->capacity == INLINE_SIZE) {
+		GHashTable *tmp = g_hash_table_new (NULL, NULL);
+		for (int i = 0; i < INLINE_SIZE; ++i)
+			g_hash_table_insert (tmp, set->data [i], set->data [i]);
+		set->hashtable = tmp;
+		++set->capacity;
+	}
+
+	if (set->capacity > INLINE_SIZE) {
+		g_hash_table_insert (set->hashtable, val, val);
+	} else {
+		set->data [set->capacity] = val;
+		++set->capacity;
+	}
+}
+
+static gboolean
+mono_ptrset_contains (MonoPtrSet *set, gpointer val)
+{
+	if (set->capacity <= INLINE_SIZE) {
+		for (int i = 0; i < set->capacity; ++i) {
+			if (set->data [i] == val)
+				return TRUE;
+		}
+		return FALSE;
+	}
+
+	return g_hash_table_lookup (set->hashtable, val) != NULL;
+}
+
+
 static void
 optimize_initlocals (MonoCompile *cfg);
 
@@ -75,12 +137,12 @@ mono_bitset_print (MonoBitSet *set)
 }
 
 static void
-visit_bb (MonoCompile *cfg, MonoBasicBlock *bb, GSList **visited)
+visit_bb (MonoCompile *cfg, MonoBasicBlock *bb, MonoPtrSet *visited)
 {
 	int i;
 	MonoInst *ins;
 
-	if (g_slist_find (*visited, bb))
+	if (mono_ptrset_contains (visited, bb))
 		return;
 
 	for (ins = bb->code; ins; ins = ins->next) {
@@ -103,8 +165,8 @@ visit_bb (MonoCompile *cfg, MonoBasicBlock *bb, GSList **visited)
 			cfg->varinfo [vi->idx]->flags |= MONO_INST_VOLATILE;
 			if (SIZEOF_REGISTER == 4 && (var->type == STACK_I8 || (var->type == STACK_R8 && COMPILE_SOFT_FLOAT (cfg)))) {
 				/* Make the component vregs volatile as well (#612206) */
-				get_vreg_to_inst (cfg, var->dreg + 1)->flags |= MONO_INST_VOLATILE;
-				get_vreg_to_inst (cfg, var->dreg + 2)->flags |= MONO_INST_VOLATILE;
+				get_vreg_to_inst (cfg, MONO_LVREG_LS (var->dreg))->flags |= MONO_INST_VOLATILE;
+				get_vreg_to_inst (cfg, MONO_LVREG_MS (var->dreg))->flags |= MONO_INST_VOLATILE;
 			}
 		}
 			
@@ -122,14 +184,14 @@ visit_bb (MonoCompile *cfg, MonoBasicBlock *bb, GSList **visited)
 				cfg->varinfo [vi->idx]->flags |= MONO_INST_VOLATILE;
 				if (SIZEOF_REGISTER == 4 && (var->type == STACK_I8 || (var->type == STACK_R8 && COMPILE_SOFT_FLOAT (cfg)))) {
 					/* Make the component vregs volatile as well (#612206) */
-					get_vreg_to_inst (cfg, var->dreg + 1)->flags |= MONO_INST_VOLATILE;
-					get_vreg_to_inst (cfg, var->dreg + 2)->flags |= MONO_INST_VOLATILE;
+					get_vreg_to_inst (cfg, MONO_LVREG_LS (var->dreg))->flags |= MONO_INST_VOLATILE;
+					get_vreg_to_inst (cfg, MONO_LVREG_MS (var->dreg))->flags |= MONO_INST_VOLATILE;
 				}
 			}
 		}
 	}
 
-	*visited = g_slist_append (*visited, bb);
+	mono_ptrset_add (visited, bb);
 
 	/* 
 	 * Need to visit all bblocks reachable from this one since they can be
@@ -144,7 +206,6 @@ void
 mono_liveness_handle_exception_clauses (MonoCompile *cfg)
 {
 	MonoBasicBlock *bb;
-	GSList *visited = NULL;
 	MonoMethodHeader *header = cfg->header;
 	MonoExceptionClause *clause, *clause2;
 	int i, j;
@@ -179,6 +240,8 @@ mono_liveness_handle_exception_clauses (MonoCompile *cfg)
 		}
 	}
 
+	MonoPtrSet visited;
+	mono_ptrset_init (&visited);
 	/*
 	 * Variables in exception handler register cannot be allocated to registers
 	 * so make them volatile. See bug #42136. This will not be neccessary when
@@ -200,7 +263,7 @@ mono_liveness_handle_exception_clauses (MonoCompile *cfg)
 
 		visit_bb (cfg, bb, &visited);
 	}
-	g_slist_free (visited);
+	mono_ptrset_destroy (&visited);
 }
 
 static inline void
@@ -229,8 +292,7 @@ analyze_liveness_bb (MonoCompile *cfg, MonoBasicBlock *bb)
 
 #ifdef DEBUG_LIVENESS
 		if (cfg->verbose_level > 1) {
-			printf ("\t");
-			mono_print_ins (ins);
+			mono_print_ins_index (1, ins);
 		}
 #endif
 
@@ -850,13 +912,12 @@ update_liveness2 (MonoCompile *cfg, MonoInst *ins, gboolean set_volatile, int in
 static void
 mono_analyze_liveness2 (MonoCompile *cfg)
 {
-	int bnum, idx, i, j, nins, max, max_vars, block_from, block_to, pos, reverse_len;
+	int bnum, idx, i, j, nins, max, max_vars, block_from, block_to, pos;
 	gint32 *last_use;
 	static guint32 disabled = -1;
-	MonoInst **reverse;
 
 	if (disabled == -1)
-		disabled = g_getenv ("DISABLED") != NULL;
+		disabled = g_hasenv ("DISABLED");
 
 	if (disabled)
 		return;
@@ -888,9 +949,6 @@ mono_analyze_liveness2 (MonoCompile *cfg)
 	max_vars = cfg->num_varinfo;
 	last_use = g_new0 (gint32, max_vars);
 
-	reverse_len = 1024;
-	reverse = (MonoInst **)mono_mempool_alloc (cfg->mempool, sizeof (MonoInst*) * reverse_len);
-
 	for (idx = 0; idx < max_vars; ++idx) {
 		MonoMethodVar *vi = MONO_VARINFO (cfg, idx);
 
@@ -938,25 +996,13 @@ mono_analyze_liveness2 (MonoCompile *cfg)
 		if (cfg->ret)
 			last_use [cfg->ret->inst_c0] = block_to;
 
-		for (nins = 0, pos = block_from, ins = bb->code; ins; ins = ins->next, ++nins, ++pos) {
-			if (nins >= reverse_len) {
-				int new_reverse_len = reverse_len * 2;
-				MonoInst **new_reverse = (MonoInst **)mono_mempool_alloc (cfg->mempool, sizeof (MonoInst*) * new_reverse_len);
-				memcpy (new_reverse, reverse, sizeof (MonoInst*) * reverse_len);
-				reverse = new_reverse;
-				reverse_len = new_reverse_len;
-			}
-
-			reverse [nins] = ins;
-		}
+		pos = block_from + 1;
+		MONO_BB_FOR_EACH_INS (bb, ins) pos++;
 
 		/* Process instructions backwards */
-		for (i = nins - 1; i >= 0; --i) {
-			MonoInst *ins = (MonoInst*)reverse [i];
-
- 			update_liveness2 (cfg, ins, FALSE, pos, last_use);
-
-			pos --;
+		MONO_BB_FOR_EACH_INS_REVERSE (bb, ins) {
+			update_liveness2 (cfg, ins, FALSE, pos, last_use);
+			pos--;
 		}
 
 		for (idx = 0; idx < max_vars; ++idx) {
@@ -996,8 +1042,6 @@ mono_analyze_liveness2 (MonoCompile *cfg)
 
 #endif
 
-#define ALIGN_TO(val,align) ((((guint64)val) + ((align) - 1)) & ~((align) - 1))
-
 static inline void
 update_liveness_gc (MonoCompile *cfg, MonoBasicBlock *bb, MonoInst *ins, gint32 *last_use, MonoMethodVar **vreg_to_varinfo, GSList **callsites)
 {
@@ -1157,4 +1201,8 @@ mono_analyze_liveness_gc (MonoCompile *cfg)
 	g_free (vreg_to_varinfo);
 }
 
-#endif /* DISABLE_JIT */
+#else /* !DISABLE_JIT */
+
+MONO_EMPTY_SOURCE_FILE (liveness);
+
+#endif /* !DISABLE_JIT */
