@@ -1,5 +1,6 @@
-/*
- * local-propagation.c: Local constant, copy and tree propagation.
+/**
+ * \file
+ * Local constant, copy and tree propagation.
  *
  * To make some sense of the tree mover, read mono/docs/tree-mover.txt
  *
@@ -14,6 +15,8 @@
  */
 
 #include <config.h>
+#include <mono/utils/mono-compiler.h>
+
 #ifndef DISABLE_JIT
 
 #include <string.h>
@@ -25,6 +28,7 @@
 #include <mono/metadata/debug-helpers.h>
 #include <mono/metadata/mempool.h>
 #include <mono/metadata/opcodes.h>
+#include <mono/utils/unlocked.h>
 #include "mini.h"
 #include "ir-emit.h"
 
@@ -171,6 +175,8 @@ mono_strength_reduction_division (MonoCompile *cfg, MonoInst *ins)
 				ins->inst_imm = power2;
 				break;
 			}
+			if (cfg->backend->disable_div_with_mul)
+				break;
 			allocated_vregs = TRUE;
 			/*
 			 * Replacement of unsigned division with multiplication,
@@ -206,7 +212,7 @@ mono_strength_reduction_division (MonoCompile *cfg, MonoInst *ins)
 			}
 			MONO_EMIT_NEW_UNALU (cfg, OP_MOVE, ins->dreg, MONO_LVREG_LS (tmp_regl));
 #endif
-			mono_jit_stats.optimized_divisions++;
+			UnlockedIncrement (&mono_jit_stats.optimized_divisions);
 			break;
 		}
 		case OP_IDIV_IMM: {
@@ -243,6 +249,8 @@ mono_strength_reduction_division (MonoCompile *cfg, MonoInst *ins)
 				break;
 			}
 
+			if (cfg->backend->disable_div_with_mul)
+				break;
 			/*
 			 * Replacement of signed division with multiplication,
 			 * shifts and additions Hacker's Delight, chapter 10-6.
@@ -286,7 +294,7 @@ mono_strength_reduction_division (MonoCompile *cfg, MonoInst *ins)
 			MONO_EMIT_NEW_BIALU_IMM (cfg, OP_ISHR_UN_IMM, ins->dreg, tmp_regi, SIZEOF_REGISTER * 8 - 1);
 			MONO_EMIT_NEW_BIALU (cfg, OP_IADD, ins->dreg, ins->dreg, tmp_regi);
 #endif
-			mono_jit_stats.optimized_divisions++;
+			UnlockedIncrement (&mono_jit_stats.optimized_divisions);
 			break;
 		}
 	}
@@ -362,7 +370,7 @@ mono_strength_reduction_ins (MonoCompile *cfg, MonoInst *ins, const char **spec)
 	}
 	case OP_IDIV_UN_IMM:
 	case OP_IDIV_IMM: {
-		if (!COMPILE_LLVM (cfg))
+		if ((!COMPILE_LLVM (cfg)) && (!cfg->backend->optimized_div))
 			allocated_vregs = mono_strength_reduction_division (cfg, ins);
 		break;
 	}
@@ -375,10 +383,8 @@ mono_strength_reduction_ins (MonoCompile *cfg, MonoInst *ins, const char **spec)
 			ins->opcode = OP_ICONST;
 			MONO_INST_NULLIFY_SREGS (ins);
 			ins->inst_c0 = 0;
-#if __s390__
-		}
-#else
-		} else if ((ins->inst_imm > 0) && (ins->inst_imm < (1LL << 32)) && (power != -1)) {
+		} else if ((ins->inst_imm > 0) && (ins->inst_imm < (1LL << 32)) && 
+			   (power != -1) && (!cfg->backend->optimized_div)) {
 			gboolean is_long = ins->opcode == OP_LREM_IMM;
 			int compensator_reg = alloc_ireg (cfg);
 			int intermediate_reg;
@@ -403,7 +409,6 @@ mono_strength_reduction_ins (MonoCompile *cfg, MonoInst *ins, const char **spec)
 
 			allocated_vregs = TRUE;
 		}
-#endif
 		break;
 	}
 #if SIZEOF_REGISTER == 4
@@ -632,16 +637,16 @@ mono_local_cprop (MonoCompile *cfg)
 				}
 
 				/* Constant propagation */
-				/* FIXME: Make is_inst_imm a macro */
-				/* FIXME: Make is_inst_imm take an opcode argument */
 				/* is_inst_imm is only needed for binops */
-				if ((((def->opcode == OP_ICONST) || ((sizeof (gpointer) == 8) && (def->opcode == OP_I8CONST))) &&
-					 (((srcindex == 0) && (ins->sreg2 == -1)) || mono_arch_is_inst_imm (def->inst_c0))) || 
+				if ((((def->opcode == OP_ICONST) || ((sizeof (gpointer) == 8) && (def->opcode == OP_I8CONST)) || (def->opcode == OP_PCONST)))
+					||
 					(!MONO_ARCH_USE_FPSTACK && (def->opcode == OP_R8CONST))) {
 					guint32 opcode2;
 
 					/* srcindex == 1 -> binop, ins->sreg2 == -1 -> unop */
-					if ((srcindex == 1) && (ins->sreg1 != -1) && defs [ins->sreg1] && (defs [ins->sreg1]->opcode == OP_ICONST) && defs [ins->sreg2]) {
+					if ((srcindex == 1) && (ins->sreg1 != -1) && defs [ins->sreg1] &&
+						((defs [ins->sreg1]->opcode == OP_ICONST) || defs [ins->sreg1]->opcode == OP_PCONST) &&
+						defs [ins->sreg2]) {
 						/* Both arguments are constants, perform cfold */
 						mono_constant_fold_ins (cfg, ins, defs [ins->sreg1], defs [ins->sreg2], TRUE);
 					} else if ((srcindex == 0) && (ins->sreg2 != -1) && defs [ins->sreg2]) {
@@ -659,14 +664,15 @@ mono_local_cprop (MonoCompile *cfg)
 					}
 
 					opcode2 = mono_op_to_op_imm (ins->opcode);
-					if ((opcode2 != -1) && mono_arch_is_inst_imm (def->inst_c0) && ((srcindex == 1) || (ins->sreg2 == -1))) {
+					if ((opcode2 != -1) && mono_arch_is_inst_imm (ins->opcode, opcode2, def->inst_c0) && ((srcindex == 1) || (ins->sreg2 == -1))) {
 						ins->opcode = opcode2;
-						if ((def->opcode == OP_I8CONST) && (sizeof (gpointer) == 4)) {
-							ins->inst_ls_word = def->inst_ls_word;
-							ins->inst_ms_word = def->inst_ms_word;
-						} else {
+						if ((def->opcode == OP_I8CONST) && TARGET_SIZEOF_VOID_P == 4)
+							ins->inst_l = def->inst_l;
+						else if (regtype == 'l' && TARGET_SIZEOF_VOID_P == 4)
+							/* This can happen if the def was a result of an iconst+conv.i8, which is transformed into just an iconst */
+							ins->inst_l = def->inst_c0;
+						else
 							ins->inst_imm = def->inst_c0;
-						}
 						sregs [srcindex] = -1;
 						mono_inst_set_src_registers (ins, sregs);
 
@@ -692,7 +698,7 @@ mono_local_cprop (MonoCompile *cfg)
 						}
 #endif
 						opcode2 = mono_load_membase_to_load_mem (ins->opcode);
-						if ((srcindex == 0) && (opcode2 != -1) && mono_arch_is_inst_imm (def->inst_c0)) {
+						if ((srcindex == 0) && (opcode2 != -1) && mono_arch_is_inst_imm (ins->opcode, opcode2, def->inst_c0)) {
 							ins->opcode = opcode2;
 							ins->inst_imm = def->inst_c0 + ins->inst_offset;
 							ins->sreg1 = -1;
@@ -733,6 +739,14 @@ mono_local_cprop (MonoCompile *cfg)
 					dummy_arg1.inst_c0 = 1;
 
 					mono_constant_fold_ins (cfg, ins, &dummy_arg1, NULL, TRUE);
+				} else if (srcindex == 0 && ins->opcode == OP_COMPARE && defs [ins->sreg1]->opcode == OP_PCONST && defs [ins->sreg2] && defs [ins->sreg2]->opcode == OP_PCONST) {
+					/* typeof(T) == typeof(..) */
+					mono_constant_fold_ins (cfg, ins, defs [ins->sreg1], defs [ins->sreg2], TRUE);
+				} else if (ins->opcode == OP_MOVE && def->opcode == OP_LDADDR) {
+					ins->opcode = OP_LDADDR;
+					ins->sreg1 = -1;
+					ins->inst_p0 = def->inst_p0;
+					ins->klass = def->klass;
 				}
 			}
 
@@ -998,4 +1012,8 @@ mono_local_deadce (MonoCompile *cfg)
 	//mono_print_code (cfg, "AFTER LOCAL-DEADCE");
 }
 
-#endif /* DISABLE_JIT */
+#else /* !DISABLE_JIT */
+
+MONO_EMPTY_SOURCE_FILE (local_propagation);
+
+#endif /* !DISABLE_JIT */
