@@ -1,5 +1,6 @@
-/*
- * main.c: The main entry point for the mono executable
+/**
+ * \file
+ * The main entry point for the mono executable
  *
  * The main entry point does a few things:
  * 
@@ -15,11 +16,15 @@
  */
 #include <config.h>
 #include <fcntl.h>
+#ifndef HOST_WIN32
+#include <dirent.h>
+#endif
 #include <mono/metadata/assembly.h>
 #include <mono/metadata/mono-config.h>
 #include <mono/utils/mono-mmap.h>
 #include <mono/utils/mono-dl.h>
 #include "mini.h"
+#include "mini-runtime.h"
 
 #ifdef HAVE_UNISTD_H
 #  include <unistd.h>
@@ -135,23 +140,23 @@ static GSList *bundle_library_paths;
 static char *bundled_dylibrary_directory;
 
 static void
-delete_bundled_libraries ()
+delete_bundled_libraries (void)
 {
 	GSList *list;
 
 	for (list = bundle_library_paths; list != NULL; list = list->next){
-		unlink (list->data);
+		unlink ((const char*)list->data);
 	}
 	rmdir (bundled_dylibrary_directory);
 }
 
 static void
-bundle_save_library_initialize ()
+bundle_save_library_initialize (void)
 {
 	bundle_save_library_initialized = 1;
 	char *path = g_build_filename (g_get_tmp_dir (), "mono-bundle-XXXXXX", NULL);
 	bundled_dylibrary_directory = g_mkdtemp (path);
-	/* don't free path - mkdtemp modifies it in place, and bundled_dylibrary_directory is an alias of it */
+	g_free (path);
 	if (bundled_dylibrary_directory == NULL)
 		return;
 	atexit (delete_bundled_libraries);
@@ -183,6 +188,47 @@ save_library (int fd, uint64_t offset, uint64_t size, const char *destfname)
 	g_free (buffer);
 }
 
+#ifndef HOST_WIN32
+static gboolean
+search_directories(const char *envPath, const char *program, char **new_program)
+{
+	gchar **paths = NULL;
+	gint i;
+
+	paths = g_strsplit (envPath, G_SEARCHPATH_SEPARATOR_S, 0);
+	g_assert (paths);
+
+	for (i = 0; paths [i]; ++i) {
+		gchar *path = paths [i];
+		gint path_len = strlen (path);
+		DIR *dir;
+		struct dirent *ent;
+
+		if (path_len == 0)
+			continue;
+
+		dir = opendir (path);
+		if (!dir)
+			continue;
+
+		while ((ent = readdir (dir))){
+			if (!strcmp (ent->d_name, program)){
+				*new_program = g_strdup_printf ("%s%s%s", path, path [path_len - 1] == '/' ? "" : "/", program);
+				closedir (dir);
+				g_strfreev (paths);
+				return TRUE;
+			}
+		}
+
+		closedir (dir);
+	}
+
+	*new_program = NULL;
+	g_strfreev (paths);
+	return FALSE;
+}
+#endif
+
 static gboolean
 probe_embedded (const char *program, int *ref_argc, char **ref_argv [])
 {
@@ -202,6 +248,20 @@ probe_embedded (const char *program, int *ref_argc, char **ref_argv [])
 	int j;
 
 	int fd = open (program, O_RDONLY);
+#ifndef HOST_WIN32
+	if (fd == -1){
+		// Also search through the PATH in case the program was run from a different directory
+		gchar* envPath = getenv ("PATH");
+		if (envPath){
+			gchar *new_program = NULL;
+			if (search_directories (envPath, program, &new_program)){
+				fd = open (new_program, O_RDONLY);
+				g_free (new_program);
+				new_program = NULL;
+			}
+		}
+	}
+#endif
 	if (fd == -1)
 		return FALSE;
 	if ((sigstart = lseek (fd, -24, SEEK_END)) == -1)
@@ -227,7 +287,8 @@ probe_embedded (const char *program, int *ref_argc, char **ref_argv [])
 	for (i = 0; i < items; i++){
 		char *kind;
 		int strsize = STREAM_INT (p);
-		uint64_t offset, item_size;
+		uint64_t offset;
+		uint32_t item_size;
 		kind = p+4;
 		p += 4 + strsize;
 		offset = STREAM_LONG(p);
@@ -235,10 +296,15 @@ probe_embedded (const char *program, int *ref_argc, char **ref_argv [])
 		item_size = STREAM_INT (p);
 		p += 4;
 		
-		if (mapaddress == NULL){
-			mapaddress = mono_file_map (directory_location-offset, MONO_MMAP_READ | MONO_MMAP_PRIVATE, fd, offset, &maphandle);
-			if (mapaddress == NULL){
-				perror ("Error mapping file");
+		if (mapaddress == NULL) {
+			char *error_message = NULL;
+			mapaddress = (guchar*)mono_file_map_error (directory_location - offset, MONO_MMAP_READ | MONO_MMAP_PRIVATE,
+				fd, offset, &maphandle, program, &error_message);
+			if (mapaddress == NULL) {
+				if (error_message)
+					fprintf (stderr, "Error mapping file: %s\n", error_message);
+				else
+					perror ("Error mapping file");
 				exit (1);
 			}
 			baseline = offset;
@@ -261,7 +327,9 @@ probe_embedded (const char *program, int *ref_argc, char **ref_argv [])
 		} else if (strncmp (kind, "options:", strlen ("options:")) == 0){
 			mono_parse_options_from (load_from_region (fd, offset, item_size), ref_argc, ref_argv);
 		} else if (strncmp (kind, "config_dir:", strlen ("config_dir:")) == 0){
-			mono_set_dirs (getenv ("MONO_PATH"), load_from_region (fd, offset, item_size));
+			char *mono_path_value = g_getenv ("MONO_PATH");
+			mono_set_dirs (mono_path_value, load_from_region (fd, offset, item_size));
+			g_free (mono_path_value);
 		} else if (strncmp (kind, "machineconfig:", strlen ("machineconfig:")) == 0) {
 			mono_register_machine_config (load_from_region (fd, offset, item_size));
 		} else if (strncmp (kind, "env:", strlen ("env:")) == 0){
