@@ -1,5 +1,6 @@
-/*
- * tramp-ppc.c: JIT trampoline code for PowerPC
+/**
+ * \file
+ * JIT trampoline code for PowerPC
  *
  * Authors:
  *   Dietmar Maurer (dietmar@ximian.com)
@@ -22,6 +23,7 @@
 
 #include "mini.h"
 #include "mini-ppc.h"
+#include "mini-runtime.h"
 
 #if 0
 /* Same as mono_create_ftnptr, but doesn't require a domain */
@@ -35,6 +37,8 @@ mono_ppc_create_ftnptr (guint8 *code)
 	ftnptr->toc = NULL;
 	ftnptr->env = NULL;
 
+	MONO_PROFILER_RAISE (jit_code_buffer, (ftnptr, sizeof (MonoPPCFunctionDescriptor), MONO_PROFILER_CODE_BUFFER_HELPER, NULL));
+
 	return ftnptr;
 #else
 	return code;
@@ -91,29 +95,32 @@ mono_arch_get_unbox_trampoline (MonoMethod *m, gpointer addr)
 	mono_domain_unlock (domain);
 
 	if (short_branch) {
-		ppc_addi (code, this_pos, this_pos, sizeof (MonoObject));
+		ppc_addi (code, this_pos, this_pos, MONO_ABI_SIZEOF (MonoObject));
 		ppc_emit32 (code, short_branch);
 	} else {
 		ppc_load_ptr (code, ppc_r0, addr);
 		ppc_mtctr (code, ppc_r0);
-		ppc_addi (code, this_pos, this_pos, sizeof (MonoObject));
+		ppc_addi (code, this_pos, this_pos, MONO_ABI_SIZEOF (MonoObject));
 		ppc_bcctr (code, 20, 0);
 	}
 	mono_arch_flush_icache (start, code - start);
+	MONO_PROFILER_RAISE (jit_code_buffer, (start, code - start, MONO_PROFILER_CODE_BUFFER_UNBOX_TRAMPOLINE, m));
 	g_assert ((code - start) <= size);
 	/*g_print ("unbox trampoline at %d for %s:%s\n", this_pos, m->klass->name, m->name);
 	g_print ("unbox code is at %p for method at %p\n", start, addr);*/
 
+	mono_tramp_info_register (mono_tramp_info_create (NULL, start, code - start, NULL, NULL), domain);
+
 	return start;
 }
 
 /*
  * mono_arch_get_static_rgctx_trampoline:
  *
- *   Create a trampoline which sets RGCTX_REG to MRGCTX, then jumps to ADDR.
+ *   Create a trampoline which sets RGCTX_REG to ARG, then jumps to ADDR.
  */
 gpointer
-mono_arch_get_static_rgctx_trampoline (MonoMethod *m, MonoMethodRuntimeGenericContext *mrgctx, gpointer addr)
+mono_arch_get_static_rgctx_trampoline (gpointer arg, gpointer addr)
 {
 	guint8 *code, *start, *p;
 	guint8 imm_buf [128];
@@ -124,9 +131,9 @@ mono_arch_get_static_rgctx_trampoline (MonoMethod *m, MonoMethodRuntimeGenericCo
 
 	addr = mono_get_addr_from_ftnptr (addr);
 
-	/* Compute size of code needed to emit mrgctx */
+	/* Compute size of code needed to emit the arg */
 	p = imm_buf;
-	ppc_load_ptr (p, MONO_ARCH_RGCTX_REG, mrgctx);
+	ppc_load_ptr (p, MONO_ARCH_RGCTX_REG, arg);
 	imm_size = p - imm_buf;
 
 	mono_domain_lock (domain);
@@ -138,17 +145,20 @@ mono_arch_get_static_rgctx_trampoline (MonoMethod *m, MonoMethodRuntimeGenericCo
 	mono_domain_unlock (domain);
 
 	if (short_branch) {
-		ppc_load_ptr (code, MONO_ARCH_RGCTX_REG, mrgctx);
+		ppc_load_ptr (code, MONO_ARCH_RGCTX_REG, arg);
 		ppc_emit32 (code, short_branch);
 	} else {
 		ppc_load_ptr (code, ppc_r0, addr);
 		ppc_mtctr (code, ppc_r0);
-		ppc_load_ptr (code, MONO_ARCH_RGCTX_REG, mrgctx);
+		ppc_load_ptr (code, MONO_ARCH_RGCTX_REG, arg);
 		ppc_bcctr (code, 20, 0);
 	}
 	mono_arch_flush_icache (start, code - start);
+	MONO_PROFILER_RAISE (jit_code_buffer, (start, code - start, MONO_PROFILER_CODE_BUFFER_GENERICS_TRAMPOLINE, NULL));
 	g_assert ((code - start) <= size);
 
+	mono_tramp_info_register (mono_tramp_info_create (NULL, start, code - start, NULL, NULL), domain);
+
 	return start;
 }
 
@@ -198,12 +208,6 @@ mono_arch_patch_plt_entry (guint8 *code, gpointer *got, mgreg_t *regs, guint8 *a
 	*(guint8**)((guint8*)got + offset) = addr;
 }
 
-void
-mono_arch_nullify_class_init_trampoline (guint8 *code, mgreg_t *regs)
-{
-	mono_arch_patch_callsite (NULL, code, mini_get_nullified_class_init_trampoline ());
-}
-
 /* Stack size for trampoline function 
  * PPC_MINIMAL_STACK_SIZE + 16 (args + alignment to ppc_magic_trampoline)
  * + MonoLMF + 14 fp regs + 13 gregs + alignment
@@ -241,10 +245,10 @@ guchar*
 mono_arch_create_generic_trampoline (MonoTrampolineType tramp_type, MonoTrampInfo **info, gboolean aot)
 {
 	char *tramp_name;
-	guint8 *buf, *code = NULL;
-	int i, offset;
+	guint8 *buf, *code = NULL, *exception_branch;
+	int i, offset, offset_r14 = 0;
 	gconstpointer tramp_handler;
-	int size = MONO_PPC_32_64_CASE (600, 800);
+	int size = MONO_PPC_32_64_CASE (700, 900);
 	GSList *unwind_ops = NULL;
 	MonoJumpInfo *ji = NULL;
 
@@ -279,6 +283,9 @@ mono_arch_create_generic_trampoline (MonoTrampolineType tramp_type, MonoTrampInf
 	offset = GREGS_OFFSET;
 	for (i = 0; i < 31; i++) {
 		ppc_str (code, i, offset, ppc_r1);
+		if (i == ppc_r14) {
+			offset_r14 = offset;
+		}
 		offset += sizeof (mgreg_t);
 	}
 
@@ -348,10 +355,7 @@ mono_arch_create_generic_trampoline (MonoTrampolineType tramp_type, MonoTrampInf
 		ppc_ldr  (code, ppc_r4, STACK + PPC_RET_ADDR_OFFSET, ppc_r1);
 
 	/* Arg 3: trampoline argument */
-	if (tramp_type == MONO_TRAMPOLINE_GENERIC_CLASS_INIT)
-		ppc_ldr (code, ppc_r5, GREGS_OFFSET + MONO_ARCH_VTABLE_REG * sizeof (mgreg_t), ppc_r1);
-	else
-		ppc_ldr (code, ppc_r5, GREGS_OFFSET, ppc_r1);
+	ppc_ldr (code, ppc_r5, GREGS_OFFSET, ppc_r1);
 
 	if (aot) {
 		code = mono_arch_emit_load_aotconst (buf, code, &ji, MONO_PATCH_INFO_JIT_ICALL_ADDR, g_strdup_printf ("trampoline_func_%d", tramp_type));
@@ -368,16 +372,17 @@ mono_arch_create_generic_trampoline (MonoTrampolineType tramp_type, MonoTrampInf
 		ppc_blrl (code);
 	}
 		
-	/* OK, code address is now on r3. Move it to the counter reg
-	 * so it will be ready for the final jump: this is safe since we
-	 * won't do any more calls.
-	 */
+	/* OK, code address is now on r3, move it to r14 for now.  */
 	if (!MONO_TRAMPOLINE_TYPE_MUST_RETURN (tramp_type)) {
 #ifdef PPC_USES_FUNCTION_DESCRIPTOR
 		ppc_ldptr (code, ppc_r2, sizeof (gpointer), ppc_r3);
 		ppc_ldptr (code, ppc_r3, 0, ppc_r3);
 #endif
-		ppc_mtctr (code, ppc_r3);
+		ppc_mr (code, ppc_r14, ppc_r3);
+	} else {
+		// TODO: is here function descriptor unpacking necessary?
+		/* we clobber r3 during interruption checking, so move it somewhere else */
+		ppc_mr (code, ppc_r14, ppc_r3);
 	}
 
 	/*
@@ -387,12 +392,57 @@ mono_arch_create_generic_trampoline (MonoTrampolineType tramp_type, MonoTrampInf
 	 * The pointer to MonoLMF is in ppc_r12.
 	 */
 	ppc_addi (code, ppc_r12, ppc_r1, STACK - sizeof (MonoLMF));
-	/* r5 = previous_lmf */
-	ppc_ldptr (code, ppc_r5, G_STRUCT_OFFSET(MonoLMF, previous_lmf), ppc_r12);
-	/* r6 = lmf_addr */
-	ppc_ldptr (code, ppc_r6, G_STRUCT_OFFSET(MonoLMF, lmf_addr), ppc_r12);
+	/* r3 = previous_lmf */
+	ppc_ldptr (code, ppc_r3, G_STRUCT_OFFSET(MonoLMF, previous_lmf), ppc_r12);
+	/* r12 = lmf_addr */
+	ppc_ldptr (code, ppc_r12, G_STRUCT_OFFSET(MonoLMF, lmf_addr), ppc_r12);
 	/* *(lmf_addr) = previous_lmf */
-	ppc_stptr (code, ppc_r5, G_STRUCT_OFFSET(MonoLMF, previous_lmf), ppc_r6);
+	ppc_stptr (code, ppc_r3, G_STRUCT_OFFSET(MonoLMF, previous_lmf), ppc_r12);
+
+	/* thread interruption check */
+	if (aot) {
+		g_error ("Not implemented");
+	} else {
+		gconstpointer checkpoint = mono_thread_force_interruption_checkpoint_noraise;
+		ppc_load_func (code, PPC_CALL_REG, checkpoint);
+		ppc_mtlr (code, PPC_CALL_REG);
+	}
+	ppc_blrl (code);
+
+	ppc_compare_reg_imm (code, 0, ppc_r3, 0);
+	exception_branch = code;
+	ppc_bc (code, PPC_BR_TRUE, PPC_BR_EQ, 0);
+
+	/* exception case */
+
+	/* restore caller frame, as we want to throw from there */
+	ppc_ldr  (code, ppc_r14, offset_r14, ppc_r1); /* unclobber r14 */
+	ppc_ldr  (code, ppc_r1,  0, ppc_r1);
+	ppc_ldr  (code, ppc_r12, PPC_RET_ADDR_OFFSET, ppc_r1);
+	ppc_mtlr (code, ppc_r12);
+
+	if (aot) {
+		g_error ("Not implemented");
+	} else {
+		ppc_load_func (code, PPC_CALL_REG, mono_get_throw_exception_addr ());
+		ppc_ldr (code, PPC_CALL_REG, 0, PPC_CALL_REG);
+		ppc_mtctr (code, PPC_CALL_REG);
+	}
+	ppc_bcctr (code, 20, 0);
+	ppc_break (code); /* never reached */
+
+	ppc_patch (exception_branch, code);
+
+	/* no exception case */
+	if (!MONO_TRAMPOLINE_TYPE_MUST_RETURN (tramp_type)) {
+		/* we don't do any calls anymore, so code address can be safely moved
+		 * into counter register */
+		ppc_mtctr (code, ppc_r14);
+	} else {
+		ppc_mr (code, ppc_r3, ppc_r14);
+	}
+
+	ppc_addi (code, ppc_r12, ppc_r1, STACK - sizeof (MonoLMF));
 	/* restore iregs */
 	ppc_ldr_multiple (code, ppc_r13, G_STRUCT_OFFSET(MonoLMF, iregs), ppc_r12);
 	/* restore fregs */
@@ -428,15 +478,15 @@ mono_arch_create_generic_trampoline (MonoTrampolineType tramp_type, MonoTrampInf
 
 	/* Flush instruction cache, since we've generated code */
 	mono_arch_flush_icache (buf, code - buf);
-	
+	MONO_PROFILER_RAISE (jit_code_buffer, (buf, code - buf, MONO_PROFILER_CODE_BUFFER_HELPER, NULL));
+
 	/* Sanity check */
 	g_assert ((code - buf) <= size);
 
-	if (info) {
-		tramp_name = mono_get_generic_trampoline_name (tramp_type);
-		*info = mono_tramp_info_create (tramp_name, buf, code - buf, ji, unwind_ops);
-		g_free (tramp_name);
-	}
+	g_assert (info);
+	tramp_name = mono_get_generic_trampoline_name (tramp_type);
+	*info = mono_tramp_info_create (tramp_name, buf, code - buf, ji, unwind_ops);
+	g_free (tramp_name);
 
 	return buf;
 }
@@ -476,6 +526,7 @@ mono_arch_create_specific_trampoline (gpointer arg1, MonoTrampolineType tramp_ty
 	
 	/* Flush instruction cache, since we've generated code */
 	mono_arch_flush_icache (buf, code - buf);
+	MONO_PROFILER_RAISE (jit_code_buffer, (buf, code - buf, MONO_PROFILER_CODE_BUFFER_SPECIFIC_TRAMPOLINE, mono_get_generic_trampoline_simple_name (tramp_type)));
 
 	g_assert ((code - buf) <= TRAMPOLINE_SIZE);
 
@@ -506,7 +557,6 @@ emit_trampoline_jump (guint8 *code, guint8 *tramp)
 gpointer
 mono_arch_create_rgctx_lazy_fetch_trampoline (guint32 slot, MonoTrampInfo **info, gboolean aot)
 {
-#ifdef MONO_ARCH_VTABLE_REG
 	guint8 *tramp;
 	guint8 *code, *buf;
 	guint8 **rgctx_null_jumps;
@@ -596,7 +646,7 @@ mono_arch_create_rgctx_lazy_fetch_trampoline (guint32 slot, MonoTrampInfo **info
 		ppc_mtctr (code, ppc_r12);
 		ppc_bcctr (code, PPC_BR_ALWAYS, 0);
 	} else {
-		tramp = mono_arch_create_specific_trampoline (GUINT_TO_POINTER (slot),
+		tramp = (guint8*)mono_arch_create_specific_trampoline (GUINT_TO_POINTER (slot),
 			MONO_TRAMPOLINE_RGCTX_LAZY_FETCH, mono_get_root_domain (), NULL);
 
 		/* jump to the actual trampoline */
@@ -604,92 +654,13 @@ mono_arch_create_rgctx_lazy_fetch_trampoline (guint32 slot, MonoTrampInfo **info
 	}
 
 	mono_arch_flush_icache (buf, code - buf);
+	MONO_PROFILER_RAISE (jit_code_buffer, (buf, code - buf, MONO_PROFILER_CODE_BUFFER_GENERICS_TRAMPOLINE, NULL));
 
 	g_assert (code - buf <= tramp_size);
 
-	if (info) {
-		char *name = mono_get_rgctx_fetch_trampoline_name (slot);
-		*info = mono_tramp_info_create (name, buf, code - buf, ji, unwind_ops);
-		g_free (name);
-	}
-
-	return buf;
-#else
-	g_assert_not_reached ();
-#endif
-}
-
-gpointer
-mono_arch_create_generic_class_init_trampoline (MonoTrampInfo **info, gboolean aot)
-{
-	guint8 *tramp;
-	guint8 *code, *buf;
-	static int byte_offset = -1;
-	static guint8 bitmask;
-	guint8 *jump;
-	int tramp_size;
-	GSList *unwind_ops = NULL;
-	MonoJumpInfo *ji = NULL;
-
-	tramp_size = MONO_PPC_32_64_CASE (32, 44);
-	if (aot)
-		tramp_size += 32;
-
-	code = buf = mono_global_codeman_reserve (tramp_size);
-
-	if (byte_offset < 0)
-		mono_marshal_find_bitfield_offset (MonoVTable, initialized, &byte_offset, &bitmask);
-
-	ppc_lbz (code, ppc_r4, byte_offset, MONO_ARCH_VTABLE_REG);
-	ppc_andid (code, ppc_r4, ppc_r4, bitmask);
-	jump = code;
-	ppc_bc (code, PPC_BR_TRUE, PPC_BR_EQ, 0);
-
-	ppc_blr (code);
-
-	ppc_patch (jump, code);
-
-	if (aot) {
-		code = mono_arch_emit_load_aotconst (buf, code, &ji, MONO_PATCH_INFO_JIT_ICALL_ADDR, "specific_trampoline_generic_class_init");
-		/* Branch to the trampoline */
-#ifdef PPC_USES_FUNCTION_DESCRIPTOR
-		ppc_ldptr (code, ppc_r12, 0, ppc_r12);
-#endif
-		ppc_mtctr (code, ppc_r12);
-		ppc_bcctr (code, PPC_BR_ALWAYS, 0);
-	} else {
-		tramp = mono_arch_create_specific_trampoline (NULL, MONO_TRAMPOLINE_GENERIC_CLASS_INIT,
-			mono_get_root_domain (), NULL);
-
-		/* jump to the actual trampoline */
-		code = emit_trampoline_jump (code, tramp);
-	}
-
-	mono_arch_flush_icache (buf, code - buf);
-
-	g_assert (code - buf <= tramp_size);
-
-	if (info)
-		*info = mono_tramp_info_create ("generic_class_init_trampoline", buf, code - buf, ji, unwind_ops);
-
-	return buf;
-}
-
-gpointer
-mono_arch_get_nullified_class_init_trampoline (MonoTrampInfo **info)
-{
-	guint8 *code, *buf;
-	guint32 tramp_size = 64;
-
-	code = buf = mono_global_codeman_reserve (tramp_size);
-	ppc_blr (code);
-
-	mono_arch_flush_icache (buf, code - buf);
-
-	g_assert (code - buf <= tramp_size);
-
-	if (info)
-		*info = mono_tramp_info_create ("nullified_class_init_trampoline", buf, code - buf, NULL, NULL);
+	char *name = mono_get_rgctx_fetch_trampoline_name (slot);
+	*info = mono_tramp_info_create (name, buf, code - buf, ji, unwind_ops);
+	g_free (name);
 
 	return buf;
 }
