@@ -1,5 +1,6 @@
-/*
- * unwind.c: Stack Unwinding Interface
+/**
+ * \file
+ * Stack Unwinding Interface
  *
  * Authors:
  *   Zoltan Varga (vargaz@gmail.com)
@@ -13,6 +14,7 @@
 #include <mono/utils/mono-counters.h>
 #include <mono/utils/freebsd-dwarf.h>
 #include <mono/utils/hazard-pointer.h>
+#include <mono/utils/mono-logger-internals.h>
 #include <mono/metadata/threads-types.h>
 #include <mono/metadata/mono-endian.h>
 
@@ -31,8 +33,6 @@ typedef struct {
 	guint8 info [MONO_ZERO_LEN_ARRAY];
 } MonoUnwindInfo;
 
-#define ALIGN_TO(val,align) ((((size_t)val) + ((align) - 1)) & ~((align) - 1))
-
 static mono_mutex_t unwind_mutex;
 
 static MonoUnwindInfo **cached_info;
@@ -92,8 +92,15 @@ static int map_hw_reg_to_dwarf_reg [ppc_lr + 1] = { 0, 1, 2, 3, 4, 5, 6, 7, 8,
 #endif
 #define NUM_DWARF_REGS (DWARF_PC_REG + 1)
 #elif defined (TARGET_S390X)
-static int map_hw_reg_to_dwarf_reg [] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 };
-#define NUM_DWARF_REGS 16
+/*
+ * 0-15 = GR0-15
+ * 16-31 = FP0-15
+ */
+static int map_hw_reg_to_dwarf_reg [] = {  0,  1,  2,  3,  4,  5,  6,  7, 
+					   8,  9, 10, 11, 12, 13, 14, 15,
+					  16, 17, 18, 19, 20, 21, 22, 23, 
+					  24, 25, 26, 27, 28, 29, 30, 31};
+#define NUM_DWARF_REGS 32
 #define DWARF_DATA_ALIGN (-8)
 #define DWARF_PC_REG (mono_hw_reg_to_dwarf_reg (14))
 #elif defined (TARGET_MIPS)
@@ -449,6 +456,13 @@ mono_unwind_ops_encode_full (GSList *unwind_ops, guint32 *out_len, gboolean enab
 			g_assert (op->val == 0);
 			*p ++ = op->op;
 			break;
+#if defined(TARGET_WIN32) && defined(TARGET_AMD64)
+		case DW_CFA_mono_sp_alloc_info_win64:
+		case DW_CFA_mono_fp_alloc_info_win64:
+			// Drop Windows specific unwind op's. These op's are currently
+			// only used when registering unwind info with Windows OS unwinder.
+			break;
+#endif
 		default:
 			g_assert_not_reached ();
 			break;
@@ -502,8 +516,10 @@ typedef struct {
  * N was saved, or NULL, if it was not saved by this frame.
  * MARK_LOCATIONS should contain the locations marked by mono_emit_unwind_op_mark_loc (), if any.
  * This function is signal safe.
+ *
+ * It returns FALSE on failure
  */
-void
+gboolean
 mono_unwind_frame (guint8 *unwind_info, guint32 unwind_info_len, 
 				   guint8 *start_ip, guint8 *end_ip, guint8 *ip, guint8 **mark_locations,
 				   mono_unwind_reg_t *regs, int nregs,
@@ -561,7 +577,10 @@ mono_unwind_frame (guint8 *unwind_info, guint32 unwind_info_len,
 				reg = decode_uleb128 (p, &p);
 				hwreg = mono_dwarf_reg_to_hw_reg (reg);
 				offset = decode_sleb128 (p, &p);
-				g_assert (reg < NUM_DWARF_REGS);
+				if (reg >= NUM_DWARF_REGS) {
+					mono_runtime_printf_err ("Unwind failure. Assertion at %s %d\n.", __FILE__, __LINE__);
+					return FALSE;
+				}
 				reg_saved [hwreg] = TRUE;
 				locations [hwreg].loc_type = LOC_OFFSET;
 				locations [hwreg].offset = offset * DWARF_DATA_ALIGN;
@@ -570,7 +589,10 @@ mono_unwind_frame (guint8 *unwind_info, guint32 unwind_info_len,
 				reg = decode_uleb128 (p, &p);
 				hwreg = mono_dwarf_reg_to_hw_reg (reg);
 				offset = decode_uleb128 (p, &p);
-				g_assert (reg < NUM_DWARF_REGS);
+				if (reg >= NUM_DWARF_REGS) {
+					mono_runtime_printf_err ("Unwind failure. Assertion at %s %d\n.", __FILE__, __LINE__);
+					return FALSE;
+				}
 				reg_saved [hwreg] = TRUE;
 				locations [hwreg].loc_type = LOC_OFFSET;
 				locations [hwreg].offset = offset * DWARF_DATA_ALIGN;
@@ -592,7 +614,10 @@ mono_unwind_frame (guint8 *unwind_info, guint32 unwind_info_len,
 				p += 4;
 				break;
 			case DW_CFA_remember_state:
-				g_assert (state_stack_pos == 0);
+				if (state_stack_pos != 0) {
+					mono_runtime_printf_err ("Unwind failure. Assertion at %s %d\n.", __FILE__, __LINE__);
+					return FALSE;
+				}
 				memcpy (&state_stack [0].locations, &locations, sizeof (locations));
 				memcpy (&state_stack [0].reg_saved, &reg_saved, sizeof (reg_saved));
 				state_stack [0].cfa_reg = cfa_reg;
@@ -600,7 +625,10 @@ mono_unwind_frame (guint8 *unwind_info, guint32 unwind_info_len,
 				state_stack_pos ++;
 				break;
 			case DW_CFA_restore_state:
-				g_assert (state_stack_pos == 1);
+				if (state_stack_pos != 1) {
+					mono_runtime_printf_err ("Unwind failure. Assertion at %s %d\n.", __FILE__, __LINE__);
+					return FALSE;
+				}
 				state_stack_pos --;
 				memcpy (&locations, &state_stack [0].locations, sizeof (locations));
 				memcpy (&reg_saved, &state_stack [0].reg_saved, sizeof (reg_saved));
@@ -608,28 +636,40 @@ mono_unwind_frame (guint8 *unwind_info, guint32 unwind_info_len,
 				cfa_offset = state_stack [0].cfa_offset;
 				break;
 			case DW_CFA_mono_advance_loc:
-				g_assert (mark_locations [0]);
+				if (!mark_locations [0]) {
+					mono_runtime_printf_err ("Unwind failure. Assertion at %s %d\n.", __FILE__, __LINE__);
+					return FALSE;
+				}
 				pos = mark_locations [0] - start_ip;
 				break;
 			default:
-				g_assert_not_reached ();
+				mono_runtime_printf_err ("Unwind failure. Illegal value for switch statement, assertion at %s %d\n.", __FILE__, __LINE__);
+				return FALSE;
 			}
 			break;
 		}
 		default:
-			g_assert_not_reached ();
+			mono_runtime_printf_err ("Unwind failure. Illegal value for switch statement, assertion at %s %d\n.", __FILE__, __LINE__);
+			return FALSE;
 		}
 	}
 
 	if (save_locations)
 		memset (save_locations, 0, save_locations_len * sizeof (mgreg_t*));
 
-	g_assert (cfa_reg != -1);
+	if (cfa_reg == -1) {
+		mono_runtime_printf_err ("Unset cfa_reg in method %s. Memory around ip (%p):", mono_get_method_from_ip (ip), ip);
+		mono_dump_mem (ip - 0x10, 0x40);
+		return FALSE;
+	}
 	cfa_val = (guint8*)regs [mono_dwarf_reg_to_hw_reg (cfa_reg)] + cfa_offset;
 	for (hwreg = 0; hwreg < NUM_HW_REGS; ++hwreg) {
 		if (reg_saved [hwreg] && locations [hwreg].loc_type == LOC_OFFSET) {
 			int dwarfreg = mono_hw_reg_to_dwarf_reg (hwreg);
-			g_assert (hwreg < nregs);
+			if (hwreg >= nregs) {
+				mono_runtime_printf_err ("Unwind failure. Assertion at %s %d\n.", __FILE__, __LINE__);
+				return FALSE;
+			}
 			if (IS_DOUBLE_REG (dwarfreg))
 				regs [hwreg] = *(guint64*)(cfa_val + locations [hwreg].offset);
 			else
@@ -640,6 +680,9 @@ mono_unwind_frame (guint8 *unwind_info, guint32 unwind_info_len,
 	}
 
 	*out_cfa = cfa_val;
+
+	// Success
+	return TRUE;
 }
 
 void
@@ -859,9 +902,10 @@ read_encoded_val (guint32 encoding, guint8 *p, guint8 **endp)
  * decode_lsda:
  *
  *   Decode the Mono specific Language Specific Data Area generated by LLVM.
+ * This function is async safe.
  */
 static void
-decode_lsda (guint8 *lsda, guint8 *code, MonoJitExceptionInfo **ex_info, guint32 *ex_info_len, gpointer **type_info, int *this_reg, int *this_offset)
+decode_lsda (guint8 *lsda, guint8 *code, MonoJitExceptionInfo *ex_info, gpointer *type_info, guint32 *ex_info_len, int *this_reg, int *this_offset)
 {
 	guint8 *p;
 	int i, ncall_sites, this_encoding;
@@ -897,12 +941,8 @@ decode_lsda (guint8 *lsda, guint8 *code, MonoJitExceptionInfo **ex_info, guint32
 	ncall_sites = decode_uleb128 (p, &p);
 	p = (guint8*)ALIGN_TO ((mgreg_t)p, 4);
 
-	if (ex_info) {
-		*ex_info = (MonoJitExceptionInfo *)g_malloc0 (ncall_sites * sizeof (MonoJitExceptionInfo));
+	if (ex_info_len)
 		*ex_info_len = ncall_sites;
-	}
-	if (type_info)
-		*type_info = (gpointer *)g_malloc0 (ncall_sites * sizeof (gpointer));
 
 	for (i = 0; i < ncall_sites; ++i) {
 		int block_start_offset, block_size, landing_pad;
@@ -922,11 +962,11 @@ decode_lsda (guint8 *lsda, guint8 *code, MonoJitExceptionInfo **ex_info, guint32
 		//printf ("X: %p %d\n", landing_pad, *(int*)tinfo);
 
 		if (ex_info) {
-			if (*type_info)
-				(*type_info) [i] = tinfo;
-			(*ex_info)[i].try_start = code + block_start_offset;
-			(*ex_info)[i].try_end = code + block_start_offset + block_size;
-			(*ex_info)[i].handler_start = code + landing_pad;
+			if (type_info)
+				type_info [i] = tinfo;
+			ex_info[i].try_start = code + block_start_offset;
+			ex_info[i].try_end = code + block_start_offset + block_size;
+			ex_info[i].handler_start = code + landing_pad;
 		}
 	}
 }
@@ -1065,7 +1105,16 @@ mono_unwind_decode_fde (guint8 *fde, guint32 *out_len, guint32 *code_len, MonoJi
 		if (lsda_offset != 0) {
 			lsda = fde_aug + lsda_offset;
 
-			decode_lsda (lsda, code, ex_info, ex_info_len, type_info, this_reg, this_offset);
+			/* Get the lengths first */
+			guint32 len;
+			decode_lsda (lsda, code, NULL, NULL, &len, this_reg, this_offset);
+
+			if (ex_info)
+				*ex_info = (MonoJitExceptionInfo *)g_malloc0 (len * sizeof (MonoJitExceptionInfo));
+			if (type_info)
+				*type_info = (gpointer *)g_malloc0 (len * sizeof (gpointer));
+
+			decode_lsda (lsda, code, ex_info ? *ex_info : NULL, type_info ? *type_info : NULL, ex_info_len, this_reg, this_offset);
 		}
 	}
 
@@ -1108,13 +1157,15 @@ mono_unwind_decode_fde (guint8 *fde, guint32 *out_len, guint32 *code_len, MonoJi
  * mono_unwind_decode_mono_fde:
  *
  *   Decode an FDE entry in the LLVM emitted mono EH frame.
- * info->ex_info is set to a malloc-ed array of MonoJitExceptionInfo structures,
- * only try_start, try_end and handler_start is set.
- * info->type_info is set to a malloc-ed array containing the ttype table from the
- * LSDA.
+ * If EI/TYPE_INFO/UNW_INFO are NULL, compute only the value of the scalar fields in INFO.
+ * Otherwise:
+ * - Fill out EX_INFO with try_start, try_end and handler_start.
+ * - Fill out TYPE_INFO with the ttype table from the LSDA.
+ * - Fill out UNW_INFO with the unwind info.
+ * This function is async safe.
  */
 void
-mono_unwind_decode_llvm_mono_fde (guint8 *fde, int fde_len, guint8 *cie, guint8 *code, MonoLLVMFDEInfo *res)
+mono_unwind_decode_llvm_mono_fde (guint8 *fde, int fde_len, guint8 *cie, guint8 *code, MonoLLVMFDEInfo *res, MonoJitExceptionInfo *ex_info, gpointer *type_info, guint8 *unw_info)
 {
 	guint8 *p, *fde_aug, *cie_cfi, *fde_cfi, *buf;
 	int has_aug, aug_len, cie_cfi_len, fde_cfi_len;
@@ -1144,7 +1195,10 @@ mono_unwind_decode_llvm_mono_fde (guint8 *fde, int fde_len, guint8 *cie, guint8
 		/* The LSDA is embedded directly into the FDE */
 		lsda = fde_aug;
 
-		decode_lsda (lsda, code, &res->ex_info, &res->ex_info_len, &res->type_info, &res->this_reg, &res->this_offset);
+		/* Get the lengths first */
+		decode_lsda (lsda, code, NULL, NULL, &res->ex_info_len, &res->this_reg, &res->this_offset);
+
+		decode_lsda (lsda, code, ex_info, type_info, NULL, &res->this_reg, &res->this_offset);
 	}
 
 	/* Decode CIE */
@@ -1175,12 +1229,13 @@ mono_unwind_decode_llvm_mono_fde (guint8 *fde, int fde_len, guint8 *cie, guint8
 	cie_cfi_len = p - cie_cfi;
 	fde_cfi_len = (fde + fde_len - fde_cfi);
 
-	buf = (guint8 *)g_malloc0 (cie_cfi_len + fde_cfi_len);
-	memcpy (buf, cie_cfi, cie_cfi_len);
-	memcpy (buf + cie_cfi_len, fde_cfi, fde_cfi_len);
+	buf = unw_info;
+	if (buf) {
+		memcpy (buf, cie_cfi, cie_cfi_len);
+		memcpy (buf + cie_cfi_len, fde_cfi, fde_cfi_len);
+	}
 
 	res->unw_info_len = cie_cfi_len + fde_cfi_len;
-	res->unw_info = buf;
 }
 
 /*
