@@ -1,5 +1,6 @@
-/*
- * mono-context.c: plat independent machine state definitions
+/**
+ * \file
+ * plat independent machine state definitions
  *
  *
  * Copyright (c) 2011 Novell, Inc (http://www.novell.com)
@@ -35,7 +36,7 @@
 void
 mono_sigctx_to_monoctx (void *sigctx, MonoContext *mctx)
 {
-#if defined (__native_client__) || defined (HOST_WATCHOS)
+#if defined (HOST_WATCHOS)
 	printf("WARNING: mono_arch_sigctx_to_monoctx() called!\n");
 	mctx->eax = 0xDEADBEEF;
 	mctx->ebx = 0xDEADBEEF;
@@ -60,6 +61,18 @@ mono_sigctx_to_monoctx (void *sigctx, MonoContext *mctx)
 	mctx->esi = UCONTEXT_REG_ESI (ctx);
 	mctx->edi = UCONTEXT_REG_EDI (ctx);
 	mctx->eip = UCONTEXT_REG_EIP (ctx);
+#ifdef UCONTEXT_HAS_XMM
+	if (UCONTEXT_HAS_XMM (ctx)) {
+		mctx->fregs [0] = UCONTEXT_REG_XMM0 (ctx);
+		mctx->fregs [1] = UCONTEXT_REG_XMM1 (ctx);
+		mctx->fregs [2] = UCONTEXT_REG_XMM2 (ctx);
+		mctx->fregs [3] = UCONTEXT_REG_XMM3 (ctx);
+		mctx->fregs [4] = UCONTEXT_REG_XMM4 (ctx);
+		mctx->fregs [5] = UCONTEXT_REG_XMM5 (ctx);
+		mctx->fregs [6] = UCONTEXT_REG_XMM6 (ctx);
+		mctx->fregs [7] = UCONTEXT_REG_XMM7 (ctx);
+	}
+#endif
 #elif defined(HOST_WIN32)
 	CONTEXT *context = (CONTEXT*)sigctx;
 
@@ -84,13 +97,13 @@ mono_sigctx_to_monoctx (void *sigctx, MonoContext *mctx)
 	mctx->esi = ctx->SC_ESI;
 	mctx->edi = ctx->SC_EDI;
 	mctx->eip = ctx->SC_EIP;
-#endif /* if defined(__native_client__) */
+#endif
 }
 
 void
 mono_monoctx_to_sigctx (MonoContext *mctx, void *sigctx)
 {
-#if defined(__native_client__) || defined(HOST_WATCHOS)
+#if defined(HOST_WATCHOS)
 	printf("WARNING: mono_arch_monoctx_to_sigctx() called!\n");
 #elif MONO_CROSS_COMPILE
 	g_assert_not_reached ();
@@ -106,6 +119,18 @@ mono_monoctx_to_sigctx (MonoContext *mctx, void *sigctx)
 	UCONTEXT_REG_ESI (ctx) = mctx->esi;
 	UCONTEXT_REG_EDI (ctx) = mctx->edi;
 	UCONTEXT_REG_EIP (ctx) = mctx->eip;
+#ifdef UCONTEXT_HAS_XMM
+	if (UCONTEXT_HAS_XMM (ctx)) {
+		UCONTEXT_REG_XMM0 (ctx) = mctx->fregs [0];
+		UCONTEXT_REG_XMM1 (ctx) = mctx->fregs [1];
+		UCONTEXT_REG_XMM2 (ctx) = mctx->fregs [2];
+		UCONTEXT_REG_XMM3 (ctx) = mctx->fregs [3];
+		UCONTEXT_REG_XMM4 (ctx) = mctx->fregs [4];
+		UCONTEXT_REG_XMM5 (ctx) = mctx->fregs [5];
+		UCONTEXT_REG_XMM6 (ctx) = mctx->fregs [6];
+		UCONTEXT_REG_XMM7 (ctx) = mctx->fregs [7];
+	}
+#endif
 #elif defined(HOST_WIN32)
 	CONTEXT *context = (CONTEXT*)sigctx;
 
@@ -130,7 +155,7 @@ mono_monoctx_to_sigctx (MonoContext *mctx, void *sigctx)
 	ctx->SC_ESI = mctx->esi;
 	ctx->SC_EDI = mctx->edi;
 	ctx->SC_EIP = mctx->eip;
-#endif /* __native_client__ */
+#endif
 }
 
 #elif (defined(__x86_64__) && !defined(MONO_CROSS_COMPILE)) || (defined(TARGET_AMD64)) /* defined(__i386__) */
@@ -144,10 +169,6 @@ mono_monoctx_to_sigctx (MonoContext *mctx, void *sigctx)
 void
 mono_sigctx_to_monoctx (void *sigctx, MonoContext *mctx)
 {
-#if defined(__native_client_codegen__) || defined(__native_client__)
-	printf("WARNING: mono_arch_sigctx_to_monoctx() called!\n");
-#endif
-
 #ifdef MONO_CROSS_COMPILE
 	g_assert_not_reached ();
 #elif defined(MONO_SIGNAL_USE_UCONTEXT_T)
@@ -171,23 +192,25 @@ mono_sigctx_to_monoctx (void *sigctx, MonoContext *mctx)
 	mctx->gregs [AMD64_R15] = UCONTEXT_REG_R15 (ctx);
 	mctx->gregs [AMD64_RIP] = UCONTEXT_REG_RIP (ctx);
 
-#ifdef UCONTEXT_REG_XMM
-	mctx->fregs [0] = UCONTEXT_REG_XMM0 (ctx);
-	mctx->fregs [1] = UCONTEXT_REG_XMM1 (ctx);
-	mctx->fregs [2] = UCONTEXT_REG_XMM2 (ctx);
-	mctx->fregs [3] = UCONTEXT_REG_XMM3 (ctx);
-	mctx->fregs [4] = UCONTEXT_REG_XMM4 (ctx);
-	mctx->fregs [5] = UCONTEXT_REG_XMM5 (ctx);
-	mctx->fregs [6] = UCONTEXT_REG_XMM6 (ctx);
-	mctx->fregs [7] = UCONTEXT_REG_XMM7 (ctx);
-	mctx->fregs [8] = UCONTEXT_REG_XMM8 (ctx);
-	mctx->fregs [9] = UCONTEXT_REG_XMM9 (ctx);
-	mctx->fregs [10] = UCONTEXT_REG_XMM10 (ctx);
-	mctx->fregs [11] = UCONTEXT_REG_XMM11 (ctx);
-	mctx->fregs [12] = UCONTEXT_REG_XMM12 (ctx);
-	mctx->fregs [13] = UCONTEXT_REG_XMM13 (ctx);
-	mctx->fregs [14] = UCONTEXT_REG_XMM14 (ctx);
-	mctx->fregs [15] = UCONTEXT_REG_XMM15 (ctx);
+#ifdef UCONTEXT_HAS_XMM
+	if (UCONTEXT_HAS_XMM (ctx)) {
+		mctx->fregs [0] = UCONTEXT_REG_XMM0 (ctx);
+		mctx->fregs [1] = UCONTEXT_REG_XMM1 (ctx);
+		mctx->fregs [2] = UCONTEXT_REG_XMM2 (ctx);
+		mctx->fregs [3] = UCONTEXT_REG_XMM3 (ctx);
+		mctx->fregs [4] = UCONTEXT_REG_XMM4 (ctx);
+		mctx->fregs [5] = UCONTEXT_REG_XMM5 (ctx);
+		mctx->fregs [6] = UCONTEXT_REG_XMM6 (ctx);
+		mctx->fregs [7] = UCONTEXT_REG_XMM7 (ctx);
+		mctx->fregs [8] = UCONTEXT_REG_XMM8 (ctx);
+		mctx->fregs [9] = UCONTEXT_REG_XMM9 (ctx);
+		mctx->fregs [10] = UCONTEXT_REG_XMM10 (ctx);
+		mctx->fregs [11] = UCONTEXT_REG_XMM11 (ctx);
+		mctx->fregs [12] = UCONTEXT_REG_XMM12 (ctx);
+		mctx->fregs [13] = UCONTEXT_REG_XMM13 (ctx);
+		mctx->fregs [14] = UCONTEXT_REG_XMM14 (ctx);
+		mctx->fregs [15] = UCONTEXT_REG_XMM15 (ctx);
+	}
 #endif
 
 #elif defined(HOST_WIN32)
@@ -210,6 +233,27 @@ mono_sigctx_to_monoctx (void *sigctx, MonoContext *mctx)
 	mctx->gregs [AMD64_R13] = context->R13;
 	mctx->gregs [AMD64_R14] = context->R14;
 	mctx->gregs [AMD64_R15] = context->R15;
+#elif defined(__HAIKU__)
+	// Haiku uses sigcontext because there's no ucontext
+	struct sigcontext *ctx = (struct sigcontext *)sigctx;
+
+	mctx->gregs [AMD64_RIP] = ctx->regs.rip;
+	mctx->gregs [AMD64_RAX] = ctx->regs.rax;
+	mctx->gregs [AMD64_RCX] = ctx->regs.rcx;
+	mctx->gregs [AMD64_RDX] = ctx->regs.rdx;
+	mctx->gregs [AMD64_RBX] = ctx->regs.rbx;
+	mctx->gregs [AMD64_RSP] = ctx->regs.rsp;
+	mctx->gregs [AMD64_RBP] = ctx->regs.rbp;
+	mctx->gregs [AMD64_RSI] = ctx->regs.rsi;
+	mctx->gregs [AMD64_RDI] = ctx->regs.rdi;
+	mctx->gregs [AMD64_R8] = ctx->regs.r8;
+	mctx->gregs [AMD64_R9] = ctx->regs.r9;
+	mctx->gregs [AMD64_R10] = ctx->regs.r10;
+	mctx->gregs [AMD64_R11] = ctx->regs.r11;
+	mctx->gregs [AMD64_R12] = ctx->regs.r12;
+	mctx->gregs [AMD64_R13] = ctx->regs.r13;
+	mctx->gregs [AMD64_R14] = ctx->regs.r14;
+	mctx->gregs [AMD64_R15] = ctx->regs.r15;
 #else
 	g_assert_not_reached ();
 #endif
@@ -218,10 +262,6 @@ mono_sigctx_to_monoctx (void *sigctx, MonoContext *mctx)
 void
 mono_monoctx_to_sigctx (MonoContext *mctx, void *sigctx)
 {
-#if defined(__native_client__) || defined(__native_client_codegen__)
-  printf("WARNING: mono_arch_monoctx_to_sigctx() called!\n");
-#endif
-
 #ifdef MONO_CROSS_COMPILE
 	g_assert_not_reached ();
 #elif defined(MONO_SIGNAL_USE_UCONTEXT_T)
@@ -245,23 +285,25 @@ mono_monoctx_to_sigctx (MonoContext *mctx, void *sigctx)
 	UCONTEXT_REG_R15 (ctx) = mctx->gregs [AMD64_R15];
 	UCONTEXT_REG_RIP (ctx) = mctx->gregs [AMD64_RIP];
 
-#ifdef UCONTEXT_REG_XMM
-	UCONTEXT_REG_XMM0 (ctx) = mctx->fregs [0];
-	UCONTEXT_REG_XMM1 (ctx) = mctx->fregs [1];
-	UCONTEXT_REG_XMM2 (ctx) = mctx->fregs [2];
-	UCONTEXT_REG_XMM3 (ctx) = mctx->fregs [3];
-	UCONTEXT_REG_XMM4 (ctx) = mctx->fregs [4];
-	UCONTEXT_REG_XMM5 (ctx) = mctx->fregs [5];
-	UCONTEXT_REG_XMM6 (ctx) = mctx->fregs [6];
-	UCONTEXT_REG_XMM7 (ctx) = mctx->fregs [7];
-	UCONTEXT_REG_XMM8 (ctx) = mctx->fregs [8];
-	UCONTEXT_REG_XMM9 (ctx) = mctx->fregs [9];
-	UCONTEXT_REG_XMM10 (ctx) = mctx->fregs [10];
-	UCONTEXT_REG_XMM11 (ctx) = mctx->fregs [11];
-	UCONTEXT_REG_XMM12 (ctx) = mctx->fregs [12];
-	UCONTEXT_REG_XMM13 (ctx) = mctx->fregs [13];
-	UCONTEXT_REG_XMM14 (ctx) = mctx->fregs [14];
-	UCONTEXT_REG_XMM15 (ctx) = mctx->fregs [15];
+#ifdef UCONTEXT_HAS_XMM
+	if (UCONTEXT_HAS_XMM (ctx)) {
+		UCONTEXT_REG_XMM0 (ctx) = mctx->fregs [0];
+		UCONTEXT_REG_XMM1 (ctx) = mctx->fregs [1];
+		UCONTEXT_REG_XMM2 (ctx) = mctx->fregs [2];
+		UCONTEXT_REG_XMM3 (ctx) = mctx->fregs [3];
+		UCONTEXT_REG_XMM4 (ctx) = mctx->fregs [4];
+		UCONTEXT_REG_XMM5 (ctx) = mctx->fregs [5];
+		UCONTEXT_REG_XMM6 (ctx) = mctx->fregs [6];
+		UCONTEXT_REG_XMM7 (ctx) = mctx->fregs [7];
+		UCONTEXT_REG_XMM8 (ctx) = mctx->fregs [8];
+		UCONTEXT_REG_XMM9 (ctx) = mctx->fregs [9];
+		UCONTEXT_REG_XMM10 (ctx) = mctx->fregs [10];
+		UCONTEXT_REG_XMM11 (ctx) = mctx->fregs [11];
+		UCONTEXT_REG_XMM12 (ctx) = mctx->fregs [12];
+		UCONTEXT_REG_XMM13 (ctx) = mctx->fregs [13];
+		UCONTEXT_REG_XMM14 (ctx) = mctx->fregs [14];
+		UCONTEXT_REG_XMM15 (ctx) = mctx->fregs [15];
+	}
 #endif
 
 #elif defined(HOST_WIN32)
@@ -284,6 +326,27 @@ mono_monoctx_to_sigctx (MonoContext *mctx, void *sigctx)
 	context->R13 = mctx->gregs [AMD64_R13];
 	context->R14 = mctx->gregs [AMD64_R14];
 	context->R15 = mctx->gregs [AMD64_R15];
+#elif defined(__HAIKU__)
+	// Haiku uses sigcontext because there's no ucontext
+	struct sigcontext *ctx = (struct sigcontext *)sigctx;
+
+	ctx->regs.rip = mctx->gregs [AMD64_RIP];
+	ctx->regs.rax = mctx->gregs [AMD64_RAX];
+	ctx->regs.rcx = mctx->gregs [AMD64_RCX];
+	ctx->regs.rdx = mctx->gregs [AMD64_RDX];
+	ctx->regs.rbx = mctx->gregs [AMD64_RBX];
+	ctx->regs.rsp = mctx->gregs [AMD64_RSP];
+	ctx->regs.rbp = mctx->gregs [AMD64_RBP];
+	ctx->regs.rsi = mctx->gregs [AMD64_RSI];
+	ctx->regs.rdi = mctx->gregs [AMD64_RDI];
+	ctx->regs.r8 = mctx->gregs [AMD64_R8];
+	ctx->regs.r9 = mctx->gregs [AMD64_R9];
+	ctx->regs.r10 = mctx->gregs [AMD64_R10];
+	ctx->regs.r11 = mctx->gregs [AMD64_R11];
+	ctx->regs.r12 = mctx->gregs [AMD64_R12];
+	ctx->regs.r13 = mctx->gregs [AMD64_R13];
+	ctx->regs.r14 = mctx->gregs [AMD64_R14];
+	ctx->regs.r15 = mctx->gregs [AMD64_R15];
 #else
 	g_assert_not_reached ();
 #endif
@@ -332,15 +395,26 @@ mono_monoctx_to_sigctx (MonoContext *mctx, void *ctx)
 #include <mono/arch/arm/arm-codegen.h>
 #include <mono/arch/arm/arm-vfp-codegen.h>
 
+#ifdef HOST_WIN32
+#include <windows.h>
+#endif
+
 void
 mono_sigctx_to_monoctx (void *sigctx, MonoContext *mctx)
 {
 #ifdef MONO_CROSS_COMPILE
 	g_assert_not_reached ();
-#elif defined(__native_client__)
-	g_assert_not_reached ();
+#elif defined(HOST_WIN32)
+	CONTEXT *context = (CONTEXT*)sigctx;
+
+	mctx->pc = context->Pc;
+	mctx->cpsr = context->Cpsr;
+	memcpy (&mctx->regs, &context->R0, sizeof (DWORD) * 16);
+	
+	/* Why are we only copying 16 registers?! There are 32! */
+	memcpy (&mctx->fregs, &context->D, sizeof (double) * 16);
 #else
-	arm_ucontext *my_uc = sigctx;
+	arm_ucontext *my_uc = (arm_ucontext*)sigctx;
 
 	mctx->pc = UCONTEXT_REG_PC (my_uc);
 	mctx->regs [ARMREG_SP] = UCONTEXT_REG_SP (my_uc);
@@ -357,10 +431,17 @@ mono_monoctx_to_sigctx (MonoContext *mctx, void *ctx)
 {
 #ifdef MONO_CROSS_COMPILE
 	g_assert_not_reached ();
-#elif defined(__native_client__)
-	g_assert_not_reached ();
+#elif defined(HOST_WIN32)
+	CONTEXT *context = (CONTEXT*)ctx;
+
+	context->Pc = mctx->pc;
+	context->Cpsr = mctx->cpsr;
+	memcpy (&context->R0, &mctx->regs, sizeof (DWORD) * 16);
+	
+	/* Why are we only copying 16 registers?! There are 32! */
+	memcpy (&context->D, &mctx->fregs, sizeof (double) * 16);
 #else
-	arm_ucontext *my_uc = ctx;
+	arm_ucontext *my_uc = (arm_ucontext*)ctx;
 
 	UCONTEXT_REG_PC (my_uc) = mctx->pc;
 	UCONTEXT_REG_SP (my_uc) = mctx->regs [ARMREG_SP];
@@ -392,8 +473,7 @@ mono_sigctx_to_monoctx (void *sigctx, MonoContext *mctx)
 
 	g_assert (fpctx->head.magic == FPSIMD_MAGIC);
 	for (i = 0; i < 32; ++i)
-		/* Only store the bottom 8 bytes for now */
-		*(guint64*)&(mctx->fregs [i]) = fpctx->vregs [i];
+		mctx->fregs [i] = fpctx->vregs [i];
 #endif
 	/* FIXME: apple */
 #endif
@@ -452,8 +532,9 @@ mono_sigctx_to_monoctx (void *sigctx, MonoContext *mctx)
 
 	mctx->sc_ir = UCONTEXT_REG_NIP(uc);
 	mctx->sc_sp = UCONTEXT_REG_Rn(uc, 1);
-	memcpy (&mctx->regs, &UCONTEXT_REG_Rn(uc, 13), sizeof (mgreg_t) * MONO_SAVED_GREGS);
-	memcpy (&mctx->fregs, &UCONTEXT_REG_FPRn(uc, 14), sizeof (double) * MONO_SAVED_FREGS);
+
+	memcpy (&mctx->regs, &UCONTEXT_REG_Rn(uc, 0), sizeof (mgreg_t) * MONO_MAX_IREGS);
+	memcpy (&mctx->fregs, &UCONTEXT_REG_FPRn(uc, 0), sizeof (double) * MONO_MAX_FREGS);
 }
 
 void
@@ -461,10 +542,28 @@ mono_monoctx_to_sigctx (MonoContext *mctx, void *sigctx)
 {
 	os_ucontext *uc = sigctx;
 
+	memcpy (&UCONTEXT_REG_Rn(uc, 0), &mctx->regs, sizeof (mgreg_t) * MONO_MAX_IREGS);
+	memcpy (&UCONTEXT_REG_FPRn(uc, 0), &mctx->fregs, sizeof (double) * MONO_MAX_FREGS);
+
+	/* The valid values for pc and sp are stored here and not in regs array */
 	UCONTEXT_REG_NIP(uc) = mctx->sc_ir;
 	UCONTEXT_REG_Rn(uc, 1) = mctx->sc_sp;
-	memcpy (&UCONTEXT_REG_Rn(uc, 13), &mctx->regs, sizeof (mgreg_t) * MONO_SAVED_GREGS);
-	memcpy (&UCONTEXT_REG_FPRn(uc, 14), &mctx->fregs, sizeof (double) * MONO_SAVED_FREGS);
+}
+
+#elif defined (TARGET_WASM)
+
+#include <mono/utils/mono-context.h>
+
+void
+mono_sigctx_to_monoctx (void *sigctx, MonoContext *mctx)
+{
+	g_error ("MonoContext not supported");
+}
+
+void
+mono_monoctx_to_sigctx (MonoContext *mctx, void *sigctx)
+{
+	g_error ("MonoContext not supported");
 }
 
 #endif /* #if defined(__i386__) */
