@@ -1,5 +1,6 @@
-/*
- * mono-context.h: plat independent machine state definitions
+/**
+ * \file
+ * plat independent machine state definitions
  *
  *
  * Copyright (c) 2011 Novell, Inc (http://www.novell.com)
@@ -21,9 +22,34 @@
 #define MONO_CONTEXT_OFFSET(field, index, field_type) \
     "i" (offsetof (MonoContext, field) + (index) * sizeof (field_type))
 
+#if defined(TARGET_X86)
 #if defined(__APPLE__)
+#define MONO_HAVE_SIMD_REG
 typedef struct __darwin_xmm_reg MonoContextSimdReg;
 #endif
+#elif defined(TARGET_AMD64)
+#if defined(__APPLE__)
+#define MONO_HAVE_SIMD_REG
+typedef struct __darwin_xmm_reg MonoContextSimdReg;
+#elif defined(__linux__) && defined(__GLIBC__)
+#define MONO_HAVE_SIMD_REG
+typedef struct _libc_xmmreg MonoContextSimdReg;
+#elif defined(HOST_WIN32)
+#define MONO_HAVE_SIMD_REG
+#include <emmintrin.h>
+typedef __m128d MonoContextSimdReg;
+#elif defined(HOST_ANDROID)
+#define MONO_HAVE_SIMD_REG
+typedef struct _libc_xmmreg MonoContextSimdReg;
+#elif defined(__linux__) || defined(__OpenBSD__)
+#define MONO_HAVE_SIMD_REG
+#include <emmintrin.h>
+typedef __m128d MonoContextSimdReg;
+#endif
+#elif defined(TARGET_ARM64)
+#define MONO_HAVE_SIMD_REG
+typedef __uint128_t MonoContextSimdReg;
+#endif
 
 /*
  * General notes about mono-context.
@@ -33,7 +59,25 @@ typedef struct __darwin_xmm_reg MonoContextSimdReg;
  * MONO_CONTEXT_GET_CURRENT captures the current context as close as possible. One reg might be clobbered
  *  to hold the address of the target MonoContext. It will be a caller save one, so should not be a problem.
  */
-#if (defined(__i386__) && !defined(MONO_CROSS_COMPILE)) || (defined(TARGET_X86))
+#if defined (TARGET_WASM)
+
+typedef struct {
+	mgreg_t wasm_sp;
+	mgreg_t wasm_bp;
+	mgreg_t llvm_exc_reg;
+	mgreg_t wasm_ip;
+	mgreg_t wasm_pc;
+} MonoContext;
+
+#define MONO_CONTEXT_SET_IP(ctx,ip) do { (ctx)->wasm_ip = (mgreg_t)(ip); } while (0);
+#define MONO_CONTEXT_SET_BP(ctx,bp) do { (ctx)->wasm_bp = (mgreg_t)(bp); } while (0);
+#define MONO_CONTEXT_SET_SP(ctx,sp) do { (ctx)->wasm_sp = (mgreg_t)(sp); } while (0);
+
+#define MONO_CONTEXT_GET_IP(ctx) ((gpointer)((ctx)->wasm_ip))
+#define MONO_CONTEXT_GET_BP(ctx) ((gpointer)((ctx)->wasm_bp))
+#define MONO_CONTEXT_GET_SP(ctx) ((gpointer)((ctx)->wasm_sp))
+
+#elif (defined(__i386__) && !defined(MONO_CROSS_COMPILE)) || (defined(TARGET_X86))
 
 /*HACK, move this to an eventual mono-signal.c*/
 #if defined( __linux__) || defined(__sun) || defined(__APPLE__) || defined(__NetBSD__) || \
@@ -43,8 +87,11 @@ typedef struct __darwin_xmm_reg MonoContextSimdReg;
 #endif
 #endif
 
-#if defined(__native_client__)
-#undef MONO_SIGNAL_USE_UCONTEXT_T
+#ifdef __HAIKU__
+/* sigcontext surrogate */
+struct sigcontext {
+	vregs regs;
+};
 #endif
 
 #ifdef HOST_WIN32
@@ -74,14 +121,14 @@ struct sigcontext {
 # define SC_ESI sc_esi
 #elif defined(__HAIKU__)
 # define SC_EAX regs.eax
-# define SC_EBX regs._reserved_2[2]
+# define SC_EBX regs.ebx
 # define SC_ECX regs.ecx
 # define SC_EDX regs.edx
 # define SC_EBP regs.ebp
 # define SC_EIP regs.eip
 # define SC_ESP regs.esp
-# define SC_EDI regs._reserved_2[0]
-# define SC_ESI regs._reserved_2[1]
+# define SC_EDI regs.edi
+# define SC_ESI regs.esi
 #else
 # define SC_EAX eax
 # define SC_EBX ebx
@@ -139,7 +186,8 @@ typedef struct {
 		 }																\
 	} while (0)
 #else
-#define MONO_CONTEXT_GET_CURRENT(ctx) \
+
+#define MONO_CONTEXT_GET_CURRENT_GREGS(ctx) \
 	__asm__ __volatile__(   \
 	"movl $0x0, %c[eax](%0)\n" \
 	"mov %%ebx, %c[ebx](%0)\n" \
@@ -162,6 +210,40 @@ typedef struct {
 		[esi] MONO_CONTEXT_OFFSET (esi, 0, mgreg_t), \
 		[edi] MONO_CONTEXT_OFFSET (edi, 0, mgreg_t) \
 	: "memory")
+
+#ifdef UCONTEXT_REG_XMM
+#define MONO_CONTEXT_GET_CURRENT_FREGS(ctx) \
+	do { \
+		__asm__ __volatile__ ( \
+			"movups %%xmm0, %c[xmm0](%0)\n"	\
+			"movups %%xmm1, %c[xmm1](%0)\n"	\
+			"movups %%xmm2, %c[xmm2](%0)\n"	\
+			"movups %%xmm3, %c[xmm3](%0)\n"	\
+			"movups %%xmm4, %c[xmm4](%0)\n"	\
+			"movups %%xmm5, %c[xmm5](%0)\n"	\
+			"movups %%xmm6, %c[xmm6](%0)\n"	\
+			"movups %%xmm7, %c[xmm7](%0)\n"	\
+			: \
+			: "a" (&(ctx)),	\
+				[xmm0] MONO_CONTEXT_OFFSET (fregs, X86_XMM0, MonoContextSimdReg), \
+				[xmm1] MONO_CONTEXT_OFFSET (fregs, X86_XMM1, MonoContextSimdReg), \
+				[xmm2] MONO_CONTEXT_OFFSET (fregs, X86_XMM2, MonoContextSimdReg), \
+				[xmm3] MONO_CONTEXT_OFFSET (fregs, X86_XMM3, MonoContextSimdReg), \
+				[xmm4] MONO_CONTEXT_OFFSET (fregs, X86_XMM4, MonoContextSimdReg), \
+				[xmm5] MONO_CONTEXT_OFFSET (fregs, X86_XMM5, MonoContextSimdReg), \
+				[xmm6] MONO_CONTEXT_OFFSET (fregs, X86_XMM6, MonoContextSimdReg), \
+				[xmm7] MONO_CONTEXT_OFFSET (fregs, X86_XMM7, MonoContextSimdReg)); \
+	} while (0)
+#else
+#define MONO_CONTEXT_GET_CURRENT_FREGS(ctx)
+#endif
+
+#define MONO_CONTEXT_GET_CURRENT(ctx) \
+    do {	\
+		MONO_CONTEXT_GET_CURRENT_GREGS(ctx);	\
+		MONO_CONTEXT_GET_CURRENT_FREGS(ctx);	\
+	} while (0)
+
 #endif
 
 #define MONO_ARCH_HAS_MONO_CONTEXT 1
@@ -170,17 +252,28 @@ typedef struct {
 
 #include <mono/arch/amd64/amd64-codegen.h>
 
-#if !defined( HOST_WIN32 ) && !defined(__native_client__) && !defined(__native_client_codegen__)
+#if !defined( HOST_WIN32 )
 
-#if defined(HAVE_SIGACTION) || defined(__APPLE__)  // the __APPLE__ check is required for the tvos simulator, which has ucontext_t but not sigaction
+// the __APPLE__ check is required for the tvos simulator, which has ucontext_t but not sigaction
+#if defined(HAVE_SIGACTION) || defined(__APPLE__)
 #define MONO_SIGNAL_USE_UCONTEXT_T 1
 #endif
 
 #endif
 
+#ifdef __HAIKU__
+/* sigcontext surrogate */
+struct sigcontext {
+	vregs regs;
+};
+
+// Haiku doesn't support this
+#undef MONO_SIGNAL_USE_UCONTEXT_T
+#endif
+
 typedef struct {
 	mgreg_t gregs [AMD64_NREG];
-#ifdef __APPLE__
+#if defined(MONO_HAVE_SIMD_REG)
 	MonoContextSimdReg fregs [AMD64_XMM_NREG];
 #else
 	double fregs [AMD64_XMM_NREG];
@@ -197,33 +290,10 @@ typedef struct {
 
 #if defined (HOST_WIN32) && !defined(__GNUC__)
 /* msvc doesn't support inline assembly, so have to use a separate .asm file */
-extern void mono_context_get_current (void *);
+// G_EXTERN_C due to being written in assembly.
+G_EXTERN_C void mono_context_get_current (void *);
 #define MONO_CONTEXT_GET_CURRENT(ctx) do { mono_context_get_current((void*)&(ctx)); } while (0)
 
-#elif defined(__native_client__)
-#define MONO_CONTEXT_GET_CURRENT(ctx)	\
-	__asm__ __volatile__(	\
-		"movq $0x0,  %%nacl:0x00(%%r15, %0, 1)\n"	\
-		"movq %%rcx, %%nacl:0x08(%%r15, %0, 1)\n"	\
-		"movq %%rdx, %%nacl:0x10(%%r15, %0, 1)\n"	\
-		"movq %%rbx, %%nacl:0x18(%%r15, %0, 1)\n"	\
-		"movq %%rsp, %%nacl:0x20(%%r15, %0, 1)\n"	\
-		"movq %%rbp, %%nacl:0x28(%%r15, %0, 1)\n"	\
-		"movq %%rsi, %%nacl:0x30(%%r15, %0, 1)\n"	\
-		"movq %%rdi, %%nacl:0x38(%%r15, %0, 1)\n"	\
-		"movq %%r8,  %%nacl:0x40(%%r15, %0, 1)\n"	\
-		"movq %%r9,  %%nacl:0x48(%%r15, %0, 1)\n"	\
-		"movq %%r10, %%nacl:0x50(%%r15, %0, 1)\n"	\
-		"movq %%r11, %%nacl:0x58(%%r15, %0, 1)\n"	\
-		"movq %%r12, %%nacl:0x60(%%r15, %0, 1)\n"	\
-		"movq %%r13, %%nacl:0x68(%%r15, %0, 1)\n"	\
-		"movq %%r14, %%nacl:0x70(%%r15, %0, 1)\n"	\
-		"movq %%r15, %%nacl:0x78(%%r15, %0, 1)\n"	\
-		"leaq (%%rip), %%rdx\n"	\
-		"movq %%rdx, %%nacl:0x80(%%r15, %0, 1)\n"	\
-		: 	\
-		: "a" ((int64_t)&(ctx))	\
-		: "rdx", "memory")
 #else
 
 #define MONO_CONTEXT_GET_CURRENT_GREGS(ctx) \
@@ -346,7 +416,7 @@ typedef struct {
 
 #define MONO_CONTEXT_GET_CURRENT(ctx) do { \
 	gpointer _dummy; \
-    ctx.regs [ARMREG_SP] = &_dummy; \
+    ctx.regs [ARMREG_SP] = (mgreg_t)&_dummy; \
 } while (0);
 
 #else
@@ -384,7 +454,8 @@ typedef struct {
 
 typedef struct {
 	mgreg_t regs [32];
-	double fregs [32];
+	/* FIXME not fully saved in trampolines */
+	MonoContextSimdReg fregs [32];
 	mgreg_t pc;
 	/*
 	 * fregs might not be initialized if this context was created from a
@@ -405,15 +476,27 @@ typedef struct {
 
 #define MONO_CONTEXT_GET_CURRENT(ctx) do { \
 	arm_unified_thread_state_t thread_state;	\
+	arm_neon_state64_t thread_fpstate;		\
 	int state_flavor = ARM_UNIFIED_THREAD_STATE;	\
+	int fpstate_flavor = ARM_NEON_STATE64;	\
 	unsigned state_count = ARM_UNIFIED_THREAD_STATE_COUNT;	\
+	unsigned fpstate_count = ARM_NEON_STATE64_COUNT;	\
 	thread_port_t self = mach_thread_self ();	\
 	kern_return_t ret = thread_get_state (self, state_flavor, (thread_state_t) &thread_state, &state_count);	\
 	g_assert (ret == 0);	\
-	mono_mach_arch_thread_states_to_mono_context ((thread_state_t)&thread_state, (thread_state_t)NULL, &ctx); \
+	ret = thread_get_state (self, fpstate_flavor, (thread_state_t) &thread_fpstate, &fpstate_count);	\
+	g_assert (ret == 0);	\
+	mono_mach_arch_thread_states_to_mono_context ((thread_state_t) &thread_state, (thread_state_t) &thread_fpstate, &ctx); \
 	mach_port_deallocate (current_task (), self);	\
 } while (0);
 
+#elif defined(HOST_WATCHOS)
+
+#define MONO_CONTEXT_GET_CURRENT(ctx) do { \
+	gpointer _dummy; \
+	ctx.regs [ARMREG_SP] = (mgreg_t)(gsize)&_dummy; \
+} while (0);
+
 #else
 
 #define MONO_CONTEXT_GET_CURRENT(ctx)	do { 	\
@@ -434,12 +517,28 @@ typedef struct {
 		"stp x24, x25, [x16], #16\n"	\
 		"stp x26, x27, [x16], #16\n"	\
 		"stp x28, x29, [x16], #16\n"	\
-		"stp x30, xzr, [x16]\n"	\
+		"stp x30, xzr, [x16], #8\n"	\
 		"mov x30, sp\n"				\
-		"str x30, [x16, #8]\n"		\
+		"str x30, [x16], #8\n"		\
+		"stp q0, q1, [x16], #32\n"	\
+		"stp q2, q3, [x16], #32\n"	\
+		"stp q4, q5, [x16], #32\n"	\
+		"stp q6, q7, [x16], #32\n"	\
+		"stp q8, q9, [x16], #32\n"	\
+		"stp q10, q11, [x16], #32\n"	\
+		"stp q12, q13, [x16], #32\n"	\
+		"stp q14, q15, [x16], #32\n"	\
+		"stp q16, q17, [x16], #32\n"	\
+		"stp q18, q19, [x16], #32\n"	\
+		"stp q20, q21, [x16], #32\n"	\
+		"stp q22, q23, [x16], #32\n"	\
+		"stp q24, q25, [x16], #32\n"	\
+		"stp q26, q27, [x16], #32\n"	\
+		"stp q28, q29, [x16], #32\n"	\
+		"stp q30, q31, [x16], #32\n"		\
 		:							\
 		: "r" (&ctx.regs)			\
-		: "x30", "memory"			\
+		: "x16", "x30", "memory"		\
 	);								\
 	__asm__ __volatile__( \
 		"adr %0, L0%=\n" \
@@ -477,7 +576,7 @@ typedef struct {
 #define MONO_CONTEXT_SET_SP(ctx,sp) do { (ctx)->sc_sp = (gulong)sp; } while (0);
 
 #define MONO_CONTEXT_GET_IP(ctx) ((gpointer)((ctx)->sc_ir))
-#define MONO_CONTEXT_GET_BP(ctx) ((gpointer)((ctx)->regs [ppc_r31-13]))
+#define MONO_CONTEXT_GET_BP(ctx) ((gpointer)((ctx)->regs [ppc_r31]))
 #define MONO_CONTEXT_GET_SP(ctx) ((gpointer)((ctx)->sc_sp))
 
 #define MONO_CONTEXT_GET_CURRENT(ctx)	\
@@ -568,45 +667,45 @@ typedef struct {
 #define MONO_CONTEXT_SET_SP(ctx,sp) do { (ctx)->sc_sp = (mgreg_t)sp; } while (0);
 
 #define MONO_CONTEXT_GET_IP(ctx) ((gpointer)((ctx)->sc_ir))
-#define MONO_CONTEXT_GET_BP(ctx) ((gpointer)((ctx)->regs [ppc_r31-13]))
+#define MONO_CONTEXT_GET_BP(ctx) ((gpointer)((ctx)->regs [ppc_r31]))
 #define MONO_CONTEXT_GET_SP(ctx) ((gpointer)((ctx)->sc_sp))
 
 #define MONO_CONTEXT_GET_CURRENT(ctx)	\
 	__asm__ __volatile__(	\
-		"std 0, 0(%0)\n"	\
-		"std 1, 4(%0)\n"	\
-		"std 0, 4*0+8(%0)\n"	\
-		"std 1, 4*1+8(%0)\n"	\
-		"std 2, 4*2+8(%0)\n"	\
-		"std 3, 4*3+8(%0)\n"	\
-		"std 4, 4*4+8(%0)\n"	\
-		"std 5, 4*5+8(%0)\n"	\
-		"std 6, 4*6+8(%0)\n"	\
-		"std 7, 4*7+8(%0)\n"	\
-		"std 8, 4*8+8(%0)\n"	\
-		"std 9, 4*9+8(%0)\n"	\
-		"std 10, 4*10+8(%0)\n"	\
-		"std 11, 4*11+8(%0)\n"	\
-		"std 12, 4*12+8(%0)\n"	\
-		"std 13, 4*13+8(%0)\n"	\
-		"std 14, 4*14+8(%0)\n"	\
-		"std 15, 4*15+8(%0)\n"	\
-		"std 16, 4*16+8(%0)\n"	\
-		"std 17, 4*17+8(%0)\n"	\
-		"std 18, 4*18+8(%0)\n"	\
-		"std 19, 4*19+8(%0)\n"	\
-		"std 20, 4*20+8(%0)\n"	\
-		"std 21, 4*21+8(%0)\n"	\
-		"std 22, 4*22+8(%0)\n"	\
-		"std 23, 4*23+8(%0)\n"	\
-		"std 24, 4*24+8(%0)\n"	\
-		"std 25, 4*25+8(%0)\n"	\
-		"std 26, 4*26+8(%0)\n"	\
-		"std 27, 4*27+8(%0)\n"	\
-		"std 28, 4*28+8(%0)\n"	\
-		"std 29, 4*29+8(%0)\n"	\
-		"std 30, 4*30+8(%0)\n"	\
-		"std 31, 4*31+8(%0)\n"	\
+		"stw 0, 0(%0)\n"	\
+		"stw 1, 4(%0)\n"	\
+		"stw 0, 4*0+8(%0)\n"	\
+		"stw 1, 4*1+8(%0)\n"	\
+		"stw 2, 4*2+8(%0)\n"	\
+		"stw 3, 4*3+8(%0)\n"	\
+		"stw 4, 4*4+8(%0)\n"	\
+		"stw 5, 4*5+8(%0)\n"	\
+		"stw 6, 4*6+8(%0)\n"	\
+		"stw 7, 4*7+8(%0)\n"	\
+		"stw 8, 4*8+8(%0)\n"	\
+		"stw 9, 4*9+8(%0)\n"	\
+		"stw 10, 4*10+8(%0)\n"	\
+		"stw 11, 4*11+8(%0)\n"	\
+		"stw 12, 4*12+8(%0)\n"	\
+		"stw 13, 4*13+8(%0)\n"	\
+		"stw 14, 4*14+8(%0)\n"	\
+		"stw 15, 4*15+8(%0)\n"	\
+		"stw 16, 4*16+8(%0)\n"	\
+		"stw 17, 4*17+8(%0)\n"	\
+		"stw 18, 4*18+8(%0)\n"	\
+		"stw 19, 4*19+8(%0)\n"	\
+		"stw 20, 4*20+8(%0)\n"	\
+		"stw 21, 4*21+8(%0)\n"	\
+		"stw 22, 4*22+8(%0)\n"	\
+		"stw 23, 4*23+8(%0)\n"	\
+		"stw 24, 4*24+8(%0)\n"	\
+		"stw 25, 4*25+8(%0)\n"	\
+		"stw 26, 4*26+8(%0)\n"	\
+		"stw 27, 4*27+8(%0)\n"	\
+		"stw 28, 4*28+8(%0)\n"	\
+		"stw 29, 4*29+8(%0)\n"	\
+		"stw 30, 4*30+8(%0)\n"	\
+		"stw 31, 4*31+8(%0)\n"	\
 		"stfd 0, 8*0+4*32+8(%0)\n"	\
 		"stfd 1, 8*1+4*32+8(%0)\n"	\
 		"stfd 2, 8*2+4*32+8(%0)\n"	\
@@ -640,7 +739,7 @@ typedef struct {
 		"stfd 30, 8*30+4*32+8(%0)\n"	\
 		"stfd 31, 8*31+4*32+8(%0)\n"	\
 		: : "r" (&(ctx))	\
-		: "memory"			\
+		: "memory", "r0"	\
 	)
 
 #endif
@@ -712,89 +811,9 @@ typedef struct MonoContext {
 
 #define MONO_ARCH_HAS_MONO_CONTEXT 1
 
-#elif defined(__ia64__) /*defined(__sparc__) || defined(sparc) */
+#elif ((defined(__mips__) && !defined(MONO_CROSS_COMPILE)) || (defined(TARGET_MIPS))) && SIZEOF_REGISTER == 4
 
-#ifndef UNW_LOCAL_ONLY
-
-#define UNW_LOCAL_ONLY
-#include <libunwind.h>
-
-#endif
-
-typedef struct MonoContext {
-	unw_cursor_t cursor;
-	/* Whenever the ip in 'cursor' points to the ip where the exception happened */
-	/* This is true for the initial context for exceptions thrown from signal handlers */
-	gboolean precise_ip;
-} MonoContext;
-
-/*XXX SET_BP is missing*/
-#define MONO_CONTEXT_SET_IP(ctx,eip) do { int err = unw_set_reg (&(ctx)->cursor, UNW_IA64_IP, (unw_word_t)(eip)); g_assert (err == 0); } while (0)
-#define MONO_CONTEXT_SET_SP(ctx,esp) do { int err = unw_set_reg (&(ctx)->cursor, UNW_IA64_SP, (unw_word_t)(esp)); g_assert (err == 0); } while (0)
-
-#define MONO_CONTEXT_GET_IP(ctx) ((gpointer)(mono_ia64_context_get_ip ((ctx))))
-#define MONO_CONTEXT_GET_BP(ctx) ((gpointer)(mono_ia64_context_get_fp ((ctx))))
-#define MONO_CONTEXT_GET_SP(ctx) ((gpointer)(mono_ia64_context_get_sp ((ctx))))
-
-static inline unw_word_t
-mono_ia64_context_get_ip (MonoContext *ctx)
-{
-	unw_word_t ip;
-	int err;
-
-	err = unw_get_reg (&ctx->cursor, UNW_IA64_IP, &ip);
-	g_assert (err == 0);
-
-	if (ctx->precise_ip) {
-		return ip;
-	} else {
-		/* Subtrack 1 so ip points into the actual instruction */
-		return ip - 1;
-	}
-}
-
-static inline unw_word_t
-mono_ia64_context_get_sp (MonoContext *ctx)
-{
-	unw_word_t sp;
-	int err;
-
-	err = unw_get_reg (&ctx->cursor, UNW_IA64_SP, &sp);
-	g_assert (err == 0);
-
-	return sp;
-}
-
-static inline unw_word_t
-mono_ia64_context_get_fp (MonoContext *ctx)
-{
-	unw_cursor_t new_cursor;
-	unw_word_t fp;
-	int err;
-
-	{
-		unw_word_t ip, sp;
-
-		err = unw_get_reg (&ctx->cursor, UNW_IA64_SP, &sp);
-		g_assert (err == 0);
-
-		err = unw_get_reg (&ctx->cursor, UNW_IA64_IP, &ip);
-		g_assert (err == 0);
-	}
-
-	/* fp is the SP of the parent frame */
-	new_cursor = ctx->cursor;
-
-	err = unw_step (&new_cursor);
-	g_assert (err >= 0);
-
-	err = unw_get_reg (&new_cursor, UNW_IA64_SP, &fp);
-	g_assert (err == 0);
-
-	return fp;
-}
-
-#elif ((defined(__mips__) && !defined(MONO_CROSS_COMPILE)) || (defined(TARGET_MIPS))) && SIZEOF_REGISTER == 4 /* defined(__ia64__) */
+#define MONO_ARCH_HAS_MONO_CONTEXT 1
 
 #include <mono/arch/mips/mips-codegen.h>
 
@@ -854,7 +873,13 @@ typedef struct {
 
 #define MONO_ARCH_HAS_MONO_CONTEXT 1
 
+#include <sys/ucontext.h>
+
+#if __GLIBC_PREREQ(2, 27)
+typedef ucontext_t MonoContext;
+#else
 typedef struct ucontext MonoContext;
+#endif
 
 #define MONO_CONTEXT_SET_IP(ctx,ip) 					\
 	do {								\
