@@ -1,6 +1,7 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
-/*
- * mono-os-mutex.h: Portability wrappers around POSIX Mutexes
+/**
+ * \file
+ * Portability wrappers around POSIX Mutexes
  *
  * Authors: Jeffrey Stedfast <fejj@ximian.com>
  *
@@ -40,6 +41,10 @@ G_BEGIN_DECLS
 
 #if !defined(HOST_WIN32)
 
+#if !defined(CLOCK_MONOTONIC) || defined(HOST_DARWIN) || defined(HOST_ANDROID) || defined(HOST_WASM)
+#define BROKEN_CLOCK_SOURCE
+#endif
+
 typedef pthread_mutex_t mono_mutex_t;
 typedef pthread_cond_t mono_cond_t;
 
@@ -76,16 +81,14 @@ mono_os_mutex_init_recursive (mono_mutex_t *mutex)
 		g_error ("%s: pthread_mutexattr_destroy failed with \"%s\" (%d)", __func__, g_strerror (res), res);
 }
 
-static inline int
+static inline void
 mono_os_mutex_destroy (mono_mutex_t *mutex)
 {
 	int res;
 
 	res = pthread_mutex_destroy (mutex);
-	if (G_UNLIKELY (res != 0 && res != EBUSY))
+	if (G_UNLIKELY (res != 0))
 		g_error ("%s: pthread_mutex_destroy failed with \"%s\" (%d)", __func__, g_strerror (res), res);
-
-	return res != 0 ? -1 : 0;
 }
 
 static inline void
@@ -125,21 +128,41 @@ mono_os_cond_init (mono_cond_t *cond)
 {
 	int res;
 
+#ifdef BROKEN_CLOCK_SOURCE
 	res = pthread_cond_init (cond, NULL);
 	if (G_UNLIKELY (res != 0))
 		g_error ("%s: pthread_cond_init failed with \"%s\" (%d)", __func__, g_strerror (res), res);
+#else
+	/* POSIX standard does not compel to have CLOCK_MONOTONIC */
+	pthread_condattr_t attr;
+
+	res = pthread_condattr_init (&attr);
+	if (G_UNLIKELY (res != 0))
+		g_error ("%s: pthread_condattr_init failed with \"%s\" (%d)", __func__, g_strerror (res), res);
+
+	res = pthread_condattr_setclock (&attr, CLOCK_MONOTONIC);
+	if (G_UNLIKELY (res != 0))
+		g_error ("%s: pthread_condattr_setclock failed with \"%s\" (%d)", __func__, g_strerror (res), res);
+
+	/* Attach an attribute having CLOCK_MONOTONIC to condition */
+	res = pthread_cond_init (cond, &attr);
+	if (G_UNLIKELY (res != 0))
+		g_error ("%s: pthread_cond_init failed with \"%s\" (%d)", __func__, g_strerror (res), res);
+
+	res = pthread_condattr_destroy (&attr);
+	if (G_UNLIKELY (res != 0))
+		g_error ("%s: pthread_condattr_destroy failed with \"%s\" (%d)", __func__, g_strerror (res), res);
+#endif
 }
 
-static inline int
+static inline void
 mono_os_cond_destroy (mono_cond_t *cond)
 {
 	int res;
 
 	res = pthread_cond_destroy (cond);
-	if (G_UNLIKELY (res != 0 && res != EBUSY))
+	if (G_UNLIKELY (res != 0))
 		g_error ("%s: pthread_cond_destroy failed with \"%s\" (%d)", __func__, g_strerror (res), res);
-
-	return res != 0 ? -1 : 0;
 }
 
 static inline void
@@ -152,40 +175,8 @@ mono_os_cond_wait (mono_cond_t *cond, mono_mutex_t *mutex)
 		g_error ("%s: pthread_cond_wait failed with \"%s\" (%d)", __func__, g_strerror (res), res);
 }
 
-static inline int
-mono_os_cond_timedwait (mono_cond_t *cond, mono_mutex_t *mutex, guint32 timeout_ms)
-{
-	struct timeval tv;
-	struct timespec ts;
-	gint64 usecs;
-	int res;
-
-	if (timeout_ms == MONO_INFINITE_WAIT) {
-		mono_os_cond_wait (cond, mutex);
-		return 0;
-	}
-
-	/* ms = 10^-3, us = 10^-6, ns = 10^-9 */
-
-	res = gettimeofday (&tv, NULL);
-	if (G_UNLIKELY (res != 0))
-		g_error ("%s: gettimeofday failed with \"%s\" (%d)", __func__, g_strerror (errno), errno);
-
-	tv.tv_sec += timeout_ms / 1000;
-	usecs = tv.tv_usec + ((timeout_ms % 1000) * 1000);
-	if (usecs >= 1000000) {
-		usecs -= 1000000;
-		tv.tv_sec ++;
-	}
-	ts.tv_sec = tv.tv_sec;
-	ts.tv_nsec = usecs * 1000;
-
-	res = pthread_cond_timedwait (cond, mutex, &ts);
-	if (G_UNLIKELY (res != 0 && res != ETIMEDOUT))
-		g_error ("%s: pthread_cond_timedwait failed with \"%s\" (%d)", __func__, g_strerror (res), res);
-
-	return res != 0 ? -1 : 0;
-}
+int
+mono_os_cond_timedwait (mono_cond_t *cond, mono_mutex_t *mutex, guint32 timeout_ms);
 
 static inline void
 mono_os_cond_signal (mono_cond_t *cond)
@@ -296,11 +287,10 @@ mono_os_mutex_init_recursive (mono_mutex_t *mutex)
 		g_error ("%s: InitializeCriticalSectionEx failed with error %d", __func__, GetLastError ());
 }
 
-static inline int
+static inline void
 mono_os_mutex_destroy (mono_mutex_t *mutex)
 {
 	DeleteCriticalSection (mutex);
-	return 0;
 }
 
 static inline void
@@ -327,11 +317,10 @@ mono_os_cond_init (mono_cond_t *cond)
 	InitializeConditionVariable (cond);
 }
 
-static inline int
+static inline void
 mono_os_cond_destroy (mono_cond_t *cond)
 {
 	/* Beauty of win32 API: do not destroy it */
-	return 0;
 }
 
 static inline void
